// noErfRelativeOrientation.h++

#ifndef _noErfRelativeOrientation_H_
#define _noErfRelativeOrientation_H_

#include <math.h>
#include "qmNonlinearOptimize/noErrorFunction.h++"

class noErfRelativeOrientation : public noErrorFunction
{
    qmVector uL;
    qmVector vL;
    qmVector uR;
    qmVector vR;
    int pointCount;
    real beta;
    static int verbose;
    // int fixedPointIndex;
    int fixedPointCount;
    int* fixedPointIndices;
    // real* fixedPointDepths;
public:
    enum { TX, TY, TZB, Q0, QX, QY, QZ, ALPHA };
    enum { DTX, DTY, DTZB, DWX, DWY, DWZ, DALPHA };
    noErfRelativeOrientation() {}
    void init( qmVectorRef& uL_in,
	       qmVectorRef& vL_in,
	       qmVectorRef& uR_in,
	       qmVectorRef& vR_in );
    const qmVectorRef& operator()( const qmVectorRef& x );
    const qmMatrixRef& jacobian( const qmVectorRef& x );
    const qmVectorRef& add( const qmVectorRef& x, const qmVectorRef& dx );
    static void setVerbose( int v ) { verbose = v; }
    // inline void setFixedPoint( int );
//     static inline int nMmts( int pointCount ) { return 2*pointCount; }
//     static inline int nStates( int pointCount ) { return 7+pointCount; }
//     static inline int nDStates( int pointCount ); { return 6+pointCount; }
    // pointCount & fixedPointCount must be set
    inline int nMmts() { return 2 * pointCount; }
    inline int nStates() { return 7 + pointCount; }
    inline int nDStates() { return( 6 + pointCount - fixedPointCount); }
    void setBeta( real b ) { beta = b; }
    void setFixedPoint( int );
    void setFixedPoints( int, int* );
    inline int isFixedPoint( int );
};

// inline
// void
// noErfRelativeOrientation::setFixedPoint( int i )
// {
//     fixedPointIndex = i;
// }

inline
int
noErfRelativeOrientation::isFixedPoint( int id )
{
    for( int k = 0; k < fixedPointCount; k++ )
       if( id == fixedPointIndices[k] )
	  return 1;
    return 0;
}

#endif /* _noErfRelativeOrientation_H_ */
