// IvCamera.c++

#include "IvCamera.h++"

IvCamera::IvCamera( Widget wgt )
  : representation(),
    view( wgt ),
    overlay(),
    focalLength( 1. ),
    focalLengthNode( new SoPerspectiveCamera ),
    coordinateTransform(),
    coordinateTransformNode( new SoTransform )
{
    focalLengthNode -> ref();
    coordinateTransformNode -> ref();
#if 0
    representation.addInteriorOrientationCB
       ( IvCameraView::interiorOrientationCB, &view );
    representation.addRelativeOrientationCB
       ( IvCameraView::relativeOrientationCB, &view );
#endif
    overlay.attachTo( view.getRenderArea() );

    representation.attachInteriorOrientation( focalLengthNode );
    representation.attachCoordinateTransform( coordinateTransformNode );
    view.attachInteriorOrientation( focalLengthNode );
    view.attachCoordinateTransform( coordinateTransformNode );

    representation.driveInteriorOrientation( focalLengthNode );
    representation.driveCoordinateTransform( coordinateTransformNode );

    setInteriorOrientation( 1. );
}

void
IvCamera::setInteriorOrientation( float beta )
{
    if( beta < .0001 )
       beta = .0001;
    focalLength = 1. / beta;
    focalLengthNode -> position.setValue( 0., 0., -1./beta );
    focalLengthNode -> orientation.setValue( 1., 0., 0., 0. );
    focalLengthNode -> heightAngle = 2. * atan( .375 * beta );
    focalLengthNode -> aspectRatio = 4./3.;
    focalLengthNode -> nearDistance = .1/ beta;
    focalLengthNode -> farDistance = 100.;
}


void
IvCamera::project( SbVec3f& pW, SbVec2f& q )
{
    SbVec3f pC;

    SoGetMatrixAction *gma = new SoGetMatrixAction(SbViewportRegion());
    gma->apply( coordinateTransformNode );
    coordinateTransform = gma->getMatrix();

    coordinateTransform.multVecMatrix( pW, pC );
    focalLength = (tan(focalLengthNode->heightAngle.getValue() / 2.)/.375);
    float beta = 1. / focalLength;
    q[0] = pC[0] / ( 1. + beta * pC[2] );
    q[1] = pC[1] / ( 1. + beta * pC[2] );

    delete gma;
}

const SbVec2f&
IvCamera::getProjection( SbVec3f& pW )
{
    static SbVec2f q;
    project( pW, q );
    return q;
}

const SbVec3f&
IvCamera::getCamera3DPoint( const SbVec3f& pW )
{
    static SbVec3f pC;
    SoGetMatrixAction *gma = new SoGetMatrixAction(SbViewportRegion());

    gma->apply( coordinateTransformNode );
    coordinateTransform = gma->getMatrix();
    coordinateTransform.multVecMatrix( pW, pC );

    delete gma;
    return pC;
}

void
IvCamera::getCoordinateTransform( float& tx, float& ty, float& tz,
				  float& q0, float& q1,
				  float& q2, float& q3 )
{
    coordinateTransformNode -> translation.getValue().getValue( tx, ty, tz );
    coordinateTransformNode -> rotation.getValue().getValue( q1, q2, q3, q0 );
}

void
IvCamera::getCoordinateTransform( SbVec3f& t, SbRotation& q )
{
    t = coordinateTransformNode -> translation.getValue();
    q = coordinateTransformNode -> rotation.getValue();
}

int
IvCamera::drawBlob( int id,
		    float um,
		    float vm,
		    float x0x,
		    float x0y, 
		    float l0,
		    float l1 )
{
    float delta[2];
    delta[0] = 2. * sqrt( l0 );
    delta[1] = 2. * sqrt( l1 );
    float p[4][2];
    p[0][0] = um - delta[0] * x0x;
    p[0][1] = vm - delta[0] * x0y;
    p[1][0] = um + delta[0] * x0x;
    p[1][1] = vm + delta[0] * x0y;
    p[2][0] = um - delta[1] * -x0y;
    p[2][1] = vm - delta[1] * x0x;
    p[3][0] = um + delta[1] * -x0y;
    p[3][1] = vm + delta[1] * x0x;
    if( id < 0 )
       {
	   id = newLine( p[0][0], p[0][1], p[1][0], p[1][1] );
	   newLine( p[2][0], p[2][1], p[3][0], p[3][1] );
	   return id;
       }
    redrawLine(   id, p[0][0], p[0][1], p[1][0], p[1][1] );
    redrawLine( id+1, p[2][0], p[2][1], p[3][0], p[3][1] );
    return id;
}

void
IvCamera::setSize( int h, int w, int y, int x )
{
    view.setSize( h, w, y, x );
    overlay.setSize( h, w );
    representation.setSize( h, w );
}
