// IvCamera.h++

// In this object, "reference transform" and "coordinate transform" are
// of the camera frame relative to the world.  Thusly, rt is where the camera
// is relative to the world, ct is where the world is relative to the camera

#ifndef _IvCamera_H_
#define _IvCamera_H_

#include "IvCamera/IvCameraRep.h++"
#include "IvCamera/IvCameraView.h++"
#include "IvCamera/IvCameraOverlay.h++"

class IvCamera
{
    IvCameraRep representation;
    IvCameraView view;
    IvCameraOverlay overlay;
    float focalLength;
    SbMatrix coordinateTransform;
public:
    SoPerspectiveCamera* focalLengthNode;
    SoTransform* coordinateTransformNode;

public:
    IvCamera( Widget = NULL );

    // Representation only
    inline void setColor( float r, float g, float b )
    { representation.setColor( r, g, b ); }
    inline int select( SoPath* p ) { return representation.select(p); }
    inline int deselect( SoPath* p ) { return representation.deselect(p); }
    SoSeparator* getRep() { return representation.getNode(); }
    IvCameraRep& getRepClass() { return representation; }
    inline int drawRay( int i, float u, float v, float depth = 4 );

    // View only
    inline void setBackground( void *p, GLenum f = GL_RGB ) 
    { view.setBackground( p, f ); }
    inline void addToView( SoNode* n ) { view.addToView( n ); }
    inline void setTitle( char *s ) { view.setTitle( s ); }
    inline int findColorBlob( SbColor& c, float* mean, float* cov )
    { return view.findColorBlob( c, mean, cov ); }
  // inline void setBackground( void* p ) { view.setBackground(p); }
    inline void setBackgroundColor( float r, float g, float b ) 
    { view.setBackgroundColor( r, g, b ); }
    inline Widget getWidget() { return view.getWidget(); }
    inline void setKeyCB( int (*fn)(void*, SoEventCallback*), void* d= 0 )
    { view.setKeyCB( fn, d ); }
    inline void setMouseCB( int (*fn)(void*, SoEventCallback*), void* d= 0 )
    { view.setMouseCB( fn, d ); }
    inline SoXtRenderArea* getRenderArea() { return view.getRenderArea(); }
    inline void handle() { view.handle(); }
    IvCameraView& getView() { return view; }
    inline void setLightModel( int m ) { view.setLightModel(m); }
    inline SoSeparator* getViewRoot() { return view.getRoot(); }

    // Overlay only
    inline int newLine( float x0, float y0, float x1, float y1, int c=1 )
    { return overlay.newLine( x0, y0, x1, y1, c ); }
    inline void redrawLine( int id, float x0, float y0, float x1, float y1,
			    int c=1 )
    { overlay.redrawLine( id, x0, y0, x1, y1, c ); }
    inline void eraseLines( int stid = 0 ) { overlay.eraseLines(stid); }
    inline IvCameraOverlay* getOverlay() { return &overlay; }
    int drawBlob( int id, float um, float vm, float x0x, float x0y, 
		  float l0, float l1 );

    // Coordinated
    void setSize( int h, int w, int y = 0, int x = 0 );
    // { view.setSize( h, w, y, x ); }
    void setInteriorOrientation( float beta );
    inline void setCoordinateTransform( float tx, float ty, float tz,
					float q0, float q1,
					float q2, float q3 );
    inline void setReferenceTransform( float tx, float ty, float tz,
				       float q0, float q1,
				       float q2, float q3 );
    inline float getFocalLength() { return focalLength; }
    void project( SbVec3f&, SbVec2f& );
    const SbVec2f& getProjection( SbVec3f& );
    const SbVec3f& getCamera3DPoint( const SbVec3f& );
    void getCoordinateTransform( float& tx, float& ty, float& tz,
				 float& q0, float& q1,
				 float& q2, float& q3 );
    void getCoordinateTransform( SbVec3f&, SbRotation& );
    const SbMatrix& getCoordinateTransform() { return coordinateTransform; }
};

inline
void
IvCamera::setCoordinateTransform( float tx, float ty, float tz,
				  float q0, float q1, float q2, float q3 )
{
    SbVec3f scale( 1., 1., 1. );
    SbRotation rot( q1, q2, q3, q0 );
    SbVec3f tr( tx, ty, tz );
    coordinateTransform.setTransform( tr, rot, scale );
    coordinateTransformNode -> setMatrix( coordinateTransform );
    coordinateTransformNode -> touch();
}

inline
void
IvCamera::setReferenceTransform( float tx, float ty, float tz,
				 float q0, float q1, float q2, float q3 )
{
    SbVec3f scale( 1., 1., 1. );
    SbRotation rot( q1, q2, q3, q0 );
    SbVec3f tr( tx, ty, tz );
    coordinateTransform.setTransform( tr, rot, scale );
    coordinateTransform = coordinateTransform.inverse();
    coordinateTransformNode -> setMatrix( coordinateTransform );
    coordinateTransformNode -> touch();
}

inline
int
IvCamera::drawRay( int i, float u, float v, float depth )
{
    return representation.drawRay( i, u, v, depth );
}
 
#endif /* _IvCamera_H_ */
