// IvCameraOverlay.h++

#ifndef _IvCameraOverlay_H_
#define _IvCameraOverlay_H_

#include "IvCamera/IvCameraOverlayLine.h++"
#include "IvCamera/IvCameraTextField.h++"
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoColorIndex.h>
#include <Inventor/Xt/SoXtRenderArea.h>

const int MAXTEXTFIELDS = 16;

class IvCameraOverlay
{
private:

    // Scene graph for overlay on top of camera view
    // SEP
    //  |
    // ORTHO-CAM...LIGHT-MODEL...COLOR-INDEX...SEP
    //                                          |
    //                                       CamLine...CamLine...
    static const int maxOverlayLines;
    IvCameraOverlayLine** lines;
    int nLines;
    SoSeparator* root;
    SoOrthographicCamera* camera;
    SoLightModel* lightModel;
    SoColorIndex* colorIndex;
    SoSeparator* linesRoot;
    // SbColor lineColor;
    SbColor* lineColors;

    IvCameraTextField* textFields[ MAXTEXTFIELDS ];
    int textFieldCount;
    SoSeparator* textRoot;

public:
    IvCameraOverlay();
    ~IvCameraOverlay();
    void attachTo( SoXtRenderArea* );
    int newLine( float x0, float y0, float x1, float y1, int c = 1 );
    void redrawLine( int id, float x0, float y0, float x1, float y1, int c=1 );
    void eraseLines( int startid = 0 );
    void setSize( int h, int w );
    int newTextField();
    void setText( int, char* );
};

#endif /* _IvCameraOverlay_H_ */
