// IvCameraRep.c++

#include "IvCameraRep.h++"

#include <stream.h>
#if 0
static inline void printPath( SoPath* path )
{
    cerr << "Path: " << endl;
    int n = path -> getLength();
    for( int i = 0; i < n; i++ )
       cerr << "  node: " << path->getNode(i)->getName().getString() << endl;
}
#endif

float IvCameraRep::axisCoordinateValues[20][3] =
{ { .5, .375, 0. }, { .5, -.375, 0. }, { -.5, -.375, 0. }, { -.5, .375, 0. },
      { 0., 0., 0. }, { .2, 0., 0. }, { 0., .2, 0. }, { 0., 0., .2 }, 
      { .075, -.025, .2 }, { .025, -.025, .2 }, { .075, -.075, .2 },
      { .025, -.075, .2 }, 
      { .075, .15, 0. }, { .05, .175, 0. }, { .05, .2, 0. }, { .025, .15, 0. },
      { .15, -.025, 0. }, { .2, -.075, 0. }, { .15, -.075, 0. },
      { .2, -.025, 0 } };

// const long IvCameraRep::axisLineCoords[33] =
const int32_t IvCameraRep::axisLineCoords[33] =
{ 0, 1, 2, 3, 0, -1, 4, 5, -1, 4, 6, -1, 4, 7, -1, 8,
      9, 10, 11, -1, 12, 13, 14, -1, 15, 13, -1, 16, 17, -1, 18, 19, -1 };

SoSeparator* IvCameraRep::axisGeometry = 0;
SoCoordinate3* IvCameraRep::axisCoordinates = 0;
SoIndexedLineSet* IvCameraRep::axisLines = 0;
SoSphere* IvCameraRep::copGeometry = 0;
const SbName IvCameraRep::rootName = "Camera_root";
const SbName IvCameraRep::copName = "Camera_copGeometry";

void
IvCameraRep::copDragValueCB( void* cptr, SoDragger* dragger )
{
    IvCameraRep* c = (IvCameraRep*) cptr;
    SoTranslate1Dragger* std = (SoTranslate1Dragger*) dragger;
    float f = std->translation.getValue()[0] / 10. + 1.;
    if( f < .001 )
       f = .001;
    c -> setInteriorOrientation( 1./f );
}

void
IvCameraRep::imagePlaneDragValueCB( void* cptr, SoDragger* dragger )
{
    IvCameraRep* c = (IvCameraRep*) cptr;
    // SoJackDragger* sjd = (SoJackDragger*) dragger;
    SoTransformBoxDragger* sjd = (SoTransformBoxDragger*) dragger;
    SbMatrix tmpm;
    tmpm.setTransform( sjd->translation.getValue(),
		       sjd->rotation.getValue(),
		       SbVec3f( 1., 1., 1. ) );
    c -> referenceTransform -> setMatrix( tmpm );
}

int
IvCameraRep::initClass()
{
    static int inited = 0;
    if( inited )
       return 0;
    axisGeometry = new SoSeparator;
    axisCoordinates = new SoCoordinate3;
    axisLines = new SoIndexedLineSet;
    axisGeometry -> ref();
    axisGeometry -> addChild( axisCoordinates );
    axisGeometry -> addChild( axisLines );
    axisGeometry -> setName( "IvCameraRep_axisGeometry" );
    axisCoordinates -> setName( "IvCameraRep_axisCoordinates" );
    axisLines -> setName( "IvCameraRep_axisLines" );
    axisCoordinates -> point.setValues( 0, 20, axisCoordinateValues );
    axisLines -> coordIndex.setValues( 0, 33, axisLineCoords );
    copGeometry = new SoSphere;
    copGeometry -> ref();
    copGeometry -> setName( copName );
    copGeometry->radius.setValue( .03 );
    inited = 1;
    return 0;
}


IvCameraRep::IvCameraRep()
  : root( new SoSeparator ),
    imagePlaneDraggerRoot( new SoSeparator ),
    imagePlaneDraggerTransform( new SoTransform ),
    // imagePlaneDragger( new SoJackDragger ),
    imagePlaneDragger( new SoTransformBoxDragger ),
    material( new SoMaterial ),
    referenceTransform( new SoTransform ),
    copRoot( new SoTransformSeparator ),
    interiorOrientation( new SoTransform ),
    copDraggerRoot( new SoSeparator ),
    copDraggerTransform( new SoTransform ),
    copDragger( new SoTranslate1Dragger ),
    copDragUserCB( 0 ),
    copDragUserData( 0 ),
    imagePlaneDragUserCB( 0 ),
    imagePlaneDragUserData( 0 ),
    interiorOrientationSensor( 0 ),
    relativeOrientationSensor( 0 ),
    rayCount( 0 ),
    rayGeometry( new SoSeparator ),
    rayCoordinates( new SoCoordinate3 ),
    rayLines( new SoIndexedLineSet )
{
    initClass();

    if( !root || !imagePlaneDraggerRoot || !imagePlaneDraggerTransform ||
	!imagePlaneDragger || !material || !referenceTransform ||
	!copRoot || !interiorOrientation || !copDraggerRoot || 
	!copDraggerTransform || ! copDragger )
       fprintf( stderr, "IvCameraRep: init failed\n" ), exit(1);

    // Main scene graph (no draggers)
    root -> ref();
    root -> addChild( material );
    root -> addChild( referenceTransform );
    root -> addChild( axisGeometry );
    root -> addChild( copRoot );
    copRoot -> addChild( interiorOrientation );
    copRoot -> addChild( copGeometry );
    root -> setName( rootName );
    material -> setName( "IvCameraRep_material" );
    referenceTransform -> setName( "IvCameraRep_referenceTransform" );
    copRoot -> setName( "IvCameraRep_copRoot" );
    interiorOrientation -> setName( "IvCameraRep_interiorOrientation" );
    setColor( .7, .5, 0. );
    setInteriorOrientation( 1. );
    setReferenceTransform( 0., 0., 0., 1., 0., 0., 0. );

    // COP Dragger
    copDraggerRoot -> ref();
    copDraggerRoot -> addChild( copDraggerTransform );
    copDraggerRoot -> addChild( copDragger );
    copDraggerTransform -> translation.setValue( .1, 0., -1. );
    copDraggerTransform -> rotation.setValue( 0., .70710678, 0., .70710678 );
    copDraggerTransform -> scaleFactor.setValue( .1, .1, .1 );
    copDraggerRoot -> setName( "IvCameraRep_copDraggerRoot" );
    copDraggerTransform -> setName( "IvCameraRep_copDraggerTransform" );
    copDragger -> setName( "IvCameraRep_copDragger" );

    // Image plane Dragger
    imagePlaneDraggerRoot -> ref();
    imagePlaneDraggerRoot -> addChild( imagePlaneDraggerTransform );
    imagePlaneDraggerRoot -> addChild( imagePlaneDragger );
    imagePlaneDraggerRoot -> setName( "IvCameraRep_imagePlaneDraggerRoot" );
    imagePlaneDraggerTransform -> 
       setName( "IvCameraRep_imagePlaneDraggerTransform" );
    imagePlaneDragger -> setName( "IvCameraRep_imagePlaneDragger" );

    // Rays
    root -> addChild( rayGeometry );
    rayGeometry -> addChild( rayCoordinates );
    rayGeometry -> addChild( rayLines );
    rayCoordinates -> point.setValue( 0, 0, -getFocalLength() );
    
}

IvCameraRep::~IvCameraRep()
{
    root -> unref();
    copDraggerRoot -> unref();
    imagePlaneDraggerRoot -> unref();
}

void
IvCameraRep::setReferenceTransform( float tx, float ty, float tz, 
				    float q0, float q1, float q2, float q3 )
{
    SbVec3f scale( 1., 1., 1. );
    SbRotation rot( q1, q2, q3, q0 );
    SbVec3f tr( tx, ty, tz );
    SbMatrix lrxform;
    lrxform.setTransform( tr, rot, scale );
    referenceTransform -> setMatrix( lrxform );
}
    
void
IvCameraRep::setCoordinateTransform( float tx, float ty, float tz, 
				     float q0, float q1, float q2, float q3 )
{
    SbVec3f scale( 1., 1., 1. );
    SbRotation rot( q1, q2, q3, q0 );
    SbVec3f tr( tx, ty, tz );
    SbMatrix lrxform;
    lrxform.setTransform( tr, rot, scale );
    lrxform = lrxform.inverse();
    referenceTransform -> setMatrix( lrxform );
}

int
IvCameraRep::select( SoPath* path )
{
    // cerr << "IvCameraRep::select: " << endl;
    int n = path -> getLength();
    for( int i = 0; i < n; i++ )
       if( path->getNode(i)->getName() == rootName &&
	   path->getNode(i) == root )
	  {
	      // cerr << "IvCameraRep::select: is a camera" << endl;
	      // cerr << "IvCameraRep::select: is this camera " << endl;
	      // printPath( path );
	      if( path->getTail()->getName() == copName )
		 {
		     // cerr << "IvCameraRep::select: is a cop " <<endl;
		     copDragger->translation.setValue
			( (getFocalLength()-1.)*10, 0., 0. );
		     copDragger -> 
			addValueChangedCallback( copDragValueCB, this );
		     if( copDragUserCB )
			copDragger -> 
			   addValueChangedCallback( copDragUserCB,
						    copDragUserData );
		     root -> addChild( copDraggerRoot );
		 }
	      else
		 {
		     // cerr << "IvCameraRep::select: is not a cop " << endl;
		     imagePlaneDragger->scaleFactor.setValue( .4, .2, .2 );
		     imagePlaneDragger->rotation = 
			referenceTransform->rotation.getValue();
		     imagePlaneDragger->translation.setValue
			( referenceTransform->translation.getValue() );
		     imagePlaneDragger ->
			addValueChangedCallback( imagePlaneDragValueCB, this );
		     if( imagePlaneDragUserCB )
			imagePlaneDragger ->
			   addValueChangedCallback( imagePlaneDragUserCB,
						    imagePlaneDragUserData);
		     root -> insertChild( imagePlaneDraggerRoot, 0 );
		 }
	      return 1;
	  }
    return 0;
}

int
IvCameraRep::deselect( SoPath* path )
{
    int n = path -> getLength();
    for( int i = 0; i < n; i++ )
       if( path->getNode( i )->getName() == rootName &&
	   path->getNode(i) == root )
	  {
	      if( path->getTail()->getName() == copName )
		 {
		     root -> removeChild( copDraggerRoot );
		     copDragger -> 
			removeValueChangedCallback( copDragValueCB, this );
		     if( copDragUserCB )
			copDragger -> 
			   removeValueChangedCallback( copDragUserCB,
						       copDragUserData );
		 }
	      else
		 {
		     root -> removeChild( imagePlaneDraggerRoot );
		     imagePlaneDragger -> 
			removeValueChangedCallback
			( imagePlaneDragValueCB, this );
		     if( imagePlaneDragUserCB )
			imagePlaneDragger ->
			   removeValueChangedCallback( imagePlaneDragUserCB,
						       imagePlaneDragUserData);
		 }
	      return 1;
	  }
    return 0;
}

void
IvCameraRep::attachInteriorOrientation( SoPerspectiveCamera* pc )
{
    if( interiorOrientationSensor )
       {
	   interiorOrientationSensor -> detach();
	   delete interiorOrientationSensor;
       }
    interiorOrientationSensor = new SoNodeSensor
       ( &interiorOrientationSensorCB, this );
    interiorOrientationSensor -> attach( pc );
}

void
IvCameraRep::attachReferenceTransform( SoTransform* rt )
{
    if( relativeOrientationSensor )
       {
	   relativeOrientationSensor -> detach();
	   delete relativeOrientationSensor;
       }
    relativeOrientationSensor = new SoNodeSensor
       ( &referenceTransformSensorCB, this );
    relativeOrientationSensor -> attach( rt );
}

void
IvCameraRep::attachCoordinateTransform( SoTransform* ct )
{
    if( relativeOrientationSensor )
       {
	   relativeOrientationSensor -> detach();
	   delete relativeOrientationSensor;
       }
    relativeOrientationSensor = new SoNodeSensor
       ( coordinateTransformSensorCB, this );
    relativeOrientationSensor -> attach( ct );
}

void
IvCameraRep::interiorOrientationSensorCB( void* icrptr, SoSensor* sensor )
{
    IvCameraRep* icr = (IvCameraRep*) icrptr;
    SoNodeSensor* nodesensor = (SoNodeSensor*) sensor;
    SoPerspectiveCamera* pc = 
       (SoPerspectiveCamera*) nodesensor -> getAttachedNode();
    icr -> setInteriorOrientation( pc );
}

void
IvCameraRep::referenceTransformSensorCB( void* icrptr, SoSensor* sensor )
{
    IvCameraRep* icr = (IvCameraRep*) icrptr;
    SoNodeSensor* nodesensor = (SoNodeSensor*) sensor;
    SoTransform* rt = 
       (SoTransform*) nodesensor -> getAttachedNode();
    SoGetMatrixAction *fuckingstupid = new SoGetMatrixAction( SbViewportRegion() );
    fuckingstupid->apply( rt );
    icr -> setReferenceTransform( fuckingstupid->getMatrix() );
    delete fuckingstupid;
}

void
IvCameraRep::coordinateTransformSensorCB( void* icrptr, SoSensor* sensor )
{
    IvCameraRep* icr = (IvCameraRep*) icrptr;
    SoNodeSensor* nodesensor = (SoNodeSensor*) sensor;
    SoTransform* ct = 
       (SoTransform*) nodesensor -> getAttachedNode();
    SoGetMatrixAction *fuckingstupid = new SoGetMatrixAction( SbViewportRegion() );
    fuckingstupid->apply( ct );
    icr -> setReferenceTransform( fuckingstupid->getInverse() );
    delete fuckingstupid;
}

void
IvCameraRep::driveInteriorOrientation( SoPerspectiveCamera* pc )
{
    copDragger -> addValueChangedCallback( interiorOrientationDraggerCB, pc );
}

void
IvCameraRep::driveReferenceTransform( SoTransform* rt )
{
    imagePlaneDragger -> 
       addValueChangedCallback( referenceTransformDraggerCB, rt );
}

void
IvCameraRep::driveCoordinateTransform( SoTransform* ct )
{
    imagePlaneDragger -> 
       addValueChangedCallback( coordinateTransformDraggerCB, ct );
}

void
IvCameraRep::interiorOrientationDraggerCB( void* pcptr,
					   SoDragger* dragger )
{
    SoPerspectiveCamera* pc = (SoPerspectiveCamera*) pcptr;
    SoTranslate1Dragger* std = (SoTranslate1Dragger*) dragger;
    float f = std->translation.getValue()[0] / 10. + 1.;
    if( f < .001 )
       f = .001;
    pc -> position.setValue( 0., 0., -f );
    pc -> orientation.setValue( 1., 0., 0., 0. );
    pc -> heightAngle = 2. * atan( .375 / f );
    pc -> aspectRatio = 4./3.;
    pc -> nearDistance = .1 * f;
    pc -> farDistance = 100.;
}

void
IvCameraRep::referenceTransformDraggerCB( void* rtptr, SoDragger* dragger )
{
    SoTransform* rt = (SoTransform*) rtptr;
    // SoJackDragger* sjd = (SoJackDragger*) dragger;
    SoTransformBoxDragger* sjd = (SoTransformBoxDragger*) dragger;
    SbMatrix tmpm;
    tmpm.setTransform( sjd->translation.getValue(),
		       sjd->rotation.getValue(),
		       SbVec3f( 1., 1., 1. ) );
    rt -> setMatrix( tmpm );
}

void
IvCameraRep::coordinateTransformDraggerCB( void* ctptr, SoDragger* dragger )
{
    SoTransform* ct = (SoTransform*) ctptr;
    // SoJackDragger* sjd = (SoJackDragger*) dragger;
    SoTransformBoxDragger* sjd = (SoTransformBoxDragger*) dragger;
    SbMatrix tmpm;
    tmpm.setTransform( sjd->translation.getValue(),
		       sjd->rotation.getValue(),
		       SbVec3f( 1., 1., 1. ) );
    ct -> setMatrix( tmpm.inverse() );
}

void
IvCameraRep::setSize( int h, int w )
{
    float halfHeight = float(h) / float(w) / 2.;
    // cerr << "IvCameraRep::setSize " << halfHeight << endl;
    axisCoordinateValues[0][1] = halfHeight;
    axisCoordinateValues[1][1] = -halfHeight;
    axisCoordinateValues[2][1] = -halfHeight;
    axisCoordinateValues[3][1] = halfHeight;
    axisCoordinates -> point.setValues( 0, 20, axisCoordinateValues );
}

