// IvCameraView.h++

// In this object, "reference transform" and "coordinate transform" are
// of the camera frame relative to the world.  Thusly, rt is where the camera
// is relative to the world, ct is where the world is relative to the camera

#ifndef _IvCameraView_H_
#define _IvCameraView_H_

#include <Inventor/Xt/SoXtRenderArea.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoCallback.h>
#include <Inventor/nodes/SoSpotLight.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoTransformSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/events/SoMouseButtonEvent.h>
#include <Inventor/SoOffscreenRenderer.h>
#include <Inventor/draggers/SoJackDragger.h>
#include <Inventor/draggers/SoTranslate1Dragger.h>
#include <Inventor/sensors/SoNodeSensor.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <X11/IntrinsicP.h>
#include "QuickMath/qmMinMax.h++"

class IvCameraView
{
private:
    static const int defaultHeight;
    static const int defaultWidth;
    int height;
    int width;
    // Scene graph for what is viewed BY the camera
    // SEP
    //  |
    // CAM..BGCB..LIGHT..L-MODEL..SCENE
    //                              |
    //                            REL-OR ...
    SoXtRenderArea* renderArea;
    SoSeparator* root;
    SoPerspectiveCamera* camera;
    SoSpotLight* light;
    SoLightModel* lightModel;
    SoTransformSeparator* sceneRoot;
    SoTransform* coordinateTransform;
    SoEventCallback* keyEventNode;
    int (*userKeyCB)( void*, SoEventCallback* );
    void* userKeyData;
    static SoEventCallbackCB keypressCB;
    int (*userMouseCB)( void*, SoEventCallback* );
    void* userMouseData;
    static SoEventCallbackCB mousepressCB;

    void* backgroundData;
    GLenum backgroundFormat;
    SoCallback* backgroundNode;
    static SoCallbackCB backgroundCB;

    // For driving my geometry from my manipulators
    // static SoDraggerCB copDragValueCB;
    // static SoDraggerCB imagePlaneDragValueCB;

    // For driving other nodes from my manipulators
    // static SoDraggerCB interiorOrientationDraggerCB;
    // static SoDraggerCB referenceTransformDraggerCB;
    // static SoDraggerCB coordinateTransformDraggerCB;

    // For driving my geometry from other nodes
    static SoSensorCB interiorOrientationSensorCB;
    static SoSensorCB referenceTransformSensorCB;
    static SoSensorCB coordinateTransformSensorCB;
    SoNodeSensor* interiorOrientationSensor;
    SoNodeSensor* relativeOrientationSensor;

public:
    IvCameraView::IvCameraView( Widget = NULL );
    IvCameraView::~IvCameraView();
    // operator SoXtRenderArea*() { return renderArea; }
    inline SoXtRenderArea* getRenderArea() { return renderArea; }
    inline Widget getWidget() { return renderArea->getWidget(); }
    // inline void setSize( int h, int w, int y, int x );
    void setSize( int h, int w, int y, int x );
    inline void addToView( SoNode* );
    // inline void setViewSize( int h, int w );
    // inline void setViewScreenLocation( int y, int x );
    // inline void setWindowConfig( int y, int x, int h, int w );
    inline void setTitle( char *s ) { renderArea -> setTitle( s ); }
    void setBackground( void*, GLenum = GL_RGB );
    void setBackgroundColor( float r, float g, float b );
    void setLightModel( int model ); // SoLightModel::{BASE_COLOR,PHONG}

    // void updateBackground();
    inline void setKeyCB( int (*)(void*, SoEventCallback*), void* = 0 );
    inline void setMouseCB( int (*)(void*, SoEventCallback*), void* = 0 );

    void setInteriorOrientation( float beta );

    inline void setReferenceTransform( SbMatrix& );
    inline void setCoordinateTransform( SbMatrix& );
    void setCoordinateTransform( float tx, float ty, float tz,
				 float q0, float q1, float q2, float q3 );

    void attachInteriorOrientation( SoPerspectiveCamera* );
    void attachReferenceTransform( SoTransform* );
    void attachCoordinateTransform( SoTransform* );
    // void driveInteriorOrientation( SoPerspectiveCamera* );
    // void driveReferenceTransform( SoTransform* );
    // void driveCoordinateTransform( SoTransform* );

    int findColorBlob( SbColor&, float* mean, float* cov );
    // inline SoPerspectiveCamera* getCamera() { return camera; }
    // inline SoTransform* getCoordinateTransform() { return coordinateTransform; }
    void handle();
    SoSeparator* getRoot() { return root; }
};

#if 0
inline
void
IvCameraView::setSize( int h, int w, int y, int x )
{
    XtConfigureWidget( getWidget(), x, y, w, h, 0 );
    height = h;
    width = w;
}
#endif

inline
void
IvCameraView::addToView( SoNode* node )
{
    sceneRoot -> addChild( node );
}

inline
void
IvCameraView::setReferenceTransform( SbMatrix& m )
{
    coordinateTransform -> setMatrix( m.inverse() );
}

inline
void
IvCameraView::setCoordinateTransform( SbMatrix& m )
{
    coordinateTransform -> setMatrix( m );
}

inline 
void
IvCameraView::setKeyCB( int (*fn)(void*, SoEventCallback* ),
		    void* data )
{
    userKeyCB = fn;
    userKeyData = data;
}

inline 
void
IvCameraView::setMouseCB( int (*fn)(void*, SoEventCallback* ),
			  void* data )
{
    userMouseCB = fn;
    userMouseData = data;
}

#if 0
inline
void
IvCameraView::setViewSize( int h, int w )
{
    XtResizeWidget( renderArea->getParentWidget(), w, h, 0 );
}

inline
void
IvCameraView::setViewScreenLocation( int y, int x )
{
    XtMoveWidget( renderArea->getParentWidget(), x, y );
}

#endif

#endif /* _IvCameraView_H_ */
