// IvStereoCamera.c++

#include <stream.h>
#include "IvStereoCamera.h++"

static
void
error( char* fname, char* msg )
{
    cerr << "IvStereoCamera::" << fname << ": " << msg << endl;
}

void
IvStereoCamera::initialize( float bl, float br,
			    float tx, float ty, float tz,
			    float q0, float q1, float q2, float q3,
			    Widget wgt, int w, int h )
{
    Widget parent = wgt;
    if( !wgt )
       parent = XtVaCreateManagedWidget( "IvStereoCamera",
					 vendorShellWidgetClass,
					 SoXt::getTopLevelWidget(),
					 NULL );
    if( !parent )
       error( "initialize", "failed to create parent shell widget" );

    form = XtVaCreateManagedWidget( "Form",
				    xmFormWidgetClass, 
				    parent,
				    NULL );
    if( !form )
       error( "initialize", "failed to create xmForm widget" );

    // int viewHeight = 340;

    Iv3DViewer::initialize( form );
    Iv3DViewer::resize( 0, 0, 2*w+5, viewHeight );
    leftCamera = new IvCamera( form );
    leftCamera -> setSize( h, w, viewHeight, 0 );
    rightCamera = new IvCamera( form );
    rightCamera -> setSize( h, w, viewHeight, w+5 );
    leftCamera->setInteriorOrientation( bl );
    leftCamera->setTitle( "Left Camera (Yellow)" );
    rightCamera->setCoordinateTransform( tx, ty, tz, q0, q1, q2, q3 );
    rightCamera->setInteriorOrientation( br );
    rightCamera->setColor( 1., 0., .2 );
    rightCamera->setTitle( "Right Camera (Red)" );
    selectRoot -> addChild( leftCamera->getRep() );
    selectRoot -> addChild( rightCamera->getRep() );
    leftCamera->addToView( sceneRoot );
    rightCamera->addToView( sceneRoot );
    selectRoot -> removeSelectionCallback
       ( Iv3DViewer::selectCB, this );
    selectRoot -> addSelectionCallback( selectCB, this );
    selectRoot -> removeDeselectionCallback
       ( Iv3DViewer::deselectCB, this );
    selectRoot -> addDeselectionCallback( deselectCB, this );
    view3Dcamera-> position.setValue( 0., -3., 0. );
    view3Dcamera -> pointAt( SbVec3f( 0., 0., 0. ) );

    if( !wgt )
       SoXt::show( parent );
}

IvStereoCamera::IvStereoCamera( Widget wgt, int w, int h, int vh )
  : Iv3DViewer( 0, 0 ),
    leftCamera( 0 ),
    rightCamera( 0 ),
    viewHeight( vh )
{
    initialize( 1., 1., -.5, 0., 0., 1., 0., 0., 0., wgt, w, h );
}

IvStereoCamera::IvStereoCamera( float bl, float br,
				float tx, float ty, float tz,
				float q0, float q1, float q2, float q3 )
  : Iv3DViewer( 0, 0 ),
    leftCamera( 0 ),
    rightCamera( 0 ),
    viewHeight( 340 )
{
    initialize( bl, br, tx, ty, tz, q0, q1, q2, q3, NULL, 240, 180 );
}

IvStereoCamera::~IvStereoCamera()
{
    root -> unref();
}

#if 0
int
IvStereoCamera::addObject( SoNode* node )
{
    // Note: this is not the most efficient way, because we add a SEP
    // and SoTransform even to objects which may already have a suitable
    // internal SoTransform.
    // We plan to deal with only a few objects, so not a big concern
    // This scheme facilitates the easy selection/manipulation technique below

    SoSeparator* sep = new SoSeparator;
    SoTransform* xform = new SoTransform;
    sceneRoot -> addChild( sep );
    sep -> addChild( xform );
    sep -> addChild( node );
    sep -> setName( "IvStereoCamera_object" );
    xform -> setName( "IvStereoCamera_objectXform" );
    return( objectCount++ );
}
#endif

#if 0
SoTransform*
IvStereoCamera::getObjectTransform( int i )
{
    if( i < 0 || i >= objectCount )
       return 0;
    return( (SoTransform*)
	    ( ((SoSeparator*)(sceneRoot->getChild( i )))->getChild( 0 ) ) );
}
#endif

#if 0
static
SoPath*
createPathToXform( SoPath* objectPath )
{
    // The first three elements of every object path are:
    // selectRoot -> sceneRoot -> objSep
    SoPath* pathToXform = objectPath -> copy( 0, 3 );
    pathToXform -> ref();
    // The SoTransform is the first child of objSep
    pathToXform->append( ((SoGroup *)(pathToXform->getTail()))->getChild(0) );
    pathToXform -> unrefNoDelete();
    return( pathToXform );
}
#endif

void
IvStereoCamera::selectCB( void *svptr, SoPath *selectionPath )
{
    IvStereoCamera* sv = (IvStereoCamera*) svptr;
    if( sv -> rightCamera->select( selectionPath ) )
       return;
    if( sv -> leftCamera->select( selectionPath ) )
       return;
    Iv3DViewer::selectCB( svptr, selectionPath );
}

void
IvStereoCamera::deselectCB( void *svptr, SoPath *deselectionPath )
{
    IvStereoCamera* sv = (IvStereoCamera*) svptr;
    if( sv -> rightCamera->deselect( deselectionPath ) )
       return;
    if( sv -> leftCamera->deselect( deselectionPath ) )
       return;
    Iv3DViewer::deselectCB( svptr, deselectionPath );
}

#if 0
void
IvStereoCamera::keypressCB( void* svptr, SoEventCallback* node )
{
    IvStereoCamera* sv = (IvStereoCamera*) svptr;
    if( sv->userKeyCB )
       if( sv->userKeyCB( sv->userKeyData, node ) )
	  return;
    const SoEvent *event = node -> getEvent();
    if( SO_KEY_PRESS_EVENT( event, Q ) )
       exit(0);
    if( SO_KEY_PRESS_EVENT( event, LEFT_ALT ) ||
	SO_KEY_PRESS_EVENT( event, RIGHT_ALT ) )
       sv -> altkey = 1;
    if( SO_KEY_RELEASE_EVENT( event, LEFT_ALT ) ||
	SO_KEY_RELEASE_EVENT( event, RIGHT_ALT ) )
       sv -> altkey = 0;
}
#endif

void
IvStereoCamera::getRLRefTransform( float& tx, float& ty, float& tz,
				   float& q0, float& qx, 
				   float& qy, float& qz )
{
    SbVec3f tL;
    SbVec3f tR;
    SbRotation qL;
    SbRotation qR;
    leftCamera -> getCoordinateTransform( tL, qL );
    rightCamera -> getCoordinateTransform( tR, qR );
    qL.invert();  	// Rli
    qR *= qL;		// Rr Rli
    qR.getValue( qx, qy, qz, q0 );
    SbVec3f RrRliTl;
    qR.multVec( tL, RrRliTl );
    tR -= RrRliTl;
    tR.getValue( tx, ty, tz );
}

void
IvStereoCamera::handle()
{
    XtInputMask m;
    while( m = XtAppPending( SoXt::getAppContext() ) )
       XtAppProcessEvent( SoXt::getAppContext(), m );
}

