// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.  These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.

// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988, 1992 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of the GNU C++ Library.  This library is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef _llSingle_h
#ifdef __GNUG__
//#pragma interface
#endif
#define _llSingle_h 1

#include "LinkedList/llPix.h++"

struct BasellSingleNode
{
   BasellSingleNode *tl;
   void *item() {return (void*)(this+1);} // Return ((llSingleNode<T>*)this)->hd
};

template<class T>
class llSingleNode : public BasellSingleNode
{
  public:
    T                    hd; // Data part of node
                         llSingleNode() { }
                         llSingleNode(const T& h, llSingleNode* t = 0)
			     : hd(h) { tl = t; }
                         ~llSingleNode() { }
};

extern int __llSingleLength(BasellSingleNode *ptr);

class BasellSingle {
  protected:
    BasellSingleNode *last;
    virtual void delete_node(BasellSingleNode*node) = 0;
    virtual BasellSingleNode* copy_node(const void* datum) = 0;
    virtual void copy_item(void *dst, void *src) = 0;
    virtual ~BasellSingle() { }
    BasellSingle() { last = 0; }
    void copy(const BasellSingle&);
    BasellSingle& operator = (const BasellSingle& a);
    llPix ins_after(llPix p, const void *datum);
    llPix prepend(const void *datum);
    llPix append(const void *datum);
    int remove_front(void *dst, int signal_error = 0);
    void join(BasellSingle&);
  public:
    int length() const;
    int empty() const { return last == 0; }
    void clear();
    llPix                   prepend(BasellSingleNode*);
    llPix                   append(BasellSingleNode*);
    int                   OK() const;
    void                  error(const char* msg) const;
    void                  del_after(llPix p);
    int                   owns(llPix p) const;
    void                  del_front();
};

template <class T>
class llSingle : public BasellSingle
{
  private:
    virtual void delete_node(BasellSingleNode *node) { delete (llSingleNode<T>*)node; }
    virtual BasellSingleNode* copy_node(const void *datum)
	{ return new llSingleNode<T>(*(const T*)datum); }
    virtual void copy_item(void *dst, void *src) { *(T*)dst = *(T*)src; }

public:
    llSingle() : BasellSingle() { }
    llSingle(const llSingle<T>& a) : BasellSingle() { copy(a); }
    llSingle<T>&            operator = (const llSingle<T>& a)
	{ BasellSingle::operator=((const BasellSingle&) a); return *this; }
    virtual ~llSingle() { clear(); }

    llPix prepend(const T& item) {return BasellSingle::prepend(&item);}
    llPix append(const T& item) {return BasellSingle::append(&item);}
    llPix prepend(llSingleNode<T>* node) {return BasellSingle::prepend(node);}
    llPix append(llSingleNode<T>* node) {return BasellSingle::append(node);}

    T& operator () (llPix p) {
	if (p == 0) error("null llPix");
	return ((llSingleNode<T>*)(p))->hd; }
    const T& operator () (llPix p) const {
	if (p == 0) error("null llPix");
	return ((llSingleNode<T>*)(p))->hd; }
    inline llPix first() const { return (last == 0) ? 0 : llPix(last->tl); }
    void next(llPix& p) const
	{ p = (p == 0 || p == last) ? 0 : llPix(((llSingleNode<T>*)(p))->tl); }
    llPix ins_after(llPix p, const T& item)
      { return BasellSingle::ins_after(p, &item); }
    void join(llSingle<T>& a) { BasellSingle::join(a); }
    
    T& front() {
	if (last == 0) error("front: empty list");
	return ((llSingleNode<T>*)last->tl)->hd; }
    T& rear() {
	if (last == 0) error("rear: empty list");
	return ((llSingleNode<T>*)last)->hd; }
    const T& front() const {
	if (last == 0) error("front: empty list");
	return ((llSingleNode<T>*)last->tl)->hd; }
    const T& rear() const {
	if (last == 0) error("rear: empty list");
	return ((llSingleNode<T>*)last)->hd; }
    int remove_front(T& x) { return BasellSingle::remove_front(&x); }
    T remove_front() { T dst; BasellSingle::remove_front(&dst, 1); return dst; }
};

#endif
