// Parse.c++
// To do
//
// (*) make a util.PrintArgs which will print the *value* of all the
//     args, much like PrintHelp prints them
//
// (*) allow "-???" to not pass as an ordered argument if "-" is a
//     registered flag
//
// (*) add flags in Parse that change the error reporting
//     behavior, e.g.
//       cannotFindFlagOrTag = { *Error*, Alert, Ignore, UseAsOrdered }
//       cannotFindArgBehavior = { *Error*, Alert, Ignore }
//       notEnoughOrderedArgs = { *Error*, Alert, Ignore }
//       tooManyOrderedArgs = { *Error*, Alert, Ignore } (same as cfab)


#include "Parse/Parse.h++"

/***************************************************************************\
 *
 * class AssignData
 *
\***************************************************************************/

char *
AssignData::Type()
{
  switch( convertType )
    {
     case Int:
      return "integer";
      // break;
     case Float:
      return "float";
      // break;
     default:
      return "string";
      // break;
    }
  // return 0;
}

ostream&
AssignData::PrintType( ostream& os )
{
  return ( os << Type() );
}

/***************************************************************************\
 *
 * class Rule
 *
\***************************************************************************/

char *
Rule::ComputeShortName( char *longName )
{
  char *buf = new char[9];
  strncpy( buf, longName, 8 );
  for( int i = 0; i < 8; i++ )
    {
      if( isspace( buf[i] ) )
	break;
      buf[i] = tolower( buf[i] );
    }
  buf[i] = '\0'; // no string will be longer than 8 chars
  return buf;
}

Rule::Rule( int acttype, AssignData *ad, char *desc, char *nm )
  : actionType( acttype )
{
  if( actionType != Assign && actionType != AssignArg )
    cerr << "Rule::Rule internal error" << endl;
  action.assign = ad;
  name = ( desc ) ? desc : "(Not Documented)";
  shortName = ( desc && nm ) ? nm : ( desc ) ? ComputeShortName( desc ) 
    : ad->Type();
}

Rule::Rule( void (*proc)(void*), void *data, char *desc, char *nm )
  : actionType( Proc )
{
  action.proc = proc;
  callData = data;
  name = ( desc ) ? desc : "(Not Documented)";
  shortName = ( desc && nm ) ? nm : ( desc ) ? ComputeShortName( desc ) : "";
}

Rule::Rule( void (*procarg)( char *,void* ), void *data, char *desc, char *nm )
  : actionType( ProcArg )
{
  action.procarg = procarg;
  callData = data;
  name = ( desc ) ? desc : "(Not Documented)";
  shortName = ( desc && nm ) ? nm : ( desc ) ? ComputeShortName( desc ) : "";
}

ostream&
Rule::PrintArgName( ostream& os )
{
  return ( os << "<" << shortName << ">" );
}

ostream&
Rule::PrintName( ostream& os )
{
  return ( os << name );
}

ostream&
Rule::PrintDefaultValue( ostream& os )
{
  if( actionType == Assign )
    return ( action.assign -> PrintValue
	     (action.assign -> PrintStorage( os << "[" ) << "|") << "]" );
  else if( actionType == AssignArg )
     return ( action.assign -> PrintStorage( os << "[" )  << "]" );
  else
     return os;
}

ostream&
Rule::PrintAction( ostream& os )
{
  if( actionType == Assign || actionType == Proc )
    {
      PrintName( os );
      if( actionType == Assign )
	PrintDefaultValue( os << " " );
    }
  else
    {
      PrintArgName( os ) << " = ";
      PrintName( os ) << " ";
      if( actionType == AssignArg )
	PrintDefaultValue( action.assign->PrintType( os << "(" ) << ") " );
      else
	os << "(string)";
    }
    return os;
}

void
Rule::Execute()
{
  if( actionType == Proc )
    action.proc( callData );
  else if ( actionType == Assign )
    action.assign -> Execute();
  else
    cerr << "Rule::Execute: internal error" << endl;
}

void
Rule::Execute( char *arg )
{
  if( actionType == ProcArg )
    action.procarg( arg, callData );
  else if ( actionType == AssignArg )
    action.assign -> Execute( arg );
  else
    cerr << "Rule::Execute: internal error" << endl;
}

/***************************************************************************\
 *
 * class RuleList
 *
\***************************************************************************/

void
RuleList::AddRule( Rule& rule )
{
  list.append( rule );
  if( rule.Type() == Rule::AssignArg ||
      rule.Type() == Rule::ProcArg )
    argCount++;
}

RuleList::RuleList( char, int n )
  : truth( 1 ), switchChar( '\0' ), idType( OrderedType ), argCount( 0 )
{
  id.number = n;
}

RuleList::RuleList( char sc, char fc )
  : truth( 1 ), switchChar( sc ), idType( FlagType ), argCount( 0 )
{
  id.flag = fc;
}

RuleList::RuleList( char sc, char *ts )
  : truth( 1 ), switchChar( sc ), idType( TagType ), argCount( 0 )
{
  id.tag = ts;
}

ostream&
RuleList::PrintArg( ostream& os )
{
  Pix index;

  switch( idType )
    {
     case OrderedType:
      list.front().PrintArgName( os );
      break;
     case TagType:
      os << "[" << switchChar << id.tag;
      // for( Pix index = list.first(); index != 0; list.next(index) )
      for( index = list.first(); index != 0; list.next(index) )
	{
	  Rule rule = list( index );
	  if( Rule::AssignArg == rule.Type() ||
	      Rule::ProcArg == rule.Type() )
	    rule.PrintArgName( os << " " );
	}
      os << "]";
      break;
     case FlagType:
      os << "[" << switchChar << id.flag;
      for( index = list.first(); index != 0; list.next(index) )
	{
	  Rule rule = list( index );
	  if( Rule::AssignArg == rule.Type() ||
	      Rule::ProcArg == rule.Type() )
	    {
	      rule.PrintArgName( os );
	      break;
	    }
	}
      os << "]";
      break;

    }
  return os;
}

ostream&
RuleList::PrintHelp( ostream& os )
{
  Pix index;

  switch( idType )
    {
     case OrderedType:
      list.front().PrintAction( os );
      break;

     case TagType:
      os << switchChar << id.tag;
      // for( Pix index = list.first(); index != 0; list.next(index) )
      for( index = list.first(); index != 0; list.next(index) )
	{
	  Rule rule = list( index );
	  if( rule.Type() == Rule::AssignArg || 
	      rule.Type() == Rule::ProcArg )
	    rule.PrintArgName( os << " " );
	}
      os << " : ";
      index = list.first();
      if( list( index ).Type() == Rule::Assign ||
	  list( index ).Type() == Rule::Proc )
	{
	  list( index ).PrintAction( os ) << endl;
	  list.next(index);
	}
      else
	os << endl;
      for( ; index != 0; list.next(index) )
	list(index).PrintAction( os << "\t\t" ) << endl;
      break;

     case FlagType:
      os << switchChar << id.flag;
      for( index = list.first(); index != 0; list.next(index) )
	{
	  Rule rule = list( index );
	  if( rule.Type() == Rule::AssignArg ||
	      rule.Type() == Rule::ProcArg )
	    {
	      rule.PrintArgName( os );
	      break;
	    }
	}
      os << " : ";
      index = list.first();
      if( list( index ).Type() == Rule::Assign ||
	  list( index ).Type() == Rule::Proc )
	{
	  list( index ).PrintAction( os ) << endl;
	  list.next(index);
	}
      else
	os << endl;
      for( ; index != 0; list.next(index) )
	list(index).PrintAction( os << "\t\t" ) << endl;
      break;
    }
  return os;
}

int
RuleList::ExecuteRules() // for Flags
{
  if( argCount != 0 )
    return 1;

  for( Pix index = list.first(); index != 0; list.next(index) )
    list( index ).Execute();
  return 0;
}

int
RuleList::ExecuteRules( char *arg ) // for FlaggedArgs, or OrderedArgs
{
  if( argCount < 1 )
    return 1;

  for( Pix index = list.first(); index != 0; list.next(index) )
    {
      Rule rule = list(index);
      if( rule.Type() == Rule::Assign || rule.Type() == Rule::Proc )
	rule.Execute();
      else
	rule.Execute( arg );
    }
  return 0;
}

int
RuleList::ExecuteRules( int *argn, int argc, char *argv[] ) // for Tags
{
  int argNumber = *argn;
  if( argNumber + argCount >= argc )
    return 1;

  for( Pix index = list.first(); index != 0; list.next(index) )
    {
      Rule rule = list( index );
      if( rule.Type() == Rule::Assign || rule.Type() == Rule::Proc )
	rule.Execute();
      else
	rule.Execute( argv[++argNumber] );
    }

  *argn = argNumber;
  return 0;
}

/***************************************************************************\
 *
 * class RuleMatrix
 *
\***************************************************************************/

int
RuleMatrix::IsSwitch( char sc )
{
  for( int i = 0; i < switchCount; i++ )
    if( sc == switches[i] )
      return 1;
  return 0;
}

void
RuleMatrix::AddSwitch( char sc )
{
  for( int i = 0; i < switchCount; i++ )
    if( sc == switches[i] )
      break;
  if( i >= switchCount )
    switches[switchCount++] = sc;
}

void
RuleMatrix::PrintRuleCounts()
{
  for( Pix index = list.first(); index != 0; list.next(index) )
    {
      RuleList ruleList = list( index );
      cerr << "RuleList " << index << " has " << ruleList.RuleCount()
	   << "rules" << endl;
    }
  
}

void
RuleMatrix::AddOrderedRule( Rule& rule )
{
  RuleList& ruleList = *(new RuleList( (char)0, orderedCount++ ));
  ruleList.AddRule( rule );
  list.append( ruleList );
}

void
RuleMatrix::AddFlagRule( Rule& rule, char sc, char fc )
{
  RuleList ruleList;

  for( Pix index = list.first(); index != 0; list.next(index) )
    {
      ruleList = list( index );
      if( ruleList.Type() == RuleList::FlagType &&
	  ruleList.SwitchChar() == sc &&
	  ruleList.Flag() == fc )
	break;
    }
  if( 0 == index )
    {
      ruleList = *(new RuleList( sc, fc ));
      ruleList.AddRule( rule );
      list.append( ruleList );
    }
  else
    list( index ).AddRule( rule );
  AddSwitch( sc );
}

void
RuleMatrix::AddTagRule( Rule& rule, char sc, char *ts )
{
  // First, see if the tag exists
  RuleList ruleList;

  for( Pix index = list.first(); index != 0; list.next(index) )
    {
      ruleList = list( index );
      if( ruleList.Type() == RuleList::TagType &&
	  ruleList.SwitchChar() == sc &&
	  !strcmp( ruleList.Tag(), ts ) )
	break;
    }
  if( 0 == index )
    {
      ruleList = *(new RuleList( sc, ts ));
      ruleList.AddRule( rule );
      list.append( ruleList );
    }
  else
    list( index ).AddRule( rule );
  AddSwitch( sc );
}

ParseLineStream&
RuleMatrix::PrintArgs( ParseLineStream& line )
{

  for( Pix index = list.first(); index != 0; list.next(index) )
    {
      RuleList ruleList = list( index );
      if( RuleList::OrderedType == ruleList.Type() )
	 {
	   ruleList.PrintArg( line );
	   line.Enter();
	 }
    }

  for( int switchNumber = 0; switchNumber < switchCount; switchNumber++ )
    {
      char switchChar = switches[switchNumber];
      line << "[" << switchChar << "{";
      for( index = list.first(); index != 0; list.next(index) )
	{
	  RuleList ruleList = list( index );
	  if( RuleList::FlagType == ruleList.Type() &&
	      switchChar == ruleList.SwitchChar() &&
	      0 == ruleList.ArgCount() )
	    line << ruleList.Flag();
	}
      line << "}] ";
      line.Enter();
    }

  for( index = list.first(); index != 0; list.next(index) )
    {
      RuleList ruleList = list( index );
      if( RuleList::TagType == ruleList.Type() ||
	  RuleList::FlagType == ruleList.Type() && 0 < ruleList.ArgCount() )
	 {
	   ruleList.PrintArg( line );
	   line.Enter();
	 }
    }

  return line;
}

ostream&
RuleMatrix::PrintArgs( ostream& os )
{

  for( Pix index = list.first(); index != 0; list.next(index) )
    {
      RuleList ruleList = list( index );
      if( RuleList::OrderedType == ruleList.Type() )
	ruleList.PrintArg( os ) << " ";
    }

  for( int switchNumber = 0; switchNumber < switchCount; switchNumber++ )
    {
      char switchChar = switches[switchNumber];
      os << "[" << switchChar << "{";
      for( index = list.first(); index != 0; list.next(index) )
	{
	  RuleList ruleList = list( index );
	  if( RuleList::FlagType == ruleList.Type() &&
	      switchChar == ruleList.SwitchChar() &&
	      0 == ruleList.ArgCount() )
	    os << ruleList.Flag();
	}
      os << "}] ";
    }

  for( index = list.first(); index != 0; list.next(index) )
    {
      RuleList ruleList = list( index );
      if( RuleList::TagType == ruleList.Type() ||
	  RuleList::FlagType == ruleList.Type() && 0 < ruleList.ArgCount() )
	ruleList.PrintArg( os ) << " ";
    }

  return os;
}

ostream&
RuleMatrix::PrintHelp( ostream& os )
{
  os << "Required Arguments:" << endl << endl;

  for( Pix index = list.first(); index != 0; list.next(index) )
    {
      RuleList ruleList = list( index );
      if( RuleList::OrderedType == ruleList.Type() )
	ruleList.PrintHelp( os << "\t" ) << endl;
    }

  os << endl << "Optional Arguments:" << endl << endl;

  for( int switchNumber = 0; switchNumber < switchCount; switchNumber++ )
    {
      char switchChar = switches[switchNumber];
      for( index = list.first(); index != 0; list.next(index) )
	{
	  RuleList ruleList = list( index );
	  if( RuleList::FlagType == ruleList.Type() &&
	      switchChar == ruleList.SwitchChar() &&
	      0 == ruleList.ArgCount() )
	    ruleList.PrintHelp( os << "\t" ) << endl;
	}
    }

  for( index = list.first(); index != 0; list.next(index) )
    {
      RuleList ruleList = list( index );
      if( RuleList::TagType == ruleList.Type() ||
	  RuleList::FlagType == ruleList.Type() && 0 < ruleList.ArgCount() )
	ruleList.PrintHelp( os << "\t" ) << endl;
    }
  return os;
}

#if 0
ostream&
RuleMatrix::PrintHelp( ostream& os )
{
  os << "Required Arguments:" << endl << endl;

  for( Pix index = list.first(); index != 0; list.next(index) )
    {
      RuleList ruleList = list( index );
      if( RuleList::OrderedType == ruleList.Type() )
	ruleList.PrintHelp( os << "\t" ) << endl;
    }

  os << endl << "Optional Arguments:" << endl << endl;

  for( int switchNumber = 0; switchNumber < switchCount; switchNumber++ )
    {
      char switchChar = switches[switchNumber];
      for( index = list.first(); index != 0; list.next(index) )
	{
	  RuleList ruleList = list( index );
	  if( RuleList::FlagType == ruleList.Type() &&
	      switchChar == ruleList.SwitchChar() &&
	      0 == ruleList.ArgCount() )
	    ruleList.PrintHelp( os << "\t" ) << endl;
	}
    }

  for( index = list.first(); index != 0; list.next(index) )
    {
      RuleList ruleList = list( index );
      if( RuleList::TagType == ruleList.Type() ||
	  RuleList::FlagType == ruleList.Type() && 0 < ruleList.ArgCount() )
	ruleList.PrintHelp( os << "\t" ) << endl;
    }
  return os;
}
#endif

RuleList
RuleMatrix::Find( int type, char switchChar, char *tag )
{
  // for tags only
  RuleList ruleList;
  for( Pix index = list.first(); index != 0; list.next(index) )
    {
      ruleList = list( index );
      if( ruleList.Type() == type &&
	  ruleList.SwitchChar() == switchChar &&
	  !strcmp( ruleList.Tag(), tag ) )
	return ruleList;
    }
  return 0;
}

RuleList
RuleMatrix::Find( int type, char switchChar, char flag )
{
  // for flags only
  RuleList ruleList;
  for( Pix index = list.first(); index != 0; list.next(index) )
    {
      ruleList = list( index );
      if( ruleList.Type() == type &&
	  ruleList.SwitchChar() == switchChar &&
	  ruleList.Flag() == flag )
	return ruleList;
    }
  return 0;
}

RuleList
RuleMatrix::Find( int type, int number )
{
  // for ordered only
  RuleList ruleList;
  for( Pix index = list.first(); index != 0; list.next(index) )
    {
      ruleList = list( index );
      if( ruleList.Type() == type &&
	  ruleList.Number() == number )
	return ruleList;
    }
  return 0;
}

/***************************************************************************\
 *
 * class Parse
 *
\***************************************************************************/


RuleMatrix Parse::ruleMatrix;
int Parse::verboseParse = 0;
int Parse::verboseUsage = 1;

int
Parse::ParseCmdLine( int argc, char *argv[] )
{
  // programName = argv[0];

  int orderedArgCount = 0;

  if( verboseParse )
    cerr << "Parse diagnostics begin:" << endl << endl;

  for( int argNumber = 1; argNumber < argc; argNumber++ )
    {
      // start with argNumber pointing to the FIRST ARG UNPROCESSED

      RuleList ruleList;
      char *arg = argv[argNumber];

      if( verboseParse )
	  cerr << "Arg: " << arg << endl;

      if( ruleMatrix.IsSwitch( arg[0] ) ) {
	
      if( verboseParse )
	  cerr << "\t" << arg[0] << " indicates a switch (TAG or FLAG)"<< endl;

      // -Image = ( -, Image, 0 )
      if( ruleList = ruleMatrix.Find( RuleList::TagType, arg[0], arg+1 ) )
	{
	  if( verboseParse )
	    cerr << "\tFound TAG " << arg << endl;

	  if( ruleList.ExecuteRules( &argNumber, argc, argv ) )
	    {
	      if( verboseParse )
		cerr << "\tNot enough args provided for TAG " << arg+1 
		     << endl
		     << "Parse failed" << endl << endl;
	      return 1;
	    }

	  if( verboseParse )
	    cerr << "\tExecuted rules for TAG " << arg << endl << endl;

	  continue;
	}

      if( verboseParse )
	cerr << "\tDid not find TAG " << arg << endl;

      // -Image = ( -, I, mage )
      if( (ruleList = ruleMatrix.Find( RuleList::FlagType, arg[0], arg[1] ))
	  && ruleList.ArgCount() > 0 )
	{ // We have found a flag arg[1] that takes an argument
	  if( verboseParse )
	    cerr << "\tFound FLAG " << arg[1] << " taking argument ("
		 << arg+2 << ")" << endl;
	  ruleList.ExecuteRules( arg+2 );
	  if( verboseParse )
	    cerr << "\tExecuted rules for FLAG " << arg[0] << arg[1]
		 << endl << endl;
	  continue;
	}

      if( verboseParse )
	cerr << "\tDid not find FLAG " << arg[0] << arg[1]
	     << " taking argument (" << arg+2 << ")" << endl;

      // -Image = ( -, I, 0 ), ( -, m, 0 ), ... , ( -, e, 0 )
      int flagCount = strlen( arg );
      for( int flagNumber = 1; flagNumber < flagCount; flagNumber++ )
	if( !(ruleList = 
	      ruleMatrix.Find( RuleList::FlagType, arg[0], arg[flagNumber] )) )
	  {
	    if( verboseParse )
	      cerr << "\tFlag " << arg[0] << arg[flagNumber]
		   << " not valid" << endl;
	    break;
	  }
	else if( ruleList.ArgCount() > 0 )
	  {
	    if( verboseParse )
	      cerr << "\tFlag " << arg[0] << arg[flagNumber]
		   << " takes an arg" << endl;
	    break;
	  }
      // if we did not break, ruleList is nonzero and ArgCount is zero
      if( ruleList && ruleList.ArgCount() == 0 )
	{ // ruleList found for each and every flag in arg
	  if( verboseParse )
	    cerr << "\tFound simple FLAGS {" << arg+1 << "} " 
		 << "with switch character " << arg[0] << endl;
	  for( flagNumber = 1; flagNumber < flagCount; flagNumber++ )
	    {
	      ruleList = ruleMatrix.Find( RuleList::FlagType,
					  arg[0], arg[flagNumber] );
	      ruleList.ExecuteRules();
	      if( verboseParse )
		cerr << "\tExecuted rules for FLAG " << arg[0] 
		     << arg[flagNumber]<< endl;
	    }
	  if( verboseParse )
	    cerr << endl;
	  continue;
	}
      if( verboseParse )
	cerr << "\t{" << arg+1 << "} is not a string of simple flags" << endl;
      }

      else { // No switch character

	if( verboseParse )
	  cerr << "\tArgument interpreted as an ordered argument" << endl;

      // -Image = ( 0, 0, -Image )
      if( ruleList = ruleMatrix.Find( RuleList::OrderedType,orderedArgCount ) )
	{
	  orderedArgCount++;
	  ruleList.ExecuteRules( arg );
	  if( verboseParse )
	    cerr << "\tExecuted rules for next ordered argument" 
		 << endl << endl;
	  continue;
	}
	if( verboseParse )
	  cerr << "\tCouldn't find a next ordered argument" << endl;
      
      }

      // If you end up here, generate Error, Alert, or ignore
      if( verboseParse )
	cerr << "Parse failed" << endl << endl;
      return 1;

      // end up with argNumber pointing to the LAST ARG ALREADY PROCESSED
    }

  if( orderedArgCount != ruleMatrix.OrderedArgCount() )
    {
      if( verboseParse )
	cerr << "There are " << ruleMatrix.OrderedArgCount() 
	<< " required ordered arguments"
	<< "\n\tbut only " << orderedArgCount << " were specified"
	<< endl 
	<< "Parse failed" << endl << endl;
      return 1;
    }
  if( verboseParse )
    cerr << "Parse successful" << endl;
  return 0;
}

void
Parse::Usage( char *programName, ostream& os )
{
#if 0
  ParseLineStream line( os, 78, "  " );
  line << "usage: " << programName;
  line.Enter();
  ruleMatrix.PrintArgs( line );
  line.Flush();
  os << endl;
#else
  os << "usage: " << programName << " ";
  ruleMatrix.PrintArgs( os );
  os << endl;
  os << endl;
#endif
  if( verboseUsage )
    ruleMatrix.PrintHelp( os );
}

/***************************************************************************\
 *
 * Parsing rule functions
 *
\***************************************************************************/

void
ParseRuleTag( char *tagString, 
	      void (*proc)(void*),
	      void *data,
	      char *desc,
	      char sc )
{
  Rule& rule = *(new Rule( proc, data, desc, 0 ));
  Parse::ruleMatrix.AddTagRule( rule, sc, tagString );
}

void
ParseRuleFlag( char flagChar,
	       void (*proc)(void*),
	       void *data, 
	       char *desc,
	       char sc )
{
  Rule& rule = *(new Rule( proc, data, desc, 0  ));
  Parse::ruleMatrix.AddFlagRule( rule, sc, flagChar );
}

void
ParseRuleArg( void (*proc)( char *, void* ),
	      void *data,
	      char *desc,
	      char *name )
{
  Rule& rule = *(new Rule( proc, data, desc, name ));
  Parse::ruleMatrix.AddOrderedRule( rule );
}

void
ParseRuleTaggedArg( char *tagString,
		    void (*proc)( char *, void* ),
		    void *data,
		    char *desc,
		    char *name,
		    char switchChar )
{
  Rule& rule = *(new Rule( proc, data, desc, name ));
  Parse::ruleMatrix.AddTagRule( rule, switchChar, tagString );
}

void
ParseRuleFlaggedArg( char flagChar,
		     void (*proc)( char *, void* ),
		     void *data, 
		     char *desc,
		     char *name,
		     char switchChar )
{
  Rule& rule = *(new Rule( proc, data, desc, name ));
  Parse::ruleMatrix.AddFlagRule( rule, switchChar, flagChar );
}

