// Parse.h

// Things to do
// (*) Arbitrary number of ordered arguments !!
// (*) Rules that are applied for every ordered arg or every tagged arg
// (*) Allow it to ignore args it does not know (instead of failing)
// (*) Fix Parse::Usage so that it formats long lines

#if 0

main()
{
    char s[32];
    void (*proc)( void * );
    void (*procArg)( char *, void * );
    Type var; // Type can be {int, float, char *, ...} or any class
    const Type val;
    
    // The simplest rules assign a value of some Type to a variable
    // or run a procedure when a flag or tag is present
    
    // Optional Args: ( ..., char *description, char switch )
    ParseRuleTag( "tag", &var, val );	// -tag		var = val
    ParseRuleTag( "tag", proc, 0 );		// -tag		proc()
    
    ParseRuleFlag( 'f', &var, val );		// -f		var = val
    ParseRuleFlag( 'f', proc, 0 );		// -f, -f<arg>	proc()
    
    // These rules assign the cmd line arg (a char *) to a variable
    // The "Int" and "Float" flavors, convert to int or float before
    // assigning.  More generally, a function can be called taking the
    // command line char* arg as an argument.
    
    // Optional Args: ( ..., char *description, char *label )
    ParseRuleArg( &var );			// <arg>	var = <arg>
    ParseRuleArgInt( &var );		// <arg>	var = atoi(<arg>)
    ParseRuleArgFloat( &var );		// <arg>	var = atof(<arg>)
    ParseRuleArg( procArg, 0 );		// <arg>	proc(<arg>)
    
    // Optional Args: ( ..., char *description, char *label, char switch )
    ParseRuleTaggedArg( "tag", &var );	  // -tag <arg>	var = <arg>
    ParseRuleTaggedArgInt( "tag", &var ); // -tag <arg>	var = atoi(<arg>)
    ParseRuleTaggedArgFloat("tag",&var ); // -tag <arg	var = atof(<arg>)
    ParseRuleTaggedArg("tag",procArg,0);  // -tag <arg>	proc(<arg>)
    
    ParseRuleFlaggedArg( 'f', &var );     // -f<arg>	var = <arg>
    ParseRuleFlaggedArgInt( 'f', &var );  // -f<arg>	var = atoi(<arg>)
    ParseRuleFlaggedArgFloat( 'f', &var );// -f<arg>	var = atof(<arg>)
    ParseRuleFlaggedArg( 'f',procArg,0);  // -f<arg>	proc(<arg>)

#endif
    
    
#include <stream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "Parse/ParseGnuSLList.h++"
#include "ParseLineStream.h++"
    
#ifndef _Parse_H_
#define _Parse_H_
    
/***************************************************************************\
 *
 * class AssignData
 *
 \***************************************************************************/
    
    class AssignData
       {
    private:
    protected:
	
	int convertType;
	
    public:
	
	enum { None, Int, Float, String };
	
	AssignData( int ct )
	  : convertType( ct ) {}
	
	ostream& PrintType( ostream& );
	char * Type();
	virtual void Execute() = 0;
	virtual void Execute( char * ) = 0;
	virtual ostream& PrintStorage( ostream& ) = 0;
	virtual ostream& PrintValue( ostream& ) = 0;
    };
    
/***************************************************************************\
 *
 * template < class T_ > class AssignDataT
 *
 \***************************************************************************/
    
    template < class T_ >
       class AssignDataT : public AssignData
       {
    private:
    protected:
	
	T_ *storage;
	T_ value;
	
    public:
	
	AssignDataT( int ct, T_ *s )
	  : AssignData( ct ), storage( s ) {}
	AssignDataT( T_ *s, T_ val )
	  : AssignData( AssignData::None ), storage( s ), value( val ) {}
	
	virtual void Execute();
	virtual void Execute( char * );
	virtual ostream& PrintStorage( ostream& );
	virtual ostream& PrintValue( ostream& );
	
    };
    
    template < class T_ >
       void
       AssignDataT<T_>::Execute()
       {
	   *storage = value;
       }
    
    template < class T_ >
       void
       AssignDataT<T_>::Execute( char *arg )
       {
	   *storage = *((T_ *)&arg);
       }
    
    template < class T_ >
       ostream&
       AssignDataT<T_>::PrintStorage( ostream& os )
       {
	   return( os << *storage );
       }
    
    template < class T_ >
       ostream&
       AssignDataT<T_>::PrintValue( ostream& os )
       {
	   return( os << value );
       }
    
/***************************************************************************\
 *
 * template < class T_ > class AssignDataIntT
 *
 \***************************************************************************/
    
    template < class T_ >
       class AssignDataIntT : public AssignDataT<T_>
       {
    private:
    public:
	
	AssignDataIntT( T_ *s )
	  : AssignDataT<T_>( AssignData::Int, s ) {}
	AssignDataIntT( T_ *s, T_ val )
	  : AssignDataT<T_>( s, val ) {}
	
	virtual void Execute( char * );
    };
    
    template < class T_ >
       void
       AssignDataIntT<T_>::Execute( char *arg )
       {
	   *storage = (T_) atoi(arg);
       }
    
    
/***************************************************************************\
 *
 * template < class T_ > class AssignDataFloatT
 *
 \***************************************************************************/
    
    template < class T_ >
       class AssignDataFloatT : public AssignDataT<T_>
       {
    private:
    public:
	
	AssignDataFloatT( T_ *s )
	  : AssignDataT<T_>( AssignData::Float, s ) {}
	AssignDataFloatT( T_ *s, T_ val )
	  : AssignDataT<T_>( s, val ) {}
	
	virtual void Execute( char * );
    };
    
    template < class T_ >
       void
       AssignDataFloatT<T_>::Execute( char *arg )
       {
	   *storage = (T_) atof(arg);
       }
    
/***************************************************************************\
 *
 * class Rule
 *
 \***************************************************************************/
    
    class Rule
       {
    public:
	
	enum { None, Assign, AssignArg, Proc, ProcArg };
	
    private:
	
	int actionType;
	union 
	{ 
	    AssignData *assign; 
	    void (*proc)(void *);
	    void (*procarg)(char*, void *); } action;
	    void *callData;
	    char *name;
	    char *shortName;
	    
	    char *ComputeShortName( char * );
	
    public:
	
	Rule( )
	  : actionType( None ) {}
	Rule( int acttype, AssignData *, char *desc, char *nm );
	Rule( void (*)(void *), void *data, char *desc, char *nm );
	Rule( void (*)( char *, void* ), void *data, char *desc, char *nm );
	
	ostream& PrintArgName( ostream& );
	ostream& PrintName( ostream& );
	ostream& PrintDefaultValue( ostream& );
	ostream& PrintAction( ostream& );
	char *ShortName() { return shortName; }
	char *Name() { return name; }
	int Type() { return actionType; }
	void Execute();
	void Execute( char * );
    };
    
/***************************************************************************\
 *
 * class RuleList
 *
 \***************************************************************************/
    
    class RuleList // list of Rules
       {
    public:
	
	enum { OrderedType, FlagType, TagType };
	
    private:
	
	int truth;
	char switchChar;
	int idType;
	union { int number; char flag; char *tag; } id;
	ParseGnuSLList<Rule> list;
	int argCount;
	
    public:
	
	RuleList( )
	  : truth( 1 ) {}
	RuleList( const int t )
	  : truth( t ) {}
	RuleList( char sc, int n );
	RuleList( char sc, char fc );
	RuleList( char sc, char *ts );
	
	ostream& PrintArg( ostream& );
	ostream& PrintHelp( ostream& );
	
	int Type() { return idType; }
	char SwitchChar() { return switchChar; }
	char Flag() { return id.flag; }
	char * Tag() { return id.tag; }
	int Number() { return id.number; }
	int ArgCount() { return argCount; }
	int ExecuteRules(); // for Flags
	int ExecuteRules( char * ); // for FlaggedArgs, or OrderedArgs
	int ExecuteRules( int *, int, char ** ); // for Tags
	operator int() { return truth; }
	void AddRule( Rule& );
	int RuleCount() { return list.length(); }
	
    };
    
/***************************************************************************\
 *
 * class RuleMatrix
 *
 \***************************************************************************/
    
    class RuleMatrix // list of RuleLists
       {
    private:
	
	char switches[128];
	int switchCount;
	int orderedCount;
	ParseGnuSLList<RuleList> list;
	
    public:
	
	RuleMatrix( )
	  : switchCount( 0 ), orderedCount( 0 ) {}
	
	ostream& PrintArgs( ostream& );
	ParseLineStream& PrintArgs( ParseLineStream& );
	ostream& PrintHelp( ostream& );
	RuleList Find( int, char, char * );
	RuleList Find( int, char, char );
	RuleList Find( int, int );
	void AddOrderedRule( Rule& );
	void AddFlagRule( Rule&, char, char );
	void AddTagRule( Rule&, char, char * );
	void AddSwitch( char );
	int IsSwitch( char );
	void PrintRuleCounts();
	int OrderedArgCount() { return orderedCount; }
	
    };
    
/***************************************************************************\
 *
 * class Parse
 *
 \***************************************************************************/
    
    class Parse
       {
	static int verboseParse;  // debugging info
	static int verboseUsage;  // prints long help
	
    public:
	
	static RuleMatrix ruleMatrix;
	
	static void Usage( char *programName, ostream& os = cerr );
	static void VerboseParse( int i = 1 ) { verboseParse = i; }
	static void VerboseUsage( int i = 1 ) { verboseUsage = i; }
	static int ParseCmdLine( int argc, char *argv[] );
	
    };
    
    inline
       int
       ParseCmdLine( int argc, char *argv[] )
       {
	   return Parse::ParseCmdLine( argc, argv );
       }
    
    inline
       void
       ParseUsage( char *programName, ostream& os = cerr )
       {
	   Parse::Usage( programName, os );
       }
    
    inline
       void
       ParseVerbose( int i = 1 )
       {
	   Parse::VerboseParse( i );
       }
    
    inline
       void
       ParseVerboseUsage( int i = 1 )
       {
	   Parse::VerboseUsage( i );
       }
    
/***************************************************************************\
 *
 * Parsing rule functions
 *
 \***************************************************************************/
    
    extern void ParseRuleTag( char *,
			      void (*)( void * ), 
			      void*,
			      char * = 0, 
			      char = '-');
    extern void ParseRuleFlag( char, 
			       void (*)( void * ),
			       void *,
			       char * = 0,
			       char = '-');
    extern void ParseRuleArg( void (*)( char *, void * ),
			      void *,
			      char * = 0,
			      char * = 0 );
    extern void ParseRuleTaggedArg( char*, 
				    void (*)( char *, void * ),
				    void *,
				    char* = 0,
				    char * = 0,
				    char = '-');
    extern void ParseRuleFlaggedArg( char,
				     void (*)( char *, void * ), 
				     void *,
				     char* = 0,
				     char * = 0,
				     char = '-');
    
// Tag and flag assignment rules
    
    template < class T_ >
       void
       ParseRuleTag( char *tagString,
		     T_& storage, 
		     T_ value, 
		     char *desc=0, 
		     char switchChar = '-' )
       {
	   AssignData *ad = new AssignDataT<T_>( &storage, value );
	   Rule& rule = *(new Rule( Rule::Assign, ad, desc, 0 ));
	   Parse::ruleMatrix.AddTagRule( rule, switchChar, tagString );
       }
    
    template < class T_ >
       void
       ParseRuleFlag( char flagChar,
		      T_& storage,
		      T_ value,
		      char *desc=0,
		      char switchChar = '-' )
       {
	   // When flag occurs, assign this pre-specified value to storage
	   AssignData *ad = new AssignDataT<T_>( &storage, value );
	   Rule& rule = *(new Rule( Rule::Assign, ad, desc, 0 ));
	   Parse::ruleMatrix.AddFlagRule( rule, switchChar, flagChar );
       }
    
// Ordered arguments
    
    template < class T_ >
       void
       ParseRuleArg( T_& storage, char *desc = 0, char *name = 0 )
       {
	   AssignData *ad = new AssignDataT<T_>( AssignData::String,&storage );
	   Rule& rule = *(new Rule( Rule::AssignArg, ad, desc, name ));
	   Parse::ruleMatrix.AddOrderedRule( rule );
       }
    
    template < class T_ >
       void
       ParseRuleArgInt( T_& storage, char *desc = 0, char *name = 0 )
       {
	   AssignData *ad = new AssignDataIntT<T_>( &storage );
	   Rule& rule = *(new Rule( Rule::AssignArg, ad, desc, name ));
	   Parse::ruleMatrix.AddOrderedRule( rule );
       }
    
    template < class T_ >
       void
       ParseRuleArgFloat( T_& storage, char *desc = 0, char *name = 0 )
       {
	   // Assign the value of this arg to storage
	   AssignData *ad = new AssignDataFloatT<T_>( &storage );
	   Rule& rule = *(new Rule( Rule::AssignArg, ad, desc, name ));
	   Parse::ruleMatrix.AddOrderedRule( rule );
       }
    
// Tagged arguments
    
    template < class T_ >
       void
       ParseRuleTaggedArg( char *tagString, 
			   T_& storage, 
			   char *desc=0, 
			   char *name = 0,
			   char sc = '-' )
       {
	   AssignData *ad = new AssignDataT<T_>( AssignData::String,&storage );
	   Rule& rule = *(new Rule( Rule::AssignArg, ad, desc, name ));
	   Parse::ruleMatrix.AddTagRule( rule, sc, tagString );
       }
    
    template < class T_ >
       void
       ParseRuleTaggedArgInt( char *tagString, 
			      T_& storage, 
			      char *desc=0, 
			      char *name = 0,
			      char sc = '-' )
       {
	   AssignData *ad = new AssignDataIntT<T_>( &storage );
	   Rule& rule = *(new Rule( Rule::AssignArg, ad, desc, name ));
	   Parse::ruleMatrix.AddTagRule( rule, sc, tagString );
       }
    
    template < class T_ >
       void
       ParseRuleTaggedArgFloat( char *tagString,
				T_& storage,
				char *desc=0,
				char *name = 0,
				char sc = '-' )
       {
	   AssignData *ad = new AssignDataFloatT<T_>( &storage );
	   Rule& rule = *(new Rule( Rule::AssignArg, ad, desc, name ));
	   Parse::ruleMatrix.AddTagRule( rule, sc, tagString );
       }
    
// Flagged arguments
    
    template < class T_ >
       void
       ParseRuleFlaggedArg( char flagChar,
			    T_& storage,
			    char *desc=0,
			    char *name = 0,
			    char sc = '-' )
       {
	   AssignData *ad = new AssignDataT<T_>( AssignData::String,&storage );
	   Rule& rule = *(new Rule( Rule::AssignArg, ad, desc, name ));
	   Parse::ruleMatrix.AddFlagRule( rule, sc, flagChar );
       }
    
    template < class T_ >
       void
       ParseRuleFlaggedArgInt( char flagChar,
			       T_& storage,
			       char *desc=0,
			       char *name = 0,
			       char sc = '-' )
       {
	   AssignData *ad = new AssignDataIntT<T_>( &storage );
	   Rule& rule = *(new Rule( Rule::AssignArg, ad, desc, name ));
	   Parse::ruleMatrix.AddFlagRule( rule, sc, flagChar );
       }
    
    template < class T_ >
       void
       ParseRuleFlaggedArgFloat( char flagChar,
				 T_& storage,
				 char *desc=0,
				 char *name = 0,
				 char sc = '-' )
       {
	   AssignData *ad = new AssignDataFloatT<T_>( &storage );
	   Rule& rule = *(new Rule( Rule::AssignArg, ad, desc, name ));
	   Parse::ruleMatrix.AddFlagRule( rule, sc, flagChar );
       }
    
#endif /* _Parse_H_ */
