// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988, 1992 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of the GNU C++ Library.  This library is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef _ParseGnuSLList_h
#ifdef __GNUG__
//#pragma interface
#endif
#define _ParseGnuSLList_h 1

#include "Parse/ParseGnuPix.h++"

struct AliBaseSLNode
{
   AliBaseSLNode *tl;
   void *item() {return (void*)(this+1);} // Return ((AliSLNode<T>*)this)->hd
};

template<class T>
class AliSLNode : public AliBaseSLNode
{
  public:
    T                    hd; // Data part of node
                         AliSLNode() { }
                         AliSLNode(const T& h, AliSLNode* t = 0)
			     : hd(h) { tl = t; }
                         ~AliSLNode() { }
};

extern int __ParseGnuSLListLength(AliBaseSLNode *ptr);

class BaseParseGnuSLList {
  protected:
    AliBaseSLNode *last;
    virtual void delete_node(AliBaseSLNode*node) = 0;
    virtual AliBaseSLNode* copy_node(const void* datum) = 0;
    virtual void copy_item(void *dst, void *src) = 0;
    virtual ~BaseParseGnuSLList() { }
    BaseParseGnuSLList() { last = 0; }
    void copy(const BaseParseGnuSLList&);
    BaseParseGnuSLList& operator = (const BaseParseGnuSLList& a);
    Pix ins_after(Pix p, const void *datum);
    Pix prepend(const void *datum);
    Pix append(const void *datum);
    int remove_front(void *dst, int signal_error = 0);
    void join(BaseParseGnuSLList&);
  public:
    int length() const;
    int empty() const { return last == 0; }
    void clear();
    Pix                   prepend(AliBaseSLNode*);
    Pix                   append(AliBaseSLNode*);
    int                   OK() const;
    void                  error(const char* msg) const;
    void                  del_after(Pix p);
    int                   owns(Pix p) const;
    void                  del_front();
};

template <class T>
class ParseGnuSLList : public BaseParseGnuSLList
{
  private:
    virtual void delete_node(AliBaseSLNode *node) { delete (AliSLNode<T>*)node; }
    virtual AliBaseSLNode* copy_node(const void *datum)
	{ return new AliSLNode<T>(*(const T*)datum); }
    virtual void copy_item(void *dst, void *src) { *(T*)dst = *(T*)src; }

public:
    ParseGnuSLList() : BaseParseGnuSLList() { }
    ParseGnuSLList(const ParseGnuSLList<T>& a) : BaseParseGnuSLList() { copy(a); }
    ParseGnuSLList<T>&            operator = (const ParseGnuSLList<T>& a)
	{ BaseParseGnuSLList::operator=((const BaseParseGnuSLList&) a); return *this; }
    virtual ~ParseGnuSLList() { clear(); }

    Pix prepend(const T& item) {return BaseParseGnuSLList::prepend(&item);}
    Pix append(const T& item) {return BaseParseGnuSLList::append(&item);}
    Pix prepend(AliSLNode<T>* node) {return BaseParseGnuSLList::prepend(node);}
    Pix append(AliSLNode<T>* node) {return BaseParseGnuSLList::append(node);}

    T& operator () (Pix p) {
	if (p == 0) error("null Pix");
	return ((AliSLNode<T>*)(p))->hd; }
    const T& operator () (Pix p) const {
	if (p == 0) error("null Pix");
	return ((AliSLNode<T>*)(p))->hd; }
    inline Pix first() const { return (last == 0) ? 0 : Pix(last->tl); }
    void next(Pix& p) const
	{ p = (p == 0 || p == last) ? 0 : Pix(((AliSLNode<T>*)(p))->tl); }
    Pix ins_after(Pix p, const T& item)
      { return BaseParseGnuSLList::ins_after(p, &item); }
    void join(ParseGnuSLList<T>& a) { BaseParseGnuSLList::join(a); }
    
    T& front() {
	if (last == 0) error("front: empty list");
	return ((AliSLNode<T>*)last->tl)->hd; }
    T& rear() {
	if (last == 0) error("rear: empty list");
	return ((AliSLNode<T>*)last)->hd; }
    const T& front() const {
	if (last == 0) error("front: empty list");
	return ((AliSLNode<T>*)last->tl)->hd; }
    const T& rear() const {
	if (last == 0) error("rear: empty list");
	return ((AliSLNode<T>*)last)->hd; }
    int remove_front(T& x) { return BaseParseGnuSLList::remove_front(&x); }
    T remove_front() { T dst; BaseParseGnuSLList::remove_front(&dst, 1); return dst; }
};

#endif
