/***************************************************************************
 * 
 * Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

// VLImage.c++

#include <stdlib.h>
#include <string.h>
#include "VLImage.h++"

#ifdef __sgi
VLServer VLImage::server = 0;
VLNode VLImage::drain = 0;
VLNode VLImage::source = 0;
VLPath VLImage::path = 0;
VLBuffer VLImage::buffer = 0;
VLInfoPtr VLImage::info = 0;
VLControlValue VLImage::val;
#endif
#ifdef __alpha__
LPHVIDEO VLImage::server;	
int VLImage::drain = 0;		
int VLImage::source = 0;	
int VLImage::path = 0;		
int VLImage::buffer = 0;	
int VLImage::info = 0;
int VLImage::val = 0;
#endif //__alpha__

int VLImage::capture_x = -1;
int VLImage::capture_y = -1;
int VLImage::capture_cropw = -1;
int VLImage::capture_croph = -1;
int VLImage::capture_width = 0;
int VLImage::capture_height = 0;
int VLImage::capture_size = 0;
int VLImage::capture_type = VLImage::RGB;
int VLImage::nbuffers = 40;
int VLImage::input = VLImage::INDY_SVIDEO;
int VLImage::verbose = 0;
int VLImage::zoom = 1;
int VLImage::transfer_begun = 0;

void
VLImage::Error( char *message )
{
#ifdef __sgi
  char tag[128];
  sprintf( tag, "VLImage: %s", message );
  vlPerror( tag );
  exit( 1 );
#endif
#ifdef __alpha__
  cerr << "VLImage: " <<  message << endl;
  exit( 1 );
#endif
}

void
VLImage::Init()
{
#ifdef __sgi
  // this code works for Indy and Indigo (should work for all VL)

  if( transfer_begun )
     return;

  // (0) Connect to the server

  if( !( server = vlOpenVideo("") ) )
     Error( "open video" );

  // (1) Find the "Analog Video Input" node number
  //     It is different for Indy and Indigo, and probably
  //     different still for the Sirius

  VLDevList listOfDevices;
  if( vlGetDeviceList( server, &listOfDevices ) < 0 )
     Error( "getting device list" );

  char *searchtag = NULL;
  switch (input)
    {
    case -3:
      searchtag = "Composite Video Input";
      break;
    case -2:
      searchtag = "SVideo Input";
      break;
    case -1:
      searchtag = "Digital Video Input";
      break;
    default:
      searchtag = "Analog Video Input";
      break;
    }
  for( int nodeIndex = 0;
       nodeIndex < listOfDevices.devices[0].numNodes &&
       strcmp( listOfDevices.devices[0].nodes[nodeIndex].name, searchtag);
       nodeIndex++ )
    {
      cerr << nodeIndex << ": " << searchtag << " != ";
      cerr << listOfDevices.devices[0].nodes[nodeIndex].name << endl;
    }
  if( nodeIndex >= listOfDevices.devices[0].numNodes )
     Error( "couldn't find Video Input node" );
  int analVidInputNodeNum = listOfDevices.devices[0].nodes[nodeIndex].number;

  // (2) Get the AVI node we just located, and any memory drain
  //     The path has to be setup before trying to set controls

  if( -1 == (drain = vlGetNode( server, VL_DRN, VL_MEM, VL_ANY )) )
    Error( "get drain node" );
  if( -1 == (source = vlGetNode( server, VL_SRC, VL_VIDEO, analVidInputNodeNum )))
    Error( "get source node" );
  if( -1 == (path = vlCreatePath( server, VL_ANY, source, drain )) )
    Error( "create path" );
  if( vlSetupPaths( server, (VLPathList)&path, 1, VL_SHARE, VL_SHARE ) < 0 )
     Error( "setup paths" );

  // (3) Set controls
  //     For some reason (??) you have to set Input (VL_MUXSWITCH)
  //     before the other controls or the Indy will barf (bug, IMHO)

  if (input >= 0) Input( input );
  Encoding( capture_type );
  Zoom( zoom );
  if (capture_x != -1)
    Crop( capture_x, capture_y, capture_cropw, capture_croph );

  // (4) Make and register buffers, start the transfer

  buffer = vlCreateBuffer( server, path, drain, nbuffers );
  if( !buffer )
     Error( "create buffers" );
  vlRegisterBuffer( server, path, drain, buffer );
  if( vlBeginTransfer( server, path, 0, NULL ) )
     Error( "begin transfer" );

#endif // __sgi

#ifdef __alpha__

    DWORD 	status;

  // this code works for Alpha j300

  if( transfer_begun )
     return;

  // (0) Connect to the server

  if (!(server = (LPHVIDEO)mmcAllocMem(sizeof(HVIDEO))))
    {
      Error( "malloc video data structure" );
    }
  
  status  = videoOpen(server, 0, VIDEO_IN) ; 
    if (status != DV_ERR_OK) 
    {
      if ( status == DV_ERR_ALLOCATED )
	Error("device is already in use.");
      else
	Error("videoOpen error.");
    }

    if (input != QUERY) 
    {
        status = videoSetStandard(*server, input);
        if (status != DV_ERR_OK) 
	{
	  if (status != DV_ERR_NOTSUPPORTED) 
	    {
	      Error("error selecting video input");
	    }
	  else 
	    {
	      Error("unsupported video input");
	    }
	}
    }

  if( !(InitializeStream(server) ) )
    Error( "open video" );

#endif  // __alpha__
}

void
VLImage::Close()
{
  if( !transfer_begun )
     return;

#ifdef __sgi
  vlEndTransfer( server, path );
  vlDeregisterBuffer( server, path, drain, buffer );
  vlDestroyBuffer( server, buffer );
  vlDestroyPath( server, path );
  vlCloseVideo( server );
#endif //__sgi
#ifdef __alpha__
  ShutdownVideo(server);
  videoClose( *server);
  mmcFreeAll(); /* Free All Allocated Shared Memory */
#endif // __alpha__
  transfer_begun = 0;
}

int
VLImage::NBuffers( int n )
{
  if( QUERY != n && !transfer_begun )
     nbuffers = n;
  return nbuffers;
}

int
VLImage::Encoding( int mode )
{
#ifdef __alpha__
  if( QUERY != mode )
     {
       return( capture_type = mode );
     } 
  else
    {
       return( capture_type );
    }
#endif

#ifdef __sgi
  if( QUERY != mode )
     {
       if( !server )
	  return( capture_type = mode );
       switch( mode )
	  {
	   case YUV: val.intVal = VL_PACKING_YVYU_422_8; break;
	   case RGB: default: val.intVal = VL_PACKING_RGB_8; break;
	  }
       if( vlSetControl( server, path, drain, VL_PACKING, &val ) < 0 )
	  Error( "set packing control" );
     }
  vlGetControl( server, path, drain, VL_PACKING, &val );
  switch( val.intVal )
     {
      case VL_PACKING_YVYU_422_8: capture_type = YUV; break;
      case VL_PACKING_RGB_8: capture_type = RGB; break;
      default: capture_type = -1; break;
     }
  return capture_type;
#endif
}

int
VLImage::Zoom( int decimation )
{
#ifdef __sgi
  if( QUERY != decimation )
     {
       if( !server )
	  return( zoom = decimation );
       val.fractVal.numerator = 1;
       val.fractVal.denominator = decimation;
       if( vlSetControl( server, path, drain, VL_ZOOM, &val ) < 0 )
	  Error( "set zoom control" );
       vlGetControl( server, path, drain, VL_SIZE, &val );
       capture_height = val.xyVal.y;
       capture_width = val.xyVal.x;
       capture_size = capture_height * capture_width;
     }
  vlGetControl( server, path, drain, VL_ZOOM, &val );
  return val.intVal;
#endif 
}

void 
VLImage::Crop( int x, int y, int width, int height )
{
  if( !server )
    {
      capture_y = y;
      capture_x = x;
      capture_cropw = width;
      capture_croph = height;
      return;
    }

    cerr << "Enter Crop" << endl;

    val.xyVal.y = y;
    val.xyVal.x = x;

    if( vlSetControl( server, path, drain, VL_OFFSET, &val ) < 0 )
      Error( "set OFFSET control" );

    cerr << "  set OFFSET" << endl;

    vlGetControl( server, path, drain, VL_OFFSET, &val );
    capture_y = val.xyVal.y;
    capture_x = val.xyVal.x;

    cerr << "  " << capture_x << " " << capture_y << endl;

    val.xyVal.y = height;
    val.xyVal.x = width;

    if( vlSetControl( server, path, drain, VL_SIZE, &val ) < 0 )
      Error( "set SIZE control" );

    cerr << "  set SIZE" << endl;

    vlGetControl( server, path, drain, VL_SIZE, &val );
    capture_height = val.xyVal.y;
    capture_width = val.xyVal.x;

    cerr << "  " << capture_width << " " << capture_height << endl;

    capture_size = capture_height * capture_width;

    cerr << "Exit Crop" << endl;
}

int
VLImage::Input( int in )
{
#ifdef __alpha__
  if( QUERY != in )
    return( input = in );
  else
    return( input );
#endif // __alpha__

#ifdef __sgi
  if( !server )
     return( input = in );
  if( QUERY != in )
     {
       val.intVal = in;
       if ( vlSetControl( server, path, source, VL_MUXSWITCH, &val ) < 0 )
	 ;
       // Error( "set source control" );
     }
  vlGetControl( server, path, source, VL_MUXSWITCH, &val );
     return( input = in );
  return val.intVal;
#endif // __sgi
}

int
VLImage::Verbose( int in )
{
  // Only set if 0 or 1; else just echo value
  if( 0 == in )
     verbose = 0;
  if( 1 == in )
     verbose = 1;
  return verbose;
}

VLImage::VLImage()
  : size(0), width(0), dataPtr(0)
{}

VLImage&
VLImage::Grab()
{
#ifdef __sgi
#if 0
  if( Type() != capture_type )
     Error( "type mismatch" );
#endif
#define NOLAG
#ifdef NOLAG
  vlGetLatestValid( server, buffer );
  vlPutFree(server,buffer);
#endif
  while( !( info = vlGetLatestValid( server, buffer ) ) )
    sginap( 0 );

  dataPtr = (unsigned char *) vlGetActiveRegion( server, buffer, info );
  size = capture_size;
  width = capture_width;
  return *this;
#endif __sgi
}

// -------- VLRgbImage --------

int VLRgbImage::type = VLImage::RGB;

ostream&
VLRgbImage::Write( ostream& os )
{
  unsigned char *ucp = dataPtr;
  for( int count = size; count--; )
     {
       os.put( ucp[3] );
       os.put( ucp[2] );
       os.put( ucp[1] );
       ucp += 4;
     }
  return os;
}  

VLPixel
VLRgbImage::Pixel( int i )
{
  static VLRgbPixel pixel;
  unsigned char *ucp = dataPtr + i*4 + 1;
  pixel.Blue() = *ucp++;
  pixel.Green() = *ucp++;
  pixel.Red() = *ucp++;
  return pixel;
}

// -------- VLYuvImage --------

int VLYuvImage::type = VLImage::YUV;

ostream&
VLYuvImage::Write( ostream& os )
{
  unsigned char *ucp = dataPtr;
  for( int count = size; count--; )
     {
       os.put( ucp[1] );
       os.put( ucp[0] );
       os.put( ucp[3] );
       os.put( ucp[2] );
       ucp += 4;
     }
  return os;
}

VLPixel
VLYuvImage::Pixel( int i )
{
  static VLYuvPixel pixel;
  pixel.Y() = Y(i);
  pixel.U() = U(i);
  pixel.V() = V(i);
  return pixel;
}

#ifdef TEST_VLIMAGE

void main(int argc, char **argv)

{
  VLImage::NBuffers( 10 );
  VLImage::Encoding( VLImage::RGB );
  VLImage::Zoom( 1 );
  VLImage::Input( atoi(argv[1]) );
  VLImage::Init();
  cerr << VLImage::GetHeight() << " " << VLImage::GetWidth() << endl;

  VLRgbImage picture;

  picture.Grab();
  cout << picture; ;

  VLImage::Close();
}
#endif


