/***************************************************************************
 * 
 * Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

/* backsub-Xclient.c */

extern void cleanup( );

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include <Xm/Command.h>
#include <Xm/DrawingA.h>
#include <Xm/Frame.h>
#include <Xm/PanedW.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/Scale.h>
#include <Xm/Separator.h>
#include <Xm/ToggleB.h>

#include "Xclient.h"

#define MAX_NUM_CANVAS (12)
#define EDITOR  (0)
#define DISPLAY (1)
int num_canvas = 2;

static Widget shell;			/* shell widget */
Widget widgetmanager;
static XtAppContext context;
static GC gc = NULL;
static int width;
static int height;

typedef struct {
  Widget widget;
  Window window;
  Pixmap display;
} CanvasStruct;
  
static CanvasStruct canvas[MAX_NUM_CANVAS];	/* display windows*/

static Display *display;
static int screen;
static int depth;

static double IMAGETHRESH  = 0.40;

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 *
 * GLOBAL FUNCTIONS
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

double X_GetThresh()
{
  return (IMAGETHRESH);
}

void X_SetThresh(float newvalue)
{
 IMAGETHRESH = (double) newvalue;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * local procedure decls
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void X_HandleEvents();

static void edit_action (Widget gw, 
			 XtPointer client_data, 
			 XtPointer call_data);

static void edit_next   (Widget gw, 
			 XtPointer client_data, 
			 XtPointer call_data);

static void edit_clear  (Widget gw, 
			 XtPointer client_data, 
			 XtPointer call_data);

static void edit_quit   (Widget gw, 
			 XtPointer client_data, 
			 XtPointer call_data);


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * callback structs
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

typedef struct {
  XtPointer addr;
  float mult;
  float off;
  char *title;
  int funcp;
} sliderCallbackStruct;

typedef struct {
  XtPointer addr;
  char *title;
  int funcp;
} buttonCallbackStruct;
 
typedef struct {
  PushBCallbackFunc addr;
  char *title;
} pushbCallbackStruct;
 
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * slider_notify
 * button_notify
 * draw_notify
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static void slider_notify (Widget gw, 
			   XtPointer client_data, 
			   XtPointer call_data)
{
  sliderCallbackStruct *client = (sliderCallbackStruct *) client_data;
  XmScaleCallbackStruct *call = (XmScaleCallbackStruct *) call_data;
  int value;
  float position;

  XmScaleGetValue(gw,&value);
  
  position= ((float) value)/100.0;

  if (client->funcp) 
    {
      SliderCallbackFunc callback = (SliderCallbackFunc) client->addr;
      (callback)(position);
    }
  else 
    {
      float *variable = (float *) client->addr;
      *variable = position;
    }

  if (call->reason == XmCR_VALUE_CHANGED) /* XmCR_DRAG */
    {
      fprintf (stderr, "***%s: %lf\n", client->title, position);
    }
} 

static void button_notify (Widget gw, 
			   XtPointer client_data, 
			   XtPointer )
{
  buttonCallbackStruct *data = (buttonCallbackStruct *) client_data;
  int value;
  if (XmToggleButtonGetState(gw))
    {
      value = 1;
    }
  else
    {
      value = 0;
    }
  
  if (data->funcp) 
    {
      ButtonCallbackFunc callback = (ButtonCallbackFunc) data->addr;
      (callback)(value);
    }
  else 
    {
      int *variable = (int *) data->addr;
      *variable = value;
    }
  
  fprintf (stderr, "***%s: %s\n", data->title, 
	   (value? "ON" : "OFF"));
  
}
 
static void pushb_notify (Widget , 
			  XtPointer client_data, 
			  XtPointer )
{
  pushbCallbackStruct *data = (pushbCallbackStruct *) client_data;
  
  (*data->addr)();
  fprintf (stderr, "***called: %s\n", data->title);
}
 
static void draw_notify (Widget , 
			 XtPointer client_data, 
			 XtPointer call_data)
{
  Pixmap *p = (Pixmap*) client_data;
  XmDrawingAreaCallbackStruct *call = (XmDrawingAreaCallbackStruct *) call_data;
  XCopyArea(display, *p, call->window, gc, 0, 0, width, height, 0, 0);
  XFlush(display);
} 


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * X_Initialize
 * X_HandleEvents
 *
 * X_InitializeGC
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static void
InitializeXlib()
{
  static int init = 0;

  if (!init)
    {
      init = 1;
      {				/* set basic globals */
	Arg args[10];
	int nargs;
	
	display = XtDisplay(canvas[0].widget);
	screen = DefaultScreen(display);
	
	nargs = 0;
	XtSetArg(args[nargs],XmNdepth, &depth); nargs++;
	XtGetValues(canvas[0].widget,args,nargs);
      }
      
      {				/* build GC's */
	XGCValues values;
	XtGCMask value_mask;
	
	values.background = WhitePixel(display,screen);
	values.foreground = BlackPixel(display,screen);
	values.fill_style = FillSolid;
	
	value_mask = GCBackground | GCForeground | GCFillStyle;
	
	gc = XtGetGC(canvas[0].widget,value_mask,&values);
      }
      
      {				/* build Pixmaps */
	int index;
	
	for(index=0; index<num_canvas; index++) 
	  {
	    /* get the wid of the canvas */
	    canvas[index].window = XtWindow(canvas[index].widget);
	    
	    /* make a pixmap for it */
	    canvas[index].display = XCreatePixmap(display, 
						  canvas[index].window, 
						  width, 
						  height, 
						  depth);
	    
	    /* clear it */
	    XSetForeground(display,gc,WhitePixel(display,screen));
	    XFillRectangle(display, canvas[index].display, 
			   gc, 0, 0, width, height);

	    XSetForeground(display,gc,BlackPixel(display,screen));
	    XDrawLine(display, canvas[index].display, 
		      gc, 0, 0, width, height);
	    XDrawLine(display, canvas[index].display, 
		      gc, 0, height, width, 0);

	    /* add callback */
	    XtAddCallback (canvas[index].widget, 
			   XmNexposeCallback, 
			   draw_notify, 
			   &(canvas[index].display));
	  }
      }

      XFlush(display);
      X_HandleEvents();
    }
}

#define NUMCOLORS (11)

typedef struct {
  Pixel greys[NUMCOLORS];
  Pixel colors[NUMCOLORS];
} ApplicationResources;
static ApplicationResources appres;

void
X_Initialize( int w, int h)
{

  Arg args[10];
  int nargs=0;
  Widget frame;
  Widget displaymanager;
  Widget button;
  XmString xmstitle;
  static XtResource resources[] = {
    {"color0","Color0",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[0]),XtRString,"white"},
    {"color1","Color1",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[1]),XtRString,"red"},
    {"color2","Color2",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[2]),XtRString,"green"},
    {"color3","Color3",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[3]),XtRString,"light blue"},
    {"color4","Color4",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[4]),XtRString,"orange"},
    {"color5","Color5",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[5]),XtRString,"yellow"},
    {"color6","Color6",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[6]),XtRString,"violet"},
    {"color7","Color7",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[7]),XtRString,"lightseagreen"},
    {"color8","Color8",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[8]),XtRString,"blue violet"},
    {"color9","Color9",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[9]),XtRString,"black"},
    {"color10","Color10",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[10]),XtRString,"brown"},
    {"grey0","Grey0",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[0]),XtRString,"black"},
    {"grey1","Grey1",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[1]),XtRString,"grey10"},
    {"grey2","Grey2",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[2]),XtRString,"grey20"},
    {"grey3","Grey3",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[3]),XtRString,"grey30"},
    {"grey4","Grey4",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[4]),XtRString,"grey40"},
    {"grey5","Grey5",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[5]),XtRString,"grey50"},
    {"grey6","Grey6",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[6]),XtRString,"grey60"},
    {"grey7","Grey7",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[7]),XtRString,"grey70"},
    {"grey8","Grey8",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[8]),XtRString,"grey80"},
    {"grey9","Grey9",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[9]),XtRString,"grey90"},
    {"grey10","Grey10",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[10]),XtRString,"white"}
  };    
  static char *fallback_resources[] = {
    "*foreground:            grey80",
    "*background:            black",
    "*font:                  -*-fixed-*-*-*-*-*-60-*-*-*-*-*-*",
    (char *) NULL
  };

  width = w;
  height = h;
  
  int dummy_argc = 0;

  nargs=0;
  XtSetArg(args[nargs],XmNwidth, 2*width+30); nargs++;
  XtSetArg(args[nargs],XmNheight, height+200); nargs++;
  shell = XtAppInitialize( &context,
			  "Coloredit",
			  NULL, 0,
			  &dummy_argc, NULL,
			  fallback_resources,
			  args, nargs );

  nargs=0;
  XtGetApplicationResources(shell,
			    &appres,
			    resources,
			    XtNumber(resources),
			    args, nargs);
			    
  nargs=0;
  XtSetArg(args[nargs],XmNadjustLast,False); nargs++;
  XtSetArg(args[nargs],XmNsashWidth, 1); nargs++;
  XtSetArg(args[nargs],XmNsashHeight, 1); nargs++;
  XtSetArg(args[nargs],XmNseparatorOn, False); nargs++;
  widgetmanager = XtCreateManagedWidget ("widgetpane", 
					 xmPanedWindowWidgetClass, 
					 shell, 
					 args,nargs);
  nargs=0;
  XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
  XtSetArg(args[nargs],XmNnumColumns,1); nargs++;
  XtSetArg(args[nargs],XmNpacking,XmPACK_COLUMN); nargs++;
  displaymanager = XtCreateManagedWidget ("videopane", 
					  xmRowColumnWidgetClass, 
					  widgetmanager, 
					  args,nargs);
  
  
  frame = XtCreateManagedWidget ("frame",
				 xmFrameWidgetClass,
				 displaymanager,
				 args,nargs);

  nargs=0;
  XtSetArg(args[nargs],XmNwidth, width); nargs++;
  XtSetArg(args[nargs],XmNheight, height); nargs++;
  XtSetArg(args[nargs],XmNresizePolicy, XmRESIZE_NONE); nargs++;
  canvas[0].widget = XtCreateManagedWidget ("edit",
					    xmDrawingAreaWidgetClass,
					    frame,
					    args,nargs);
  
  frame = XtCreateManagedWidget ("frame",
				 xmFrameWidgetClass,
				 displaymanager,
				 args,nargs);
  
  XtAddCallback (canvas[0].widget, XmNinputCallback,
		 edit_action, (XtPointer) NULL);

  nargs=0;
  XtSetArg(args[nargs],XmNwidth, width); nargs++;
  XtSetArg(args[nargs],XmNheight, height); nargs++;
  XtSetArg(args[nargs],XmNresizePolicy, XmRESIZE_NONE); nargs++;
  canvas[1].widget = XtCreateManagedWidget ("video",
					    xmDrawingAreaWidgetClass,
					    frame,
					    args,nargs);
  
  xmstitle = XmStringCreateSimple("Accept Examples / Next Frame");
  nargs=0;
  XtSetArg(args[nargs],XmNheight, 30); nargs++;
  XtSetArg(args[nargs],XmNpaneMinimum, 30); nargs++;
  XtSetArg(args[nargs],XmNlabelString, xmstitle); nargs++;
  button = XtCreateManagedWidget ("Next Frame", 
				  xmPushButtonWidgetClass,
				  widgetmanager, 
				  args,nargs);
  
  XmStringFree(xmstitle);
  XtAddCallback (button, XmNactivateCallback, edit_next, NULL);

  xmstitle = XmStringCreateSimple("Clear Examples");
  nargs=0;
  XtSetArg(args[nargs],XmNheight, 30); nargs++;
  XtSetArg(args[nargs],XmNpaneMinimum, 30); nargs++;
  XtSetArg(args[nargs],XmNlabelString, xmstitle); nargs++;
  button = XtCreateManagedWidget ("Clear", 
				  xmPushButtonWidgetClass,
				  widgetmanager, 
				  args,nargs);
  
  XmStringFree(xmstitle);
  XtAddCallback (button, XmNactivateCallback, edit_clear, NULL);
  
  xmstitle = XmStringCreateSimple("Dump Class and Quit");
  nargs=0;
  XtSetArg(args[nargs],XmNheight, 30); nargs++;
  XtSetArg(args[nargs],XmNpaneMinimum, 30); nargs++;
  XtSetArg(args[nargs],XmNlabelString, xmstitle); nargs++;
  button = XtCreateManagedWidget ("quit", 
				  xmPushButtonWidgetClass,
				  widgetmanager, 
				  args,nargs);
  
  XmStringFree(xmstitle);
  XtAddCallback (button, XmNactivateCallback, edit_quit, NULL);
}

void
X_HandleEvents()
{
  static init = 0;

  if (!init)
    {
      init = 1;
      XtRealizeWidget (shell);
    }
  while(XtAppPending(context))
    {
      XEvent event;
      XtAppNextEvent(context, &event);
      XtDispatchEvent(&event); 
    }
}



void 
X_Close()
{
  int index;
  for(index=0; index<num_canvas; index++) 
    {
      /* free pixmaps */
      XFreePixmap(display,canvas[index].display);
    }

  if(gc)
    XFreeGC(display,gc);
} 

  
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * X_MakeSlider
 * X_MakeButton
 * X_MakeSliderProc
 * X_MakeButtonProc
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


void
X_MakeSliderDo(XPointer addr, char *title,
		float mult, float off, float def,
		int funcp) 
{
  Widget slider;
  Arg args[10];
  int nargs=0;
  sliderCallbackStruct *scb = (sliderCallbackStruct *) 
    malloc (sizeof(sliderCallbackStruct));
  XmString xmstitle;
  scb->addr = addr;
  scb->title = strdup(title);
  scb->mult = mult;
  scb->off = off;
  scb->funcp = funcp;

  xmstitle = XmStringCreateSimple(title);

  nargs=0;
  XtSetArg(args[nargs],XmNshowValue, True); nargs++;
  XtSetArg(args[nargs],XmNtitleString, xmstitle); nargs++;
  XtSetArg(args[nargs],XmNdecimalPoints, 2); nargs++;
  XtSetArg(args[nargs],XmNorientation, XmHORIZONTAL); nargs++;
  XtSetArg(args[nargs],XmNminimum, (int) scb->off*100.0); nargs++;
  XtSetArg(args[nargs],XmNmaximum, (int) (scb->mult+scb->off)*100.0); nargs++;
  XtSetArg(args[nargs],XmNvalue,   (int) (def*100.0)); nargs++;
  XtSetArg(args[nargs],XmNscaleHeight, 20); nargs++;
  XtSetArg(args[nargs],XmNpaneMinimum, 30); nargs++;
  XtSetArg(args[nargs],XmNpaneMinimum, 30); nargs++;
  slider = XtCreateManagedWidget (scb->title, 
				  xmScaleWidgetClass, 
				  widgetmanager, 
				  args,nargs);
  
  XtAddCallback (slider, XmNdragCallback, slider_notify, scb);
  XtAddCallback (slider, XmNvalueChangedCallback, slider_notify, scb);
  XmScaleSetValue(slider, (int) (def*100.0));

  XmStringFree(xmstitle);
}

void
X_MakeSlider( float *addr, char *title,
	       float mult, float off, float def) 
{
  X_MakeSliderDo((XPointer)addr,title,mult,off,def,0);
}
void
X_MakeSliderProc(SliderCallbackFunc addr, char *title,
	       float mult, float off, float def) 
{
  X_MakeSliderDo((XPointer)addr,title,mult,off,def,1);
}


static void
X_MakeButtonDo(XPointer addr, char *title,
		int def,
		int funcp) 
{
  Widget button;
  Arg args[10];
  int nargs=0;
  buttonCallbackStruct *bcb = (buttonCallbackStruct *)
    malloc (sizeof(buttonCallbackStruct));

  bcb->addr = addr;
  bcb->title = strdup(title);
  bcb->funcp = funcp;

  nargs=0;
  XtSetArg(args[nargs],XmNindicatorType, XmN_OF_MANY); nargs++;
  XtSetArg(args[nargs],XmNset, (def ? True : False)); nargs++;
  XtSetArg(args[nargs],XmNheight, 30); nargs++;
  XtSetArg(args[nargs],XmNpaneMinimum, 30); nargs++;
  button = XtCreateManagedWidget (bcb->title, 
				  xmToggleButtonWidgetClass, 
				  widgetmanager, 
				  args,nargs);

  XtAddCallback (button, XmNvalueChangedCallback, button_notify, bcb);
  XmScaleSetValue(button, (int) (def*100.0));
}

void
X_MakeButton( int *addr, char *title,
	       int  def) 
{
  X_MakeButtonDo((XPointer)addr,title,def,0);
}
void
X_MakeButtonProc( ButtonCallbackFunc addr, char *title,
	       int def) 
{
  X_MakeButtonDo((XPointer)addr,title,def,1);
}


 void
X_MakePushButton(PushBCallbackFunc func, char *title)
{
  Widget button;
  Arg args[10];
  int nargs=0;
  pushbCallbackStruct *pcb = (pushbCallbackStruct *)
    malloc (sizeof(pushbCallbackStruct));
  
  pcb->addr = func;
  pcb->title = strdup(title);

  nargs=0;
  XtSetArg(args[nargs],XmNheight, 30); nargs++;
  XtSetArg(args[nargs],XmNpaneMinimum, 30); nargs++;
  button = XtCreateManagedWidget (pcb->title, 
				  xmPushButtonWidgetClass,
				  widgetmanager, 
				  args,nargs);

  XtAddCallback (button, XmNactivateCallback, pushb_notify, pcb);
}

void
X_MakeSeparator()
{
  Widget sep;
  Arg args[10];
  int nargs=0;

  nargs=0;
  sep = XtCreateManagedWidget ("sep", 
			       xmSeparatorWidgetClass,
			       widgetmanager, 
			       args,nargs);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * X_DisplayBitmap
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define MAXPOINTS 19360

static XPoint points[NUMCOLORS][MAXPOINTS];
static int numpoints[NUMCOLORS];

void
X_PaintBit(int color, int x, int y)
{
  color %= NUMCOLORS;
  points[color][numpoints[color]].x = (short) x;
  points[color][numpoints[color]].y = (short) y;
  numpoints[color] = (numpoints[color]+1) % MAXPOINTS;
}

void
X_Sync()
{
  XFlush(display);
}

void
X_CommitPaint(int n)
{
  static int init = 0;
  CanvasStruct *target;
  
  if (!init) 
    {
      InitializeXlib();
      init = 1;
    }
  
  if (n>num_canvas)
    {
      fprintf(stderr,"X_DisplayBitmap: arg out of range: %d\n",n);
      return;
    }
  
  target = &(canvas[n]);
  
  {
    register int i;

    for (i=0;i<NUMCOLORS;i++) {
      XSetForeground(display,gc,appres.colors[i]);
      XDrawPoints(display, target->display, gc, 
		  points[i], numpoints[i], 
		  CoordModeOrigin);
    }
    XCopyArea(display, target->display, target->window, 
	      gc, 0, 0, width, height, 0, 0);

    for (i=0;i<NUMCOLORS;i++)
      numpoints[i] = 0;
  }
}

void
X_NewCanvas(int n)
{
  static int init = 0;
  CanvasStruct *target;
  
  if (!init) 
    {
      InitializeXlib();
      init = 1;
    }
  
  if (n>num_canvas)
    {
      fprintf(stderr,"X_DisplayBitmap: arg out of range: %d\n",n);
      return;
    }
  
  target = &(canvas[n]);
  
  {
    register int i;

    XSetForeground(display,gc,appres.colors[0]);
    XFillRectangle(display, target->display, gc,
		   0,0,width,height);

    for (i=0;i<NUMCOLORS;i++)
      numpoints[i] = 0;
  } 
}

void
X_DisplayBitmap( iveBitmap *bm , int n)
{
  static int init = 0;
  CanvasStruct *target;
  
  if (!init) 
    {
      InitializeXlib();
      init = 1;
    }
  
  if (n>num_canvas)
    {
      fprintf(stderr,"X_DisplayBitmap: arg out of range: %d\n",n);
      return;
    }
  
  
  target = &(canvas[n]);
  
  {
    int i;
    int x,y;
    XPoint points[NUMCOLORS][307200];
    int numpoints[NUMCOLORS];
    {
      register int i;
      for (i=0;i<NUMCOLORS;i++)
	numpoints[i] = 0;
    }

    for (x=0; x<width; x++)
      for (y=0; y<height; y++)
	{
	  int tot=0;
	  
	  tot = iveBitmapGetBit(bm,y,x);
	  tot = ((tot<NUMCOLORS) ? tot : NUMCOLORS-1);
	  points[tot][numpoints[tot]].x = (short) x;
	  points[tot][numpoints[tot]].y = (short) y;
	  numpoints[tot]++;
	}
    
    for (i=1;i<NUMCOLORS;i++) {
      XSetForeground(display,gc,appres.colors[i]);
      XDrawPoints(display, target->display, gc, 
		  points[i], numpoints[i], 
		  CoordModeOrigin);
    }
    XCopyArea(display, target->display, target->window, 
	      gc, 0, 0, width, height, 0, 0);

  }
}

void
X_DrawCross( int x, int y, int n)
{
  static int init = 0;
  CanvasStruct *target;
  
  if (!init) 
    {
      InitializeXlib();
      init = 1;
    }
  
  if (n>num_canvas)
    {
      fprintf(stderr,"X_DrawCross: arg out of range: %d\n",n);
      return;
    }
  
  
  target = &(canvas[n]);
  
  XSetForeground(display,gc,appres.greys[5]);
  XDrawLine(display, target->display, gc, 
	    x-20,y,
	    x+20,y);
  XDrawLine(display, target->display, gc, 
	    x,y-20,
	    x,y+20);
  XCopyArea(display, target->display, target->window, 
	    gc, 0, 0, width, height, 0, 0);

  /* fprintf(stderr,"X_DrawCross: drew it dude...\n"); */
}

void
X_DrawColorCross( int x, int y, int size, int color, int n)
{
  static int init = 0;
  CanvasStruct *target;
  
  if (!init) 
    {
      InitializeXlib();
      init = 1;
    }
  if (n>num_canvas)
    {
      fprintf(stderr,"X_DrawCross: arg out of range: %d\n",n);
      return;
    }
  
  target = &(canvas[n]);
  
  XSetForeground( display, gc, appres.colors[color] );
  XDrawLine( display, target->display, gc, x-size, y, x+size, y );
  XDrawLine( display, target->display, gc, x, y-size, x, y+size );
  XCopyArea( display, target->display, target->window, 
	    gc, 0, 0, width, height, 0, 0);

  /* fprintf(stderr,"X_DrawCross: drew it dude...\n"); */
}


void
X_DrawColorLine( int x, int y, int dx, int dy, int color, int n)
{
  static int init = 0;
  CanvasStruct *target;
  
  if (!init) 
    {
      InitializeXlib();
      init = 1;
    }
  if (n>num_canvas)
    {
      fprintf(stderr,"X_DrawLine: arg out of range: %d\n",n);
      return;
    }

  target = &(canvas[n]);
  
  XSetForeground( display, gc, appres.colors[color] );
  XDrawLine( display, target->display, gc, x, y, x+dx, y+dy );
  XCopyArea( display, target->display, target->window, 
	    gc, 0, 0, width, height, 0, 0);

}

void
X_SetColor(int i, int r, int g, int b)
{
  XColor ccell;
  
  ccell.pixel = appres.colors[i];
  ccell.red = r;
  ccell.green = g;
  ccell.blue = b;
  ccell.flags = DoRed | DoGreen | DoBlue;
  
  /* XStoreColor(display, 
     DefaultColormap(display,DefaultScreen(display)),
     &ccell);
     */
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * X_DisplayImage
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void
X_DisplayImage ( VLYuvImage &picture , int n)
{
  static int init = 0;
  CanvasStruct *target;
  
  if (!init) 
    {
      InitializeXlib();
      init = 1;
    }
  
  if (n>num_canvas)
    {
      fprintf(stderr,"X_DisplayImage: arg out of range: %d\n",n);
      return;
    }
  
  
  target = &(canvas[n]);
  
  {
    int i;
    int x,y;
    XPoint points[NUMCOLORS][307200];
    int numpoints[NUMCOLORS];
    register double thresh;

    {
      register int i;
      for (i=0;i<NUMCOLORS;i++)
	numpoints[i] = 0;
    }

    int max = 0;
    int min = 255;

    for (x=0; x<width; x++)
      for (y=0; y<height; y++)
	{
	  int tot = picture.Y(picture.Index(y,x));
	  if (tot < min) min = tot;
	  if (tot > max) max = tot;
	}
    
    thresh = 0.1 * ((max-min)/2.0 + min);
    
    for (x=0; x<width; x++)
      for (y=0; y<height; y++)
	{
	  int tot=picture.Y(picture.Index(y,x)) / thresh;
	  tot = ((tot<NUMCOLORS) ? tot : NUMCOLORS-1);
	  points[tot][numpoints[tot]].x = (short) x;
	  points[tot][numpoints[tot]].y = (short) y;
	  numpoints[tot]++;
	}
    
    for (i=0;i<NUMCOLORS;i++) {
      XSetForeground(display,gc,appres.greys[i]);
      XDrawPoints(display, target->display, gc, 
		  points[i], numpoints[i], 
		  CoordModeOrigin);
    }
    XCopyArea(display, target->display, target->window, 
	      gc, 0, 0, width, height, 0, 0);

  }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * X_EditClasses
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static editing = 1;		/* -1 not editing, else class number */
static iveBitmap *oldMap = NULL; 
static iveBitmap *newMap = NULL; 
static VLYuvImage *current_picture = NULL;
static int max_classes = 1;	
				/* handle editing events and  */
				/* call the classification routines */

static void edit_quit(Widget , 
		      XtPointer , 
		      XtPointer )
{
  editing = -1;
}

static void edit_clear(Widget , 
		      XtPointer , 
		      XtPointer )
{
  iveBitmapZero( oldMap );
}

static void edit_next(Widget , 
		      XtPointer , 
		      XtPointer )
{
  editing = (editing+1) % max_classes;
  iveBitmapZero( oldMap );
}

extern void SegmentImage(VLYuvImage &picture, 
			 iveBitmap *examples, 
			 iveBitmap *segmentation);

static void edit_action (Widget , 
			 XtPointer , 
			 XtPointer call_data)

{
  int x,y;

  XmDrawingAreaCallbackStruct *call;
  call = (XmDrawingAreaCallbackStruct *) call_data;

  if (call->event->type == ButtonRelease) 
    {
      x = (int) call->event->xbutton.x;
      y = (int) call->event->xbutton.y;
      
      const int RADIUS = 2;

      for (int i=(x-RADIUS<0 ? 0 : x-RADIUS);
	       i<(x+RADIUS>width ? width : x+RADIUS);
	       i++)
	for (int j=(y-RADIUS<0 ? 0 : y-RADIUS);
	         j<(y+RADIUS>height ? height : y+RADIUS);
	         j++)
	  iveBitmapPutBit(oldMap, 
			  j,
			  i, 
			  editing+1);

      X_DisplayImage(*current_picture,EDITOR);
      X_DisplayBitmap( oldMap , EDITOR);

      iveBitmapZero( newMap );
      SegmentImage(*current_picture,oldMap,newMap);
      
      X_DisplayImage(*current_picture,DISPLAY);
      X_DisplayBitmap( newMap , DISPLAY);
      X_Sync();
    }
} 

				/* start an editing session */
				/* pops up the editor and  */
				/* kills it off when done */
				/* actual editing happes in */
				/*        edit_action */

int X_EditClasses(VLYuvImage &picture)
{
  
  editing = 1;
  
  if( oldMap == (iveBitmap *)NULL )
    {
      oldMap = iveBitmapMake( height, width );
      newMap = iveBitmapMake( height, width );
    }
  
  current_picture = &picture;

  iveBitmapZero( oldMap );
  iveBitmapZero( newMap );
  SegmentImage(*current_picture,oldMap,newMap);
  
  X_DisplayImage(*current_picture,DISPLAY);
  X_DisplayBitmap( newMap , DISPLAY);

  X_DisplayImage(picture,EDITOR);

  while(editing >0)
    {
      X_HandleEvents();
    }
  
  if (editing == -1)
    {
      return(0);
    }
  else
    {
      editing=1;
      return(1);
    }
}



