/***************************************************************************
 * 
 * Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

/* color-edit.c++ */

/* RCS: $Header: /v/projects/stive/src/color-edit/color-edit.c++,v 1.3 1999/04/12 18:33:44 cwren Exp cwren $ */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "VLImage.h++"
#include "Xclient.h"

extern "C"
{
#include "iveBitmap.h"
#include "gaussj.h"
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * misc util
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void 
cleanup( )
{
  exit(0);
}

void
help( char *progname )
{
  static char *HelpString = "\n"\
    "\t-in <int>              index of video source\n"\
    "\t                         -3: O2_COMPOSITE\n"\
    "\t                         -2: O2_SVIDEO\n"\
    "\t                         -1: DIGITAL_VIDEO\n"\
    "\t                          0: INDY_COMPOSITE\n"\
    "\t                          1: INDY_SVIDEO\n"\
    "\t                          0: GALILEO_SVIDEO_1\n"\
    "\t                          1: GALILEO_SVIDEO_2\n"\
    "\t                          2: GALILEO_SVIDEO_3\n"\
    "\t                          3: GALILEO_COMPOSITE_1\n"\
    "\t                          4: GALILEO_COMPOSITE_2\n"\
    "\t                          5: GALILEO_COMPOSITE_3\n"\
    "\t                          6: GALILEO_YRB_1\n"\
    "\t                          7: GALILEO_YRB_2\n"\
    "\t-O2S                      O2_SVIDEO\n"\
    "\t-O2C                      O2_COMPOSITE\n"\
    "\t-help                  Prints usage (this message)\n";

  cerr << "usage: " << progname << " [options]" << HelpString; 
  cleanup();
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * class manipulation stuff
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static double class_mean[3];
static double class_covar[3][3];
static long   support=0;
static double class_mean_accumulator[3];
static double class_covar_accumulator[3][3];
static long   examples=0;
static double example_mean_accumulator[3];
static double example_covar_accumulator[3][3];

void ClearClass()
{
  support=0;
  for (int i=0; i<3; i++)
    {
      class_mean[i] = 0.0;
      class_mean_accumulator[i] = 0.0;
      for (int j=0; j<3; j++)
	{
	  class_covar[i][j] = 0.0;
	  class_covar_accumulator[i][j] = 0.0;
	}
    }
}

void ClearExamples()
{
  examples=0;
  for (int i=0; i<3; i++)
    {
      example_mean_accumulator[i] = 0.0;
      for (int j=0; j<3; j++)
	  example_covar_accumulator[i][j] = 0.0;
    }
}

int BuildClass()
{
  if ((support+examples) == 0)
    return(0);

  for (int i=0; i<3; i++)
    {
      class_mean[i] = ((class_mean_accumulator[i] +
			example_mean_accumulator[i]) / 
		       ((double)support+examples));
    }

  static double **inv_covar = NULL;
  static double **dummy = NULL;
  if (inv_covar == (double **) NULL)
    {
      inv_covar = (double **) malloc (4 * sizeof(double*));
      dummy = (double **) malloc (4 * sizeof(double*));
      inv_covar[0] = (double *) malloc (4 * sizeof(double));
      inv_covar[1] = (double *) malloc (4 * sizeof(double));
      inv_covar[2] = (double *) malloc (4 * sizeof(double));
      inv_covar[3] = (double *) malloc (4 * sizeof(double));
      dummy[0] = (double *) malloc (1 * sizeof(double));
      dummy[1] = (double *) malloc (1 * sizeof(double));
      dummy[2] = (double *) malloc (1 * sizeof(double));
      dummy[3] = (double *) malloc (1 * sizeof(double));
      dummy[0][0] = 0.0;
      dummy[1][0] = 0.0;
      dummy[2][0] = 0.0;
      dummy[3][0] = 0.0;
    }
  
  for (i=0; i<3; i++)
    {
      for (int j=0; j<3; j++)
	{
	  inv_covar[i+1][j+1] =  (class_covar_accumulator[i][j] +
				  example_covar_accumulator[i][j]);
	  inv_covar[i+1][j+1] /= (double) support+examples;
	  inv_covar[i+1][j+1] -= class_mean[i]*class_mean[j];
	}
    }

  int ret = gaussj(inv_covar, 3, dummy, 1);

  for (i=0; i<3; i++)
    {
      for (int j=0; j<3; j++)
	{
	  class_covar[i][j] = inv_covar[i+1][j+1];
	}
    }
  
  return ret;
}

void AddExample(double Y, double U, double V)
{
  examples+=1;
  example_mean_accumulator[0] += Y;
  example_mean_accumulator[1] += U;
  example_mean_accumulator[2] += V;
  example_covar_accumulator[0][0] += Y*Y;
  example_covar_accumulator[0][1] += Y*U;
  example_covar_accumulator[0][2] += Y*V;
  example_covar_accumulator[1][0] += U*Y;
  example_covar_accumulator[1][1] += U*U;
  example_covar_accumulator[1][2] += U*V;
  example_covar_accumulator[2][0] += V*Y;
  example_covar_accumulator[2][1] += V*U;
  example_covar_accumulator[2][2] += V*V;
}

float THRESH = 9.0;

int InClass(double Y, double U, double V)
{
  double dist[3];
  
  dist[0] = Y - class_mean[0];
  dist[1] = U - class_mean[1];
  dist[2] = V - class_mean[2];

  double tmp[3];
  for (int i=0; i<3; i++)
    {
      tmp[i] = 0.0;
      for (int j=0; j<3; j++)
	{
	  tmp[i] += dist[i]*class_covar[i][j];
	}
    }

  double distance=0.0;
  for (i=0; i<3; i++)
    {
      distance += dist[i]*tmp[i];
    }
    
  if (distance < THRESH)
    return (1);
  else
    return(0);
}		

void RollInExamples()
{
  support += examples;
  for (int i=0; i<3; i++)
    {
      class_mean_accumulator[i] += example_mean_accumulator[i];
      for (int j=0; j<3; j++)
	{
	  class_covar_accumulator[i][j] += example_covar_accumulator[i][j];
	}
    }
  examples=0;
  for (i=0; i<3; i++)
    {
      example_mean_accumulator[i] = 0.0;
      for (int j=0; j<3; j++)
	example_covar_accumulator[i][j] = 0.0;
    }
}

void SegmentImage(VLYuvImage &picture, 
		  iveBitmap *evidence, 
		  iveBitmap *segmentation)
{
  ClearExamples();

  for(int i=0; i<evidence->width; i++)
    for(int j=0; j<evidence->height; j++)
      if (iveBitmapGetBit(evidence,j,i))
	AddExample(picture.Y(picture.Index(j,i)),
		   picture.U(picture.Index(j,i)),
		   picture.V(picture.Index(j,i)));

  if (BuildClass())
    {
      iveBitmapZero(segmentation);
      for(i=0; i<segmentation->width; i++)
	for(int j=0; j<segmentation->height; j++)
	  if (InClass(picture.Y(picture.Index(j,i)),
		      picture.U(picture.Index(j,i)),
		      picture.V(picture.Index(j,i))))
	    iveBitmapSetBit(segmentation,j,i);
    }
}

void PrintClass()
{  
  BuildClass();
  printf("%f %f %f\n", 
	 class_mean[0],
	 class_mean[1],
	 class_mean[2]);
  for(int i=0; i<3; i++)
    printf("%f %f %f\n", 
	   class_covar[i][0],
	   class_covar[i][1],
	   class_covar[i][2]);
  printf("%f\n", THRESH);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * main
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int 
main(int argc, char *argv[])
{
  int quadrant = -1;
  int zoom = 2;
  int width = 320;
  int height = 240;
  int nbuffers = 3;
  int input_connector = VLImage::INDY_SVIDEO;

  for (int idx=1; idx<argc; idx++)
    {
      if (!strcmp(argv[idx],"-in"))
	{
	  idx++;
	  if (idx<argc)
	    {
	      input_connector = atoi(argv[idx]);
	    } 
	  else 
	    {
	      cerr << "missing argument for -in <int>" << endl;
	      cleanup();
	    }
	}
      else if (!strcmp(argv[idx],"-UL"))
	{
	  zoom = 1;    
	  quadrant = 1;
	}
      else if (!strcmp(argv[idx],"-UR"))
	{
	  zoom = 1;    
	  quadrant = 2;
	}
      else if (!strcmp(argv[idx],"-O2S"))
	{
	  input_connector = -2;
	}
      else if (!strcmp(argv[idx],"-O2C"))
	{
	  input_connector = -3;
	}
      else if (!strcmp(argv[idx],"-help"))
	{
	  help(argv[0]);
	  cleanup();
	}
      else
	{
	  cerr << "unrecognized argument (try -help) :" << argv[idx] << endl;
	  cleanup();
	}	  
    }

  VLImage::NBuffers( nbuffers );
  VLImage::Encoding( VLImage::YUV );
  VLImage::Zoom( zoom );
  if (quadrant == 1)
    VLImage::Crop( 24, 12, 340, 240);
  if (quadrant == 2)
    VLImage::Crop( 360, 12, 340, 240);
  VLImage::Input( input_connector );
  VLImage::Init();
  height = VLImage::GetHeight();
  width = VLImage::GetWidth();
  
  X_Initialize(width, height);
  X_MakeSlider( &THRESH, "Threshold", 30.0, 0.0, THRESH);
  X_HandleEvents();

  ClearClass();
  ClearExamples();

  VLYuvImage picture;
  picture.Grab();
  picture.Free();
  picture.Grab();
  picture.Free();
  picture.Grab();
  picture.Free();
  picture.Grab();
  while (X_EditClasses( picture ))
    {
      RollInExamples();
      picture.Free();
      sleep(2);
      picture.Grab();
      picture.Free();
      picture.Grab();
    } 

  if ((support+examples) >9)
    PrintClass();
  else
    fprintf(stderr,"not enough examples\n");

  cleanup();
}
