#include <FfinderClient.h++>
#include <SffinderClient.h++>
#include <SffFile.h++>
#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <math.h>
#include <unistd.h>

#include <qmKalmanFilter/qmKalmanFilter.h++>

#define _BSD_SIGNALS
#include <signal.h>

extern void dump_dataarray();
extern void update_dataarray(
		      real rx, real ry, real rz, real rix, real riy, real riz,
		      real lx, real ly, real lz, real lix, real liy, real liz,
		      real hx, real hy, real hz, real hix, real hiy, real hiz);
extern void init_dataarray(real dt);

void usage()
{

  cerr << "ffclient -copy ffinderhost" << endl;
  cerr << " or " << endl;
  cerr << "ffclient -filter2 sffinderhost conf.sff" << endl;
  cerr << " or " << endl;
  cerr << "ffclient -filter3 sffinderhost conf.sff" << endl;
  exit(1);
}

int done = 0;
#if defined(__alpha) || defined(IRIX64) 
void done_sig( int )
#else
void done_sig( ... )
#endif
{
  cerr << "Caught Signal" << endl;       
  done=1;
}

qmKalmanFilter * make_filter(real u, real v)
{			     
  int nstate = 4;		// x, y, x', y'
  int nmmt   = 2;		// x, y
  qmVector x(nstate);		// initial state 
  qmMatrix p(nstate,nstate);     // initial covariance
  qmMatrix h(nmmt,nstate);	// measurement model  y=h*x+r*w
  qmMatrix r(nmmt,nmmt);	        // measurement noise 
  qmMatrix ph(nstate,nstate);	// dynamic model     x'=ph*x+q*v
  qmMatrix q(nstate,nstate);	// dynamic noise 
  
  x.setZero();
  x(0) = u;// initial state
  x(1) = v;
  
  p.setIdentity();		// initial error covarince = one pixel
  p *= 0.005;
  p(2,2) = 0.05;		// velocity not so well known
  p(3,3) = 0.05;		  
  r.setIdentity();		// covar mmt noise = one pixel
  r *= 0.005;
  q.setIdentity();		// model noise = ten pixels
  q *= 0.01;
  
  h.setZero();		// measure position only
  h(0,0) = 1.0;			
  h(1,1) = 1.0;
  
  ph.setZero();
  ph(0,0) = 1.0;		//   x   = x  + x'
  ph(0,2) = 1.0;		//
  ph(1,1) = 1.0;		//   y   = y  + y'
  ph(1,3) = 1.0;		//
  ph(2,2) = 1.0;		//   x'  = x'
  ph(3,3) = 1.0;		//   y'  = y'
  
 return new qmKalmanFilter(nstate, // number of state dimensions
			   nmmt,   // number of measurement dimensions
			   x, p, h, r, ph, q );
}

qmKalmanFilter * make_filter(real x, real y, real z)
{			     
  int nstate = 6;		// x, y, z, x', y', z'
  int nmmt   = 3;		// x, y, z
  qmVector X(nstate);		// initial state 
  qmMatrix p(nstate,nstate);     // initial covariance
  qmMatrix h(nmmt,nstate);	// measurement model  y=h*x+r*w
  qmMatrix r(nmmt,nmmt);	        // measurement noise 
  qmMatrix ph(nstate,nstate);	// dynamic model     x'=ph*x+q*v
  qmMatrix q(nstate,nstate);	// dynamic noise 
  
  X.setZero();
  X(0) = x;// initial state
  X(1) = y;
  X(2) = z;
  
  p.setIdentity();		// initial error covarince = one pixel
  p *= 0.005;
  p(3,3) = 0.05;		// velocity not so well known
  p(4,4) = 0.05;		// velocity not so well known
  p(5,5) = 0.05;		  
  r.setIdentity();		// covar mmt noise = one pixel
  r *= 0.005;
  q.setIdentity();		// model noise = ten pixels
  q *= 0.01;
  
  h.setZero();		// measure position only
  h(0,0) = 1.0;			
  h(1,1) = 1.0;
  h(2,2) = 1.0;
  
  ph.setZero();
  ph(0,0) = 1.0;		//   x   = x  + x'
  ph(0,3) = 1.0;		//
  ph(1,1) = 1.0;		//   y   = y  + y'
  ph(1,4) = 1.0;		//
  ph(2,2) = 1.0;		//   z   = z  + z'
  ph(2,5) = 1.0;		//
  ph(3,3) = 1.0;		//   x'  = x'
  ph(4,4) = 1.0;		//   y'  = y'
  ph(5,5) = 1.0;		//   z'  = z'
  
 return new qmKalmanFilter(nstate, // number of state dimensions
			   nmmt,   // number of measurement dimensions
			   X, p, h, r, ph, q );
}

void update_filter (qmKalmanFilter *filter, 
		    real& u, real& v, int& valid)
{
  qmVector sample(2);
  sample(0) = u;
  sample(1) = v;
  filter->H.setZero();	
  if (valid)
    {
      filter->H(0,0) = 1.0;			
      filter->H(1,1) = 1.0;	
    }

  filter->update(sample);
  qmVectorRef Xhatplus1 = filter->getStatePrediction();

  u = Xhatplus1(0); 
  v = Xhatplus1(1);

  real tau = filter->getErrorCovarPrediction()(0,0);
  tau += filter->getErrorCovarPrediction()(1,1);
  valid = (tau < 20.0);
}


void update_filter (qmKalmanFilter *filter, 
		    real& x, real& y, real& z, int& valid)
{
  qmVector sample(3);
  sample(0) = x;
  sample(1) = y;
  sample(2) = z;
  filter->H.setZero();	
  if (valid)
    {
      filter->H(0,0) = 1.0;			
      filter->H(1,1) = 1.0;	
      filter->H(2,2) = 1.0;	
    }

  filter->update(sample);
  qmVector state(6);
  state.copy(filter->getStateEstimate());
  filter->predict(state, 2);

  x = state(0); 
  y = state(1);
  z = state(2);

  real tau = 0.0;
  tau += filter->getErrorCovarPrediction()(0,0);
  tau += filter->getErrorCovarPrediction()(1,1);
  tau += filter->getErrorCovarPrediction()(2,2);
  valid = (tau < 30.0);
}

main (int argc, char **argv)
{
  signal(SIGINT, done_sig);

  if (argc < 2) usage();
      
  if (!strcmp("-copy",argv[1]))      
    {
      if (argc < 4) usage();

      SffFile sffile;
      if( sffile.scan( argv[3] ) )
	exit( 1 );

      FfinderClient::SetTransforms(sffile.world_t,
				   sffile.world_r,
				   sffile.world_scale,
				   sffile.stereo_t,
				   sffile.stereo_q);

      FfinderClient ffinders[2];
      ffinders[FfinderClient::RIGHT].connect(sffile.ffinder_right);
      ffinders[FfinderClient::LEFT].connect(sffile.ffinder_left);

      SffinderClient sffinder(argv[2]);
      
      ffPerson prior;

      while(!done)
	{
	  sffPersonPlus* p = sffinder.getNextPersonPlus();	
	  prior.frameNumber = p->frameNumber;

	  struct timeval tv;
	  gettimeofday(&tv);
	  prior.timestampSec = tv.tv_sec;
	  prior.timestampUsec = tv.tv_usec;
	  
	  prior.personExists = 1;
	    
	  for (int side=0; side<2; side++)
	    {
	      FfinderClient::setPriorFrom2D(prior.head, 
					    p->head,
					    side);
	      
	      FfinderClient::setPriorFrom2D(prior.leftHand, 
					    p->leftHand,
					    side);
	      
	      FfinderClient::setPriorFrom2D(prior.rightHand, 
					    p->rightHand,
					    side);

	      ffinders[side].setPriors(&prior);
	    }
	}
      
      for (int side=0; side<2; side++)
	{
	  prior.rightHand.valid=0;
	  prior.leftHand.valid=0;
	  prior.head.valid=0;
	  ffinders[side].setPriors(&prior);
	}
    }
  else if (!strcmp("-filter2",argv[1]))      
    {
      if (argc < 4) usage();

      SffFile sffile;
      if( sffile.scan( argv[3] ) )
	exit( 1 );

      FfinderClient::SetTransforms(sffile.world_t,
				   sffile.world_r,
				   sffile.world_scale,
				   sffile.stereo_t,
				   sffile.stereo_q);

      FfinderClient ffinders[2];
      ffinders[FfinderClient::RIGHT].connect(sffile.ffinder_right);
      ffinders[FfinderClient::LEFT].connect(sffile.ffinder_left);

      SffinderClient sffinder(argv[2]);
      
      ffPerson prior;

      qmKalmanFilter *lfilter[2];
      qmKalmanFilter *rfilter[2];
      
      {
	sffPersonPlus* p = sffinder.getNextPersonPlus();	
	prior.frameNumber = p->frameNumber;
	
	struct timeval tv;
	gettimeofday(&tv);
	prior.timestampSec = tv.tv_sec;
	prior.timestampUsec = tv.tv_usec;
	
	prior.personExists = 1;
	
	
	for (int side=0; side<2; side++)
	  {
	    FfinderClient::setPriorFrom2D(prior.leftHand, 
					  p->leftHand,
					  side);
	    
	    FfinderClient::setPriorFrom2D(prior.rightHand, 
					  p->rightHand,
					  side);
	    
	    lfilter[side] = make_filter(prior.leftHand.u,
					 prior.leftHand.v);
	    rfilter[side] = make_filter(prior.rightHand.u,
					 prior.rightHand.v);

	  }
      }
	      

      while (!done)
	{
	  sffPersonPlus* p = sffinder.getNextPersonPlus();	
	  prior.frameNumber = p->frameNumber;
	  
	  struct timeval tv;
	  gettimeofday(&tv);
	  prior.timestampSec = tv.tv_sec;
	  prior.timestampUsec = tv.tv_usec;
	  
	  prior.personExists = 1;
	  

	  for (int side=0; side<2; side++)
	    {
	      FfinderClient::setPriorFrom2D(prior.head, 
					    p->head,
					    side);
	      
	      FfinderClient::setPriorFrom2D(prior.leftHand, 
					    p->leftHand,
					    side);
	      
	      FfinderClient::setPriorFrom2D(prior.rightHand, 
					    p->rightHand,
					    side);
	      
	      update_filter(lfilter[side], 
			    prior.leftHand.u, 
			    prior.leftHand.v, 
			    prior.leftHand.valid);
	      
	      prior.leftHand.suu = 0.001;
	      prior.leftHand.svv = 0.001;
	      prior.leftHand.suv = 0.0;
	      
	      update_filter(rfilter[side], 
			    prior.rightHand.u, 
			    prior.rightHand.v, 
			    prior.rightHand.valid);
	      prior.rightHand.suu = 0.001;
	      prior.rightHand.svv = 0.001;
	      prior.rightHand.suv = 0.0;
	      
	      ffinders[side].setPriors(&prior);
	    }
	}

      for (int side=0; side<2; side++)
	{
	  prior.rightHand.valid=0;
	  prior.leftHand.valid=0;
	  prior.head.valid=0;
	  ffinders[side].setPriors(&prior);
	}
    }
  else if (!strcmp("-filter3",argv[1]))      
    {
      if (argc < 4) usage();

      init_dataarray(1/30.0);

      SffFile sffile;
      if( sffile.scan( argv[3] ) )
	exit( 1 );

      FfinderClient::SetTransforms(sffile.world_t,
				   sffile.world_r,
				   sffile.world_scale,
				   sffile.stereo_t,
				   sffile.stereo_q);

      FfinderClient ffinders[2];
      ffinders[FfinderClient::RIGHT].connect(sffile.ffinder_right);
      ffinders[FfinderClient::LEFT].connect(sffile.ffinder_left);

      SffinderClient sffinder(argv[2]);
      
      ffPerson prior;

      sffPersonPlus* p = sffinder.getNextPersonPlus();	 

      qmKalmanFilter *lfilter = make_filter(p->leftHand.world_ref.t[0],
					    p->leftHand.world_ref.t[1],
					    p->leftHand.world_ref.t[2]);
      qmKalmanFilter *rfilter = make_filter(p->rightHand.world_ref.t[0],
					    p->rightHand.world_ref.t[1],
					    p->rightHand.world_ref.t[2]);
      qmKalmanFilter *hfilter = make_filter(p->head.world_ref.t[0],
					    p->head.world_ref.t[1],
					    p->head.world_ref.t[2]);

      while (!done)
	{
	  sffPersonPlus* p = sffinder.getNextPersonPlus();	
	  prior.frameNumber = p->frameNumber;
	  
	  struct timeval tv;
	  gettimeofday(&tv);
	  prior.timestampSec = tv.tv_sec;
	  prior.timestampUsec = tv.tv_usec;
	  
	  prior.personExists = 1;
	  
	  real x =  p->leftHand.world_ref.t[0];
	  real y =  p->leftHand.world_ref.t[1];
	  real z =  p->leftHand.world_ref.t[2];
	  
	  update_filter(lfilter, x, y, z, p->leftHand.valid);
	  
	  p->leftHand.world_ref.t[0] = x;
	  p->leftHand.world_ref.t[1] = y;
	  p->leftHand.world_ref.t[2] = z;
	  
	  x =  p->rightHand.world_ref.t[0];
	  y =  p->rightHand.world_ref.t[1];
	  z =  p->rightHand.world_ref.t[2];
	  
	  update_filter(rfilter, x, y, z, p->rightHand.valid);
	  
	  p->rightHand.world_ref.t[0] = x;
	  p->rightHand.world_ref.t[1] = y;
	  p->rightHand.world_ref.t[2] = z;
	  
	  x =  p->head.world_ref.t[0];
	  y =  p->head.world_ref.t[1];
	  z =  p->head.world_ref.t[2];
	  
	  update_filter(hfilter, x, y, z, p->head.valid);
	  
	  p->head.world_ref.t[0] = x;
	  p->head.world_ref.t[1] = y;
	  p->head.world_ref.t[2] = z;
	  
	  {
	    qmVectorRef rpos = rfilter->getStateEstimate();
	    qmVectorRef rino = rfilter->getStateInnovation();
	    qmVectorRef lpos = lfilter->getStateEstimate();
	    qmVectorRef lino = lfilter->getStateInnovation();
	    qmVectorRef hpos = hfilter->getStateEstimate();
	    qmVectorRef hino = hfilter->getStateInnovation();
	    
	    update_dataarray(rpos(0),rpos(1),rpos(2),rino(0),rino(1),rino(2),
			     lpos(0),lpos(1),lpos(2),lino(0),lino(1),lino(2),
			     hpos(0),hpos(1),hpos(2),hino(0),hino(1),hino(2));
	  }

	  for (int side=0; side<2; side++)
	    {
	      FfinderClient::setPriorFrom3D(prior.head, 
					    p->head,
					    side);
	      
	      FfinderClient::setPriorFrom3D(prior.leftHand, 
					    p->leftHand,
					    side);
	      
	      FfinderClient::setPriorFrom3D(prior.rightHand, 
					    p->rightHand,
					    side);
	      
	      ffinders[side].setPriors(&prior);
	    }
	}

      cerr << "dumping data array" << endl;
      dump_dataarray();

      for (int side=0; side<2; side++)
	{
	  prior.rightHand.valid=0;
	  prior.leftHand.valid=0;
	  prior.head.valid=0;
	  ffinders[side].setPriors(&prior);
	}
    }
  else
    {
      usage();
    }  
}
