#include <FfinderClient.h++>
#include <SffinderClient.h++>
#include <SffFile.h++>
#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <math.h>
#include <unistd.h>

#define _BSD_SIGNALS
#include <signal.h>

void usage()
{
  cerr << "ffclient -prior  ffinderhost u v suu svv suv" << endl;
  cerr << " or " << endl;
  cerr << "ffclient -stive2 sffinderhost conf.sff" << endl;
  cerr << " or " << endl;
  cerr << "ffclient -stive3 sffinderhost conf.sff" << endl;
  cerr << " or " << endl;
  cerr << "ffclient -test   conf.sff" << endl;
  exit(1);
}

int done = 0;
#if defined(__alpha) || defined(IRIX64) 
void done_sig( int )
#else
void done_sig( ... )
#endif
{
  cerr << "Caught Signal" << endl;       
  done=1;
}

main (int argc, char **argv)
{
  signal(SIGINT, done_sig);

  if (argc < 2) usage();
      
  if (!strcmp("-stive2",argv[1]))      
    {
      if (argc < 4) usage();

      SffFile sffile;
      if( sffile.scan( argv[3] ) )
	exit( 1 );

      FfinderClient::SetTransforms(sffile.world_t,
				   sffile.world_r,
				   sffile.world_scale,
				   sffile.stereo_t,
				   sffile.stereo_q);

      FfinderClient ffinders[2];
      ffinders[FfinderClient::RIGHT].connect(sffile.ffinder_right);
      ffinders[FfinderClient::LEFT].connect(sffile.ffinder_left);

      SffinderClient sffinder(argv[2]);
      
      ffPerson prior;

      while(!done)
	{
	  sffPersonPlus* p = sffinder.getNextPersonPlus();	
	  prior.frameNumber = p->frameNumber;

	  struct timeval tv;
	  gettimeofday(&tv);
	  prior.timestampSec = tv.tv_sec;
	  prior.timestampUsec = tv.tv_usec;
	  
	  prior.personExists = 1;
	    
	  for (int side=0; side<2; side++)
	    {
	      FfinderClient::setPriorFrom2D(prior.head, 
					    p->head,
					    side);
	      
	      FfinderClient::setPriorFrom2D(prior.leftHand, 
					    p->leftHand,
					    side);
	      
	      FfinderClient::setPriorFrom2D(prior.rightHand, 
					    p->rightHand,
					    side);

	      ffinders[side].setPriors(&prior);
	    }
	}
      
      for (int side=0; side<2; side++)
	{
	  prior.rightHand.valid=0;
	  prior.leftHand.valid=0;
	  prior.head.valid=0;
	  ffinders[side].setPriors(&prior);
	}
      exit(0);
    }
  if (!strcmp("-stive3",argv[1]))      
    {
      if (argc < 3) usage();

      SffFile sffile;
      if( sffile.scan( argv[3] ) )
	exit( 1 );

      FfinderClient::SetTransforms(sffile.world_t,
				   sffile.world_r,
				   sffile.world_scale,
				   sffile.stereo_t,
				   sffile.stereo_q);

      FfinderClient ffinders[2];
      ffinders[FfinderClient::LEFT].connect(sffile.ffinder_left);
      ffinders[FfinderClient::RIGHT].connect(sffile.ffinder_right);

      SffinderClient sffinder(argv[2]);
      
      ffPerson prior;

      while(!done)
	{
	  sffPersonPlus* p = sffinder.getNextPersonPlus();	
	  prior.frameNumber = p->frameNumber;

	  struct timeval tv;
	  gettimeofday(&tv);
	  prior.timestampSec = tv.tv_sec;
	  prior.timestampUsec = tv.tv_usec;
	  
	  prior.personExists = 1;
	    
	  for (int side=0; side<2; side++)
	    {
	      FfinderClient::setPriorFrom3D(prior.head, 
					    p->head,
					    side);
	      
	      FfinderClient::setPriorFrom3D(prior.leftHand, 
					    p->leftHand,
					    side);
	      
	      FfinderClient::setPriorFrom3D(prior.rightHand, 
					    p->rightHand,
					    side);
	      
	      ffinders[side].setPriors(&prior);
	    }
	}
      
      for (int side=0; side<2; side++)
	{
	  prior.rightHand.valid=0;
	  prior.leftHand.valid=0;
	  prior.head.valid=0;
	  ffinders[side].setPriors(&prior);
	}
      exit(0);
    }
  if (!strcmp("-test",argv[1]))      
    {
      if (argc < 3) usage();

      SffFile sffile;
      if( sffile.scan( argv[2] ) )
	exit( 1 );

      FfinderClient::SetTransforms(sffile.world_t,
				   sffile.world_r,
				   sffile.world_scale,
				   sffile.stereo_t,
				   sffile.stereo_q);

      qmUnitQuaternion q;
      qmVector t(3);
      qmVector s(3);

      cin >> t;
      cin >> q;
      cin >> s;
      
      q.setNormal();

      cerr << "world trans,rot,scale" << endl;
      cerr << t << endl;
      cerr << q << endl;
      cerr << s << endl;
      
      double trans[3];
      double scale[3];
      
      for(int k=0; k<3; k++)
	{
	  trans[k] = t(k);
	  scale[k] = s(k);
	}
      
      double quat[4];

      for(int l=0; l<4; l++)
	{
	  quat[l] = q(l);
	}

      float mean[3];
      float covar[9];

      FfinderClient::PriorFromEig(mean,
				  covar,
				  trans,
				  quat,
				  scale);
      
      qmVectorRef m(3,mean);
      qmMatrixRef c(3,3,covar);

      cerr << "world mean and covar" << endl;
      cerr << m << endl;
      cerr << c << endl;

      FfinderClient::TransformFromWorld(FfinderClient::RIGHT,
      					mean,
      					covar);
      
      cerr << "world->camera mean and covar" << endl;
      cerr << m << endl;
      cerr << c << endl;

      ffPerson prior;
      FfinderClient::Project(prior.head, 
			     mean, 
			     covar);
      
      cerr << "proj mean and covar" << endl;
      cerr << prior.head.u << " ";
      cerr << prior.head.v << endl;
      cerr << prior.head.suu << " ";
      cerr << prior.head.suv << endl;
      cerr << prior.head.suv << " ";
      cerr << prior.head.svv << endl;
      cout << prior.head.u << " ";
      cout << prior.head.v << " ";
      cout << prior.head.suu << " ";
      cout << prior.head.svv << " ";
      cout << prior.head.suv << endl;

    }
  else
    {
      FfinderClient ffinder(argv[2]);

      if (argc < 8) usage();

      ffPerson prior;
      prior.frameNumber = 0;
      prior.timestampSec = 0;
      prior.timestampUsec = 0;
      prior.personExists = 1;
      prior.leftHand.valid=0;  
      prior.rightHand.valid=0;  
      prior.head.valid=1;
      prior.head.u = atof(argv[3]);
      prior.head.v = atof(argv[4]);
      prior.head.suu = atof(argv[5]);
      prior.head.svv = atof(argv[6]);
      prior.head.suv = atof(argv[7]);
      
      ffinder.setPriors(&prior);
    }
}
