// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.  These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.

// CellImage.c++

#include "CellImage.h++"
#include "QuickMath/qmMinMax.h++"
Coords CellImage::tmp;
void
CellImage::error( char* fn, char* msg )
{
    cerr << "CellImage::" << fn << "(): ERROR: " << msg << endl;
    // exit( 1 );
    assert( 0 );
}
void
CellImage::init( StateImage<Pixel>* pmap, int nw, int nh )
{
    cellWidth = nw;
    cellHeight = nh;
    int pw = pmap -> getWidth();
    int ph = pmap -> getHeight();
    if( pw<=0 || ph<=0 || nw<=0 || nh<=0 )
       error( "init", "bad args" );
    int cw = (pw%nw) ? pw/nw+1 : pw/nw;
    int ch = (ph%nh) ? ph/nh+1 : ph/nh;
    StateImage<Cell>::init( cw, ch );
    pixelMap = pmap;
    int chunk = 2*nw*nh;
    pixels = new Pixel*[ cw*ch*chunk ];
    Pixel** p = pixels;
    for( int j = 0; j < ch; j++ )
       for( int i = 0; i < cw; i++ )
	  {
	      initCell( i, j, nw, nh, pw, ph, p );
	      p += chunk;
	  }
}
void
CellImage::initCell( int i, int j, int nw, int nh, int pw, int ph, Pixel** p )
{
    Cell* cp = operator()( i, j );
    int ufirst = limit( i*nw, 0, pw-1 );
    int vfirst = limit( j*nh, 0, ph-1 );
    int upast = min( ufirst+nw, pw );
    int vpast = min( vfirst+nh, ph );
    int cw = upast-ufirst;
    int ch = vpast-vfirst;
    cp -> init( cw, ch, p );
    for( int jj = 0; jj < ch; jj++ )
       for( int ii = 0; ii < cw; ii++ )
	  cp -> setPixel( ii, jj, (*pixelMap)( ufirst+ii, vfirst+jj ) );
}
