// SffinderClient.c++

#include "FfinderClient.h++"
#include <unistd.h>
#include <limits.h>
#include <stream.h>
#include "Useless/iotools.h++"

qmMatrix FfinderClient::world_rotation(3,3);
qmVector FfinderClient::world_translation(3);
float    FfinderClient::world_scale;
qmMatrix FfinderClient::stereo_rotation(3,3);
qmVector FfinderClient::stereo_translation(3);


FfinderClient::FfinderClient( char* serverHostname , int increment_rpc_port )
  : client( 0 ), lastFrame( -1 ), lastGot( GOTNONE )
{
    person.rightHand.valid = 0;
    person.leftHand.valid = 0;
    person.head.valid = 0;
    if( serverHostname )
       if( connect( serverHostname ) )
	  cerr << "connection failed" << endl;
}

FfinderClient::~FfinderClient()
{
  if(client != (CLIENT*) NULL)
    clnt_destroy(client);
}

int
FfinderClient::connect( char* serverHostname  , int increment_rpc_port )
{
    const int tryMax = 100;
    const int tryInterval = CLK_TCK;
    int tryCount = 0;

    cerr << "connecting to ffinder server on " << serverHostname << endl;
    while( tryCount++ < tryMax &&
	   !( client = clnt_create
	      ( serverHostname, FFINDERPROG+increment_rpc_port,
		FFINDERVERS, "tcp") ) )
       {
	   clnt_pcreateerror( serverHostname );
	   ::sleep( 100 );
       }
    if( tryCount >= tryMax )
       return 1;
    cerr << "connection successful" << endl;
    return 0;
}

ffPerson*
FfinderClient::getPerson()
{
    ffPerson* p = 0;
    if( !client )
       return 0;
    if( !( p = ffget_current_1( 0, client ) ) )
       {
	   clnt_perror( client, "ffget_current_1" );
	   return 0;
       }
    person = *p;
    lastFrame = p->frameNumber;
    lastGot = GOTPERSON;

    void *result;
    if ( !( result = ffframe_1( (void *)0, client ) ) )
      clnt_perror( client, "ffframe_1" );

    return &person;
}

ffPerson*
FfinderClient::getNextPerson()
{
    ffPerson* p = 0;
    if( !client )
       return 0;
    if( !( p = ffget_frame_result_1( 0, client ) ) )
       {
	   clnt_perror( client, "ffget_frame_result_1" );
	   return 0;
       }
    person = *p;
    lastFrame = p->frameNumber;
    lastGot = GOTPERSON;
    void *result;
    if ( !( result = ffframe_1( (void *)0, client ) ) )
      clnt_perror( client, "ffframe_1" );

    return &person;
}

ostream&
FfinderClient::print( ostream& os )
{
    return os;
}

void
FfinderClient::init()
{
    void* p;
    if( !client )
       return;
    if( !( p = ffforceinit_1( 0, client ) ) )
       {
	   clnt_perror( client, "ffforceinit_1" );
       }
}

void
FfinderClient::setPriors(ffPerson *priors)
{
    void* p;
    if( !client )
       return;
    if( !( p = ffset_priors_1( priors, client ) ) )
       {
	   clnt_perror( client, "ffset_priors_1" );
       }
}

