// noErfBlob.h++

#ifndef _noErfBlob_H_
#define _noErfBlob_H_

#include <math.h>
#include "qmNonlinearOptimize/noErrorFunction.h++"

class noErfBlob : public noErrorFunction
{
 protected:
    int inited;
    qmVector y_obs;
    qmMatrix covL_obs;
    qmMatrix covR_obs;
    real beta;
    qmVector Trl;
    qmUnitQuaternion Qrl;
    qmMatrix Rrl;
    int freeCount;
    int freeMap[9];
 public:
    static int verbose;
    // Mmts: (u,v)= center, (Suu,Svv,Suv)= variances & covariance
    // State: (tx,ty,tz)= center, (q)=rotation, (l)= stdev
    enum _mmts { UL, VL, SuuL, SvvL, SuvL, UR, VR, SuuR, SvvR, SuvR };
    enum _state { TX, TY, TZ, Q0, QX, QY, QZ, LX, LY, LZ };
    enum _dstate { DTX, DTY, DTZ, DWX, DWY, DWZ, DLX, DLY, DLZ };
    noErfBlob()
      : noErrorFunction( nMmts(), nStates(), 9 ),
	inited( 0 ),
	y_obs( nMmts() ),
	covL_obs( 2, 2 ),
	covR_obs( 2, 2 ),
	beta( 1. ),
	Trl( 3 ),
	Qrl(),
	Rrl( 3, 3 ),
	freeCount( 9 )
    { for( int i = 0; i < 9; i++ ) freeMap[i] = i; }
    ~noErfBlob() {}
    void init( const qmVectorRef& mL,
	       const qmMatrixRef& c2L, 
	       const qmVectorRef& mR,
	       const qmMatrixRef& c2R,
	       const real b,
	       const qmVectorRef& t,
	       const qmUnitQuaternionRef& q );
    void init( const qmVectorRef& y,
	       const real b,
	       const qmVectorRef& t,
	       const qmUnitQuaternionRef& q );
    void initStereo( const real b, 
		     const qmVectorRef& t,
		     const qmUnitQuaternionRef& q );
    void forward( qmVectorRef& dy, 
		  qmVectorRef& t, 
		  qmUnitQuaternionRef& q,
		  qmVectorRef& l, // stdev
		  real b );
    void forward( qmVectorRef& dy, qmVectorRef& x, real b );
    void _forward( qmVectorRef& dy, 
		   qmVectorRef& t, qmMatrixRef& r, qmMatrixRef& L2, real b );
    void jacobian( qmMatrixRef& J, const qmVectorRef& x, real b, 
		   const qmVectorRef& Tc, const qmMatrixRef& Rc );
    void _forwardBoth( qmVectorRef& y, const qmVectorRef& x );
    // void _jacobianBoth( qmMatrixRef& J, const qmVectorRef& x );
    virtual const qmVectorRef& operator()( const qmVectorRef& x );
    // **
    virtual const qmMatrixRef& jacobian( const qmVectorRef& x );
    // **
    virtual const qmVectorRef& add( const qmVectorRef& x, const qmVectorRef& dx );
    static void setVerbose( int v ) { verbose = v; }
    inline int nMmts() { return 10; }
    inline int nStates() { return 10; }
    // **
    // static inline int nDStates() { return 9; }
    inline int nDStates() { return( freeCount ); }
    // **
    void computeLinearInit( qmVectorRef& x );
    void setFree( int[] );
};

#endif /* _noErfBlob_H_ */
