// noErfExponential.c++

#include "noErfExponential.h++"

void
noErfExponential::init( qmVectorRef& y_in )
{
    noErrorFunction::init( y_in.getLength(), N, N );
    y.set( y_in.getLength() );
    y = y_in;
}

const
qmVectorRef& 
noErfExponential::operator()( const qmVectorRef& x_in )
{
    // y(i) = AMPL * exp( - RATE * ti )
    qmVector x = x_in;
    for( int i = 0; i < getNdy(); i++ )
       dy_return( i ) = y(i) - x(AMPL) * exp( -x(RATE) * double(i) );
    return dy_return;
}

const
qmMatrixRef&
noErfExponential::jacobian( const qmVectorRef& x_in )
{
    // ddy(i)/dA = - exp( - RATE * ti )
    // ddy(i)/dR = A ti exp( - RATE * ti )
    qmVector x = x_in;
    double eRt;
    for( int i = 0; i < getNdy(); i++ )
       {
	   eRt = exp( -x(RATE) * double(i) );
	   J_return( i, AMPL ) = - eRt;
	   J_return( i, RATE ) = x(AMPL) * double(i) * eRt;
       }
    return J_return;
}

const
qmVectorRef&
noErfExponential::add( const qmVectorRef& x_in, const qmVectorRef& dx )
{
    return x_return.setAdd( x_in, dx );
}

qmVectorRef&
noErfExponential::generateData( double a, double r, qmVectorRef& y_out )
{
    for( int i = 0; i < y_out.getLength(); i++ )
       y_out( i ) = a * exp( -r * double(i) );
    return y_out;
}
