// noErfQuirk.h++

#ifndef _noErfQuirk_H_
#define _noErfQuirk_H_

    // x = ( P:0, ... , P:N-1, C:0, ... , C:M-1 )
    //    P:k = ( X:k, Y:k, Z:k )
    //    C:t = ( TX:t, TY:t, TZ:t, Q0:t, QX:t, QY:t, QZ:t, B:t )
    //
    // dx = ( P:0, ... , P:N-1, C:0, ... , C:M-1 )
    //    dP:k = ( dX:k, dY:k, dZ:k )
    //    dC:t = ( dTX:t, dTY:t, dTZ:t, dWX:t, dWY:t, dWZ:t, dB:t )
    //
    // y = ( (Y:0, ..., Y:Nt-1):t=0, ... , (Y:0, ..., Y:Nt-1):t=M-1 )
    // 

#include "qmNonlinearOptimize/noErrorFunction.h++"

class noErfQuirk : public noErrorFunction
{
    int pointCount;
    int frameCount;
    int pFreeCount;
    int* pFreeMap;
    int mFreeCount;
    int* mFreeMap;
    int yValidCount;
    int** yValid;
    int betaFixed;
    qmVector y_obs;

    inline int getYCount() { return 2*yValidCount; }
    inline int getXCount() { return( 3*pointCount + 8*frameCount ); }
    inline int getDxCount() { return( 3*pFreeCount + 6*mFreeCount +
				      ((betaFixed) ? 0 : 1 )); }

    inline int isFreePoint( int k ) { return( -1 != pFreeMap[k] ); }
    inline int isFreeCamera( int t ) { return( -1 != mFreeMap[t] ); }

    inline const qmVectorRef& cameraPoint( const qmVectorRef& t,
					   const qmUnitQuaternionRef& q,
					   const qmVectorRef& p );
    inline const qmVectorRef& project( real b, const qmVectorRef& p );
    static qmVector projPoint;
    static qmVector camPoint;
    static qmMatrix rotMat;
    static qmVector tmpPoint;
    inline int dxPointIndex( int k, int i );
    inline int dxTransIndex( int t, int i );
    inline int dxRotIndex( int t, int i );
    inline int dxBetaIndex( int t );
    void setdP( qmMatrixRef&, int iy, int k, qmMatrixRef& dyddx );
    void setdT( qmMatrixRef&, int iy, int t, qmMatrixRef& dyddx );
    void setdW( qmMatrixRef&, int iy, int t, qmMatrixRef& dyddx );
    void setdB( qmMatrixRef&, int iy, int t, qmMatrixRef& dyddx );
    const qmVectorRef& dxPoints( const qmVectorRef& dx );
    inline const qmVectorRef& dxTrans( const qmVectorRef& dx, int t );
    inline const qmVectorRef& dxRot( const qmVectorRef& dx, int t );
    inline real dxBeta( const qmVectorRef& dx );
public:
    noErfQuirk()
      : noErrorFunction(),
	pointCount(0), frameCount(0) {}
    enum points { X, Y, Z };
    enum cameras { TX, TY, TZ, Q0, QX, QY, QZ, B };
    // void init( int N, int M, real* y_data );
    // void init( int N, int M, real* y_data, int* map );
    // void init( int N, int M, int* plist, int** ylist, qmVectorRef& y );
    void init( int N, int M, int* plist, int** ylist, qmVectorRef& y,
	       int* mlist = 0 );
    void _forward( qmVectorRef& y, const qmVectorRef& x );
    const qmVectorRef& operator()( const qmVectorRef& x );
    const qmMatrixRef& jacobian( const qmVectorRef& x );
    const qmVectorRef& add( const qmVectorRef& x,
			    const qmVectorRef& dx );
    static ostream& printState( ostream&, qmVectorRef&, int N, int M );
    // static ostream& printObs( ostream&, qmVectorRef&, int N, int M );
    static istream& scanState( istream&, qmVectorRef&, int N, int M );
    // static istream& scanObs( istream&, qmVectorRef&, int N, int M );
    static inline qmVectorRef get3DPoint( const qmVectorRef&, int k );
    static inline qmVectorRef getCamera( const qmVectorRef&, int t, int N );
    static inline qmVectorRef getTrans( const qmVectorRef&, int t, int N );
    static inline qmUnitQuaternionRef getRot ( const qmVectorRef&, int, int );
    static inline real getBeta( const qmVectorRef&, int t, int N );
};

inline
qmVectorRef
noErfQuirk::get3DPoint( const qmVectorRef& x, int k )
{
    return x.subVector( 3*k, 3 );
}

inline
qmVectorRef
noErfQuirk::getCamera( const qmVectorRef& x, int t, int N )
{
    return x.subVector( 3*N + 8*t, 8 );
}

inline
qmVectorRef
noErfQuirk::getTrans( const qmVectorRef& x, int t, int N )
{
    return x.subVector( 3*N + 8*t, 3 );
}

inline
qmUnitQuaternionRef
noErfQuirk::getRot( const qmVectorRef& x, int t, int N )
{
    return qmUnitQuaternionRef( x.subVector( 3*N + 8*t + 3, 4 ) );
}

inline
real
noErfQuirk::getBeta( const qmVectorRef& x, int t, int N )
{
    return x.getValue( 3*N + 8*t + 7 );
}

#endif /* _noErfQuirk_H_ */

