// noErfRelativeOrientation.c++

#include "noErfRelativeOrientation.h++"
#include "QuickMath/qmMinMax.h++"

#if 0
#include <stream.h>
static
inline
ostream&
alert( char *msg, ostream& os = cerr )
{
    os << "noErfRelativeOrientation: ALERT: " << msg << endl;
    return os;
}
#endif

int noErfRelativeOrientation::verbose = 0;

void
noErfRelativeOrientation::init( qmVectorRef& uL_in,
				qmVectorRef& vL_in,
				qmVectorRef& uR_in,
				qmVectorRef& vR_in )
{
    pointCount = uL_in.getLength();
    uL = uL_in;
    vL = vL_in;
    uR = uR_in;
    vR = vR_in;
    beta = 1.;
    // fixedPointIndex = pointCount - 1;
    fixedPointCount = 1;
    fixedPointIndices = new int[1];
    fixedPointIndices[0] = 0;
    noErrorFunction::init( nMmts(), nStates(), nDStates() );
}

// This gives the error dy for the state vector x
// x --> dy:
// ( tx, ty, tzb, q0, q1, q2, q3, a_0, ... , a_N-1 ) -->
//     ( du_0, dv_0, ... , du_N-1, dv_N-1 )

const
qmVectorRef&
noErfRelativeOrientation::operator()( const qmVectorRef& x_in )
{
    qmVector x = x_in;
    // x ~ ( 7+n )
    // dy ~ ( 2n )
    real alpha;
    static qmVector p_object( 3 );
    static qmVector Rp( 3 );
    static qmVector p_camera_s( 3 ); // (X,Y,1+Zb)
    static qmVector uv( 2 );
    static qmMatrix R( 3, 3 );
    qmUnitQuaternionRef( x.getArray(Q0) ).getRotationMatrix( R );
    for( int i = 0; i < pointCount; i++ )
       {
	   alpha =  x(ALPHA+i);
	   p_object(0) = ( 1. + alpha * beta ) * uL(i);
	   p_object(1) = ( 1. + alpha * beta ) * vL(i);
	   p_object(2) = alpha;
	   Rp.setMultiply( R, p_object );
	   p_camera_s(0) = x( TX ) + Rp(0);
	   p_camera_s(1) = x( TY ) + Rp(1);
	   p_camera_s(2) = x( TZB ) + beta * Rp(2) + 1.;
	   dy_return( 2*i ) = uR(i) - p_camera_s(0) / p_camera_s(2);
	   dy_return( 2*i+1 ) = vR(i) - p_camera_s(1) / p_camera_s(2);
       }
    return dy_return;
}
 
static const int debug = 0;

// This gives the jacobian of the error dy with respect to the state
// step dx.
// x --> ddy / ddx:
// ( tx, ty, tzb, q0, q1, q2, q3, a_0, ... , a_N-1 ) -->
//
//	d( du_0, dv_0, ... , du_N-1, dv_N-1 )
//	-------------------------------------------------------------------
//	d( dtx, dty, dtz, dwx, dwy, dwz, da0,...,da_f-1,da_f+1,...,da_N-1 )
//
// a_f is "fixed", so dx is of dimension 6 + N-1 = 5 + N
// index f is given by fixedPointIndex

const
qmMatrixRef&
noErfRelativeOrientation::jacobian( const qmVectorRef& x_in )
{
    // x  ~ 7 + n
    // dy ~ 2n
    // dx ~ 3 + 3 + n-1 (which is 5+n)
    qmVector x = x_in;
    J_return.setZero(); // so that everything not explicitly set is 0
    real& tx = x( TX );
    real& ty = x( TY );
    real& tzb = x( TZB );
    real& q0 = x( Q0 );
    real& q1 = x( QX );
    real& q2 = x( QY );
    real& q3 = x( QZ );
    real* alpha = x.getArray( ALPHA );
    real q00 = q0 * q0;
    real q01 = q0 * q1;
    real q02 = q0 * q2;
    real q03 = q0 * q3;
    real q11 = q1 * q1;
    real q12 = q1 * q2;
    real q13 = q1 * q3;
    real q22 = q2 * q2;
    real q23 = q2 * q3;
    real q33 = q3 * q3;
    real c0 = 2*(q01 + q23);  
    real c1 = 2*(q01 - q23);
    real c2 = 2*(q02 + q13);  
    real c3 = 2*(q02 - q13);
    real c4 = 2*(q03 + q12);  
    real c5 = 2*(q03 - q12);
    real k0 = q00 + q11 - q22 - q33;
    real k1 = q00 - q11 + q22 - q33;
    real k2 = q00 - q11 - q22 + q33;
    static qmVector y( 2 );
    static qmVector p_obj( 3 );
    static qmVector Rp( 3 );
    static qmVector p_cam_s( 3 );
    static qmVector m( 3 );
    static qmMatrix dRdw( 3, 3 );
    static qmVector v( 3 );
    real m0, m1, m2;
    int alphaIndex = 0;
    for( int i = 0; i < pointCount; i++ ) 
       {
	   real a = alpha[ i ];
	   int u_row = 2*i;
	   int v_row = 2*i+1;
	   y(0) = uL( i );
	   y(1) = vL( i );
	   if(debug) y.print( cerr << "uv_L= " ) << endl;
	   p_obj(0) = (1. + a * beta) * y(0);
	   p_obj(1) = (1. + a * beta) * y(1);
	   p_obj(2) = a;
	   if(debug) p_obj.print( cerr << "p_obj= " ) << endl;
	   Rp(0) = k0*p_obj(0) - c5*p_obj(1) + c2*p_obj(2);
	   Rp(1) = c4*p_obj(0) + k1*p_obj(1) - c1*p_obj(2);
	   Rp(2) = (-c3)*p_obj(0) + c0*p_obj(1) + k2*p_obj(2);
	   if(debug) Rp.print( cerr << "Rp= " ) << endl;
	   p_cam_s(0) = Rp(0) + tx;
	   p_cam_s(1) = Rp(1) + ty;
	   p_cam_s(2) = beta * Rp(2) + tzb + 1.;
	   if(debug) p_cam_s.print( cerr << "p_cam_s= " ) << endl;
	   m0 = 1. / p_cam_s(2);
	   m1 = ( -m0 * m0 * p_cam_s(0) );
	   m2 = ( -m0 * m0 * p_cam_s(1) );
	   if(debug) cerr << "m0= " << m0 << endl;
	   if(debug) cerr << "m1= " << m1 << endl;
	   if(debug) cerr << "m2= " << m2 << endl;
	   J_return( u_row, DTX ) = m0;
	   J_return( u_row, DTY ) = 0.;
	   J_return( u_row, DTZB ) = m1;
	   J_return( v_row, DTX ) = 0.;
	   J_return( v_row, DTY ) = m0;
	   J_return( v_row, DTZB ) = m2;
	   dRdw( 0, 0 ) = p_obj(1) * c2 + p_obj(2) * c5;  
	   dRdw( 0, 1 ) = -p_obj(0) * c2 + p_obj(2) * k0; 
	   dRdw( 0, 2 ) = -p_obj(0) * c5 - p_obj(1) * k0;
	   dRdw( 1, 0 ) = -p_obj(1) * c1 - p_obj(2) * k1;
	   dRdw( 1, 1 ) = p_obj(0) * c1 + p_obj(2) * c4;
	   dRdw( 1, 2 ) = p_obj(0) * k1 - p_obj(1) * c4;
	   dRdw( 2, 0 ) = beta * ( p_obj(1) * k2 - p_obj(2) * c0 );
	   dRdw( 2, 1 ) = beta * ( -p_obj(0) * k2 - p_obj(2) * c3 );
	   dRdw( 2, 2 )   = beta * ( p_obj(0) * c0 + p_obj(1) * c3 );
	   J_return( u_row, DWX ) = m0 * dRdw( 0, 0 ) + m1 * dRdw( 2, 0 );
	   J_return( u_row, DWY ) = m0 * dRdw( 0, 1 ) + m1 * dRdw( 2, 1 );
	   J_return( u_row, DWZ ) = m0 * dRdw( 0, 2 ) + m1 * dRdw( 2, 2 );
	   J_return( v_row, DWX ) = m0 * dRdw( 1, 0 ) + m2 * dRdw( 2, 0 );
	   J_return( v_row, DWY ) = m0 * dRdw( 1, 1 ) + m2 * dRdw( 2, 1 );
	   J_return( v_row, DWZ ) = m0 * dRdw( 1, 2 ) + m2 * dRdw( 2, 2 );
	   v(0) = a * ( k0*y(0) - c5*y(1) );
	   v(1) = a * ( c4*y(0) + k1*y(1) );
	   v(2) = Rp(2) + a * beta * ( -c3*y(0) + c0*y(1) );
	   // J_return( u_row, DBETA ) = m0*v(0) + m1*v(2);
	   // J_return( v_row, DBETA ) = m0*v(1) + m2*v(2);
// 	   if ( i != fixedPointIndex )
// 	      {
// 		  v(0) = beta * y(0);
// 		  v(1) = beta * y(1);
// 		  v(2) = 1.;
// 		  if(debug) v.print( cerr << "v= " ) << endl;
// 		  Rp(0) = ( k0*v(0) - c5*v(1) + c2*v(2));
// 		  Rp(1) = ( c4*v(0) + k1*v(1) - c1*v(2));
// 		  Rp(2) = beta * (-c3*v(0) + c0*v(1) + k2*v(2));
// 		  if(debug) Rp.print( cerr << "Rp= " ) << endl;
// 		  J_return( u_row, DALPHA+alphaIndex ) = m0*Rp(0) + m1*Rp(2);
// 		  J_return( v_row, DALPHA+alphaIndex ) = m0*Rp(1) + m2*Rp(2);
// 		  alphaIndex++;
// 		  if(debug) cerr << endl;
// 	      }
	   if ( ! isFixedPoint(i) )
	      {
		  v(0) = beta * y(0);
		  v(1) = beta * y(1);
		  v(2) = 1.;
		  if(debug) v.print( cerr << "v= " ) << endl;
		  Rp(0) = ( k0*v(0) - c5*v(1) + c2*v(2));
		  Rp(1) = ( c4*v(0) + k1*v(1) - c1*v(2));
		  Rp(2) = beta * (-c3*v(0) + c0*v(1) + k2*v(2));
		  if(debug) Rp.print( cerr << "Rp= " ) << endl;
		  J_return( u_row, DALPHA+alphaIndex ) = m0*Rp(0) + m1*Rp(2);
		  J_return( v_row, DALPHA+alphaIndex ) = m0*Rp(1) + m2*Rp(2);
		  alphaIndex++;
		  if(debug) cerr << endl;
	      }
       }
    J_return.setNegateEq();
    return J_return;
}

static const real AngLim = .35; // 20 degrees

inline 
real
limitAngle( real angle )
{
  return limit( angle, (real)(-AngLim), (real)(AngLim) );
}

static 
qmVectorRef&
updateQuaternion( qmUnitQuaternionRef& qin, 
		  qmVectorRef& w,
		  qmUnitQuaternionRef& qout )
{
    qmUnitQuaternion dq;
    dq( 1 ) = limitAngle( w(0) ) / 2.;
    dq( 2 ) = limitAngle( w(1) ) / 2.;
    dq( 3 ) = limitAngle( w(2) ) / 2.;
    dq( 0 ) = 1. - dq(1)*dq(1) - dq(2)*dq(2) - dq(3)*dq(3);
    if( dq(0) < 0)
       cerr << "updateQuaternion: big Euler angles" << endl;
    dq(0) = sqrt( fabs( (double) dq(0) ) );
    qout(0) = qin(0)*dq(0) - qin(1)*dq(1) - qin(2)*dq(2) - qin(3)*dq(3);
    qout(1) = qin(1)*dq(0) + qin(0)*dq(1) - qin(3)*dq(2) + qin(2)*dq(3);
    qout(2) = qin(2)*dq(0) + qin(3)*dq(1) + qin(0)*dq(2) - qin(1)*dq(3);
    qout(3) = qin(3)*dq(0) - qin(2)*dq(1) + qin(1)*dq(2) + qin(0)*dq(3);
    real mag = sqrt( qout(0)*qout(0) + qout(1)*qout(1) + 
		     qout(2)*qout(2) + qout(3)*qout(3) );
    qout /= mag;
    return qout;
}

const
qmVectorRef&
noErfRelativeOrientation::add( const qmVectorRef& x_in,
			       const qmVectorRef& dx_in )
{
    qmVector x = x_in;
    qmVector dx = dx_in;
    x_return( TX ) = x( TX ) + dx( DTX );
    x_return( TY ) = x( TY ) + dx( DTY );
    x_return( TZB ) = x( TZB ) + dx( DTZB );
    qmUnitQuaternionRef qin( x.getArray(Q0) );
    qmUnitQuaternionRef qout( x_return.getArray(Q0) );
    qmVectorRef w( 3, dx.getArray(DWX) );
    updateQuaternion( qin, w, qout );
    int daIndex = 0;
    for( int i = 0; i < pointCount; i++ )
//       if( i != fixedPointIndex )
       if( ! isFixedPoint( i ) )
	  x_return( ALPHA+i ) = x( ALPHA+i ) + dx( DALPHA+daIndex++ );
       else
	  {
	      x_return( ALPHA+i ) = x( ALPHA+i );
	  }
    return x_return;
}

void
noErfRelativeOrientation::setFixedPoint( int k )
{
    setFixedPoints( 1, &k );
}

void
noErfRelativeOrientation::setFixedPoints( int n, int* klist )
{
    if( n <= 0 )
       {
	   cerr << "noErfRelativeOrientation::setFixedPoints: ERROR: " 
		<< "attempt to fix " << n << " points\n\tmust be > 0" << endl;
	   return;
       }
    if( fixedPointCount )
       {
	   delete fixedPointIndices;
	   // delete fixedPointDepths;
       }
    fixedPointCount = n;
    fixedPointIndices = new int[ n ];
    for( int i = 0; i < n; i++ )
       fixedPointIndices[i] = klist[i];
    noErrorFunction::init( nMmts(), nStates(), nDStates() );
}
