// cgTriangle.h++

#ifndef _cgTriangle_H_
#define _cgTriangle_H_

#include <stream.h>
#include <fstream.h>
#include "QuickMath/QuickMath.h++"
#include "QuickMath/qmMinMax.h++"

// three coord frames world (w), left (l), right (r)
// P_x = Txp_x		( point "p" in "x" frame is vector from "x" origin
// 			  to "p" represented in "x" basis )
// P_r  =  Trl_r  +  Rrl . P_l	( Trp = Trl + Tlp )
// P_w  =  Twl_w  +  Rwl .

class cgTriangle
{
protected:

    qmVector T;
    qmMatrix R;
    real bL;
    real bR;
    int state;
    qmUnitQuaternion q;

public:

    cgTriangle();
    cgTriangle( char *filename );
    istream& scan( istream& );
    int FScan( char * );
    real Point( real uL, real vL, real uR, real vR, qmVectorRef& p,
		real* residual = 0 );
    real PointAlt( real uL, real vL, real uR, real vR, qmVectorRef& p );
    // real FakePoint( real uL, real vL, real uR, real vR, qmVectorRef& p );
    operator int() { return state; }
    qmVectorRef& Translation() { return T; }
    qmMatrixRef& Rotation() { return R; }
    qmUnitQuaternion& Q() { return q; }
    real& BL() { return bL; }
    real& BR() { return bR; }
    istream& scanEstRecord( istream& );
    void ellipse( qmVectorRef& mL, qmMatrixRef& CL,
		  qmVectorRef& mR, qmMatrixRef& CR,
		  qmVectorRef& m3out, qmMatrixRef& C3out );
    void ellipse( qmVectorRef& mL, qmMatrixRef& CL,
		  qmVectorRef& mR, qmMatrixRef& CR,
		  qmVectorRef& m3out, qmMatrixRef& C3out,
		  qmMatrixRef& Ut, qmVectorRef& s, qmMatrixRef& Vt );
    static real defaultSV;
    static int svinvert;
    ostream& print( ostream& );

private:
    void projectionJacobian( qmMatrixRef&, qmVectorRef&, real );
    void combineJacobians( qmMatrixRef&, qmMatrixRef&, qmMatrixRef& );
    void compute3DCovariance( qmMatrixRef&, qmMatrixRef&, qmVectorRef& );
    void compute3DCovariance( qmMatrixRef&, qmMatrixRef&, qmVectorRef&,
			      qmMatrixRef&, qmVectorRef&, qmMatrixRef& );
};

#endif /* _cgTriangle_H_ */

