// qmKalmanFilter.h++
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
// at the Media Laboratory, MIT, Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.
// 

#ifndef _qmKalmanFilter_H_
#define _qmKalmanFilter_H_

// This is a simple linear Kalman filter.
// Unlike my Ekf class, it is not resizeable, so it is meant
// for simple linear problems.

#include <stream.h>
#include "QuickMath/QuickMath.h++"

class qmKalmanFilter
{
  static const float veryBig;
protected:
  int n;		// dimension of state
  int p;		// dimension of mmt
  int t;		// prediction time (1 + estimate time)
  qmVector xP;
  qmVector xE;
  qmVector dX;	
  qmVector yP;
  qmVector dY;
  qmMatrix Pp;
  qmMatrix Pe;
  qmMatrix K;
  qmMatrix I;
  qmMatrix tempNxP;
  qmMatrix tempNxN;
  qmMatrix tempPxP;
public:
  // Leave these public so the user can initialize or change them
  qmMatrix H;			// measurement model
  qmMatrix R;			// measurement noise
  qmMatrix Phi;			// dynamic model
  qmMatrix Q;			// dynamic noise

public:
  qmKalmanFilter( int nstates,      // number of state dimensions
		  int nmmts,        // number of measurement dimensions
		  qmVectorRef& xp,  // initial state 
		  qmMatrixRef& pp,  // initial covariance
		  qmMatrixRef& h,   // measurement model  y=h*x+r*w
		  qmMatrixRef& r,   // measurement noise 
		  qmMatrixRef& ph,  // dynamic model     x'=ph*x+q*v
		  qmMatrixRef& q ); // dynamic noise 

  qmKalmanFilter( int ns,           // number of state dimensions
		  float rv = 1.,    // measurement noise var diagonal
		  float qv = 1.,    // dynamic noise var diagonal
		  float dt = 1. );  // time step for Phi 

  qmKalmanFilter( const qmKalmanFilter &qmKF); //copy constructor

  // update the state of the filter with a new measurement
  qmKalmanFilter& update( const qmVectorRef& y );
  qmKalmanFilter& update( real* y ) { return update(qmVectorRef(p,y)); }

  void predict( qmVectorRef& x , int steps ) const;

  int getPredictionTime() { return t; }
  int getEstimateTime() { return( t-1 ); }

  // probe the state of the filter.  
  // These are refs in the filter - don't change them lightly!
  qmVectorRef& getStatePrediction() { return xP; }
  qmVectorRef& getStateEstimate() { return xE; }
  qmVectorRef& getInnovation() { return dY; }
  qmVectorRef& getStateInnovation() { return dX; }
  qmMatrixRef& getErrorCovarEstimate() { return Pe; }
  qmMatrixRef& getErrorCovarPrediction() { return Pp; }
  
};

#endif /* _qmKalmanFilter_H_ */
