// qmVtiKalmanFilter.h++
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
// at the Media Laboratory, MIT, Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.
// 

// variable time-interval Kalman filter
//
// Allows for measurements to be taken at a non-constant time interval
// Allows for a finer-grain time scale for state prediction

#ifndef _qmVtiKalmanFilter_H_
#define _qmVtiKalmanFilter_H_

#include "qmKalmanFilter/qmKalmanFilter.h++"
#include "QuickMath/qmTime.h++"

class qmVtiKalmanFilter : public qmKalmanFilter
{
private:
protected:
  qmTime tP;
  qmTime tE;
  qmVector xBase;
  qmMatrix PBase;
  qmMatrix dQ;
public:
  // need a more general constructor; for now just do the slick one
  qmVtiKalmanFilter( int ns, 
		     real rv = 1., 
		     real qv = 1., 
		     double to = -1. );

  qmVtiKalmanFilter& update( const qmVectorRef& y, 
			     qmTime& t, 
			     int N );

  inline qmVtiKalmanFilter& update( real* fl, 
				    qmTime& ty, 
				    int N ) 
  {
    return update( qmVectorRef(p,fl), ty, N ); 
  };

  virtual qmMatrixRef& StateTransitionMatrix( qmTime& t, 
					      qmTime& dt );
};

#endif /* _qmVtiKalmanFilter_H_ */
