#include <qmKalmanFilter.h++>

main(int argc, char *argv[])
{
  int nstate = 6;		// x, y, x', y', x'', y''
  int nmmt   = 2;		// x, y
  qmVector x(nstate);		// initial state 
  qmMatrix p(nstate,nstate);    // initial covariance
  qmMatrix h(nmmt,nstate);	// measurement model  y=h*x+r*w
  qmMatrix r(nmmt,nmmt);	// measurement noise 
  qmMatrix ph(nstate,nstate);	// dynamic model     x'=ph*x+q*v
  qmMatrix q(nstate,nstate);	// dynamic noise 

  x.setZero();			// zero initial state
  p.setIdentity();		// identity covarince
  h.setZero();			// measure position only
  h(0,0) = 1.0;			
  h(1,1) = 1.0;
  r.setIdentity();		// identity covar mmt noise
  ph.setIdentity();		// set the diagonal of the dyn model
  ph(0,2) = 1.0;		//   x   = x + dx
  ph(1,3) = 1.0;		//   y   = y + dy
  ph(2,4) = 1.0;		//   x'  = x' + x''
  ph(3,5) = 1.0;		//   y'  = y' + y''
				//   x'' = x''    constant accel
				//   y'' = y''

  q.setIdentity();		// identity model noise
  
  qmKalmanFilter filter(nstate,		// number of state dimensions
			nmmt,		// number of measurement dimensions
			x, p, h, r, ph, q );		
  
  double dt = 0.1;
  for (double u=0.0; u<10.0; u+=dt)
    {
      qmVector Y(nmmt);
      Y(0)=sin(u);		// circular motion
      Y(1)=cos(u);
      
      filter.update(Y);
      
      cout << Y << " ";		                                // posn
      cout << dt * cos(u) << " " << dt * -sin(u) << " ";        // velocity
      cout << dt*dt * -sin(u) << " " << dt*dt * -cos(u) << " ";	// accel
      cout << filter.getStateEstimate() << " ";
      cout << filter.getStatePrediction() << " ";
      cout << filter.getInnovation() << endl;
    }
}

