// ComputeGeometry.c++

#include "ComputeGeometry.h++"
real ComputeGeometry::ballSize[3] = { .1, .05, .05 };
real ComputeGeometry::lmAbsTol = .1;
int ComputeGeometry::lmIterLimit = 100;
real ComputeGeometry::yakkaT = 6;	// inches
real ComputeGeometry::yakkaW = 10;	// degrees
real ComputeGeometry::yakkaL = 0;	// inches
real ComputeGeometry::yakkaDT = 2;	// inches
real ComputeGeometry::yakkaDW = 2;	// degrees
real ComputeGeometry::yakkaDL = 0;	// inches
real ComputeGeometry::yakkaDLs = 1.0;	// ratio
real ComputeGeometry::yakkaUV = .01;	// units
real ComputeGeometry::yakkaS = .0001;	// units
int ComputeGeometry::wakkaFooMode = 1;
real ComputeGeometry::wakkaFoo[3] = { .03, .03, .06 };
real ComputeGeometry::wakkaHoo[3] = { .03, .03, .06 };
int ComputeGeometry::kalmanMode = 0;
real ComputeGeometry::mmtNoiseVar = 1;
real ComputeGeometry::dynNoiseVar = 1;
int ComputeGeometry::vtiIter = 3;
float ComputeGeometry::calibTol = .001;
float ComputeGeometry::calibAbstol = .001;
ComputeGeometry::ComputeGeometry()
  : personMode( 1 ),
    Rwc( 3, 3 ),
    Rwc_inv( 3, 3 ),
    Twc( 3 ),
    scaleToWorld( real(1.) ),
    pct( 3 )
{
    for( int i = 0; i < 3; i++ )
       {
	 mode[i] = NOMODE;
	 prevMode[i] = NOMODE;
	 yakkaX[i] = 0;
	 blobX[i] = 0;
       }
    Rwc.setIdentity();
    Rwc_inv.setIdentity();
    Twc.setZero();
}
void
ComputeGeometry::initNoMode(int i)
{
    mode[i] = NOMODE;
}
void
ComputeGeometry::initTriangle(int i)
{
    mode[i] = TRIANGLE;
}
void
ComputeGeometry::initBlobs(int i)
{
    mode[i] = BLOB;
    if( !blobX[i] )
      blobX[i] = new qmVector( 10 );
}
void
ComputeGeometry::initYakka(int i)
{
    mode[i] = YAKKA;
    if( !yakkaX[i] )
      yakkaX[i] = new qmVector( 10 );
}
void
ComputeGeometry::computeTriangle( int i,
				  qmVectorRef* y, 
				  int* val,
				  qmVectorRef* t, 
				  qmUnitQuaternionRef* q,
				  qmVectorRef* l )
{
  if( val[i] )
    triangle.point( y[i].subVector( 0, 2 ),
		    y[i].subVector( 5, 2 ),
		    t[i] );
  else
    t[i].setZero();
  q[i].setIdentity();
  l[i].setConst( ballSize[i] );
}
void
ComputeGeometry::initCalibration( real* t, real* q, real b )
{
    float* p = calibData;
    for( int i = 0; i < 3; i++ )
       *p++ = t[i];
    for( i = 0; i < 4; i++ )
       *p++ = q[i];
    *p = b;
    triangle.setRightCoordinateTransform
       ( t[0], t[1], t[2], q[0], q[1], q[2], q[3] );
    triangle.setRightBeta( b );
    triangle.setLeftBeta( b );
}
void
ComputeGeometry::computeBlob( int i,
			      qmVectorRef* y, 
			      int* val,
			      qmVectorRef* t, 
			      qmUnitQuaternionRef* q,
			      qmVectorRef* l )
{
    static noErfBlob erfBlob;
    static noLevenbergMarquardt lmBlob;
    static int inited = 0;
    const real root5 = sqrt( 5. );
    if( !inited )
      {
	lmBlob.init( erfBlob );
	inited = 1;
      }
    if( !val[i] )
      {
	t[i].setZero();
	q[i].setIdentity();
	l[i].setZero();
	return;
      }
    erfBlob.init( y[i],
		  triangle.BL(), 
		  triangle.Translation(),
		  triangle.Q() );
    erfBlob.computeLinearInit( *blobX[i] );
    lmBlob.setState( *blobX[i] );
    lmBlob.setAbsoluteTolerance( lmAbsTol );
    lmBlob.setIterationLimit( lmIterLimit );

    *blobX[i] = lmBlob.estimate();

    t[i].copy( (*blobX[i]).subVector( noErfBlob::TX, 3 ) );
    q[i].copy( (*blobX[i]).subVector( noErfBlob::Q0, 4 ) );
    l[i].copy( (*blobX[i]).subVector( noErfBlob::LX, 3 ) );
}
void
ComputeGeometry::yakkaNormalizeBlobs( int which, qmVector** b )
{
    static qmVectorRef dim;
    static int maxdim;
    static real maxval;
    static int mindim;
    static real minval;
    dim.set( b[HEAD] -> subVector( noErfBlob::LX, 3 ) );
    maxval = dim.getMax( &maxdim );
    minval = dim.getMin( &mindim );
    if (which == HEAD)
      for( int i = 0; i < 3; i++ )
	if( i != maxdim && i != mindim )
	  dim.setConst( dim(i) );

    qmVectorRef dimL = b[LEFTHAND] -> subVector( noErfBlob::LX, 3 );
    qmVectorRef dimR = b[RIGHTHAND] -> subVector( noErfBlob::LX, 3 );
    int maxL;
    int maxR;
    real maxleft = dimL.getMax( &maxL );
    real maxrite = dimR.getMax( &maxR );
    maxval = .5 * ( maxleft + maxrite );
    real minLeft = 0.;
    real minRite = 0.;
    for( int i = 0; i < 3; i++ )
       {
	   if( i != maxL )
	      minLeft += dimL(i);
	   if( i != maxR )
	      minRite += dimR(i);
       }
    minval = .25 * ( minLeft + minRite );
    for( i = 0; i < 3; i++ )
       {
	 if (which == LEFTHAND)
	   dimL(i) = ( i == maxL ) ? maxval : minval;
	 if (which == RIGHTHAND)
	   dimR(i) = ( i == maxR ) ? maxval : minval;
       }
}
void
ComputeGeometry::computeYakkaYakka( int which,
				    qmVectorRef* y, 
				    int* val,
				    qmVectorRef* t, 
				    qmUnitQuaternionRef* q,
				    qmVectorRef* l )
{
    static noErfBlob erfBlob;
    static noYakkaYakka yakkas[ 3 ];
    static int yakkaInit[ 3 ];
    const real root5 = sqrt( 5. );
    if( BLOB == prevMode[which] )
       yakkaNormalizeBlobs( which , blobX );

    if( !val[which] )
      {
	t[which].setZero();
	q[which].setIdentity();
	l[which].setZero();
	return;
      }
    qmVectorRef& xblob = (*yakkaX[which]);
    erfBlob.init( y[ which ], 
		  triangle.BL(), 
		  triangle.Translation(),
		  triangle.Q() );
    if( ! yakkaInit[ which ] || NOMODE == prevMode[which] )
      {
	static qmMatrix P( 9, 9 );
	yakkas[ which ].init( erfBlob );
	erfBlob.computeLinearInit( *yakkaX[which] );
	if( wakkaFooMode )
	  {
	    if( HEAD == which )
	      xblob.subVector( noErfBlob::LX, 3 ).setData
		( wakkaHoo );
	    else
	      xblob.subVector( noErfBlob::LX, 3 ).setData
		( wakkaFoo );
	    
	  }
	yakkas[ which ].setState( xblob );
	P.setZero();
	real s2 = scaleToWorld*scaleToWorld;
	real dtor = M_PI / 180.;
	for( int i = 0; i < 3; i++ )
	  P( i, i ) = yakkaT*yakkaT / s2;
	for( ; i < 6; i++ )
	  P( i, i ) = yakkaW*yakkaW*dtor*dtor;
	for( ; i < 9; i++ )
	  P( i, i ) = yakkaL*yakkaL / s2;
	yakkas[ which ].setCovariance( P );
	yakkas[ which ].Q.setZero();
	for( i = 0; i < 3; i++ )
	  yakkas[ which ].Q( i, i ) = yakkaDT*yakkaDT/s2;
	for( ; i < 6; i++ )
	  yakkas[ which ].Q( i, i ) = yakkaDW*yakkaDW/s2;

	yakkas[ which ].Q( i, i ) =          yakkaDL*yakkaDL/s2;  i++;
	yakkas[ which ].Q( i, i ) = yakkaDLs*yakkaDL*yakkaDL/s2;  i++;
	yakkas[ which ].Q( i, i ) = yakkaDLs*yakkaDL*yakkaDL/s2;  i++;

	yakkas[ which ].R.setZero();
	for( i = 0; i < 2; i++ )
	  yakkas[ which ].R( i, i ) = yakkaUV*yakkaUV;
	for( ; i < 5; i++ )
	  yakkas[ which ].R( i, i ) = yakkaS*yakkaS;
	for( i = 0; i < 7; i++ )
	  yakkas[ which ].R( i, i ) = yakkaUV*yakkaUV;
	for( ; i < 10; i++ )
	  yakkas[ which ].R( i, i ) = yakkaS*yakkaS;
	yakkaInit[ which ] = 1;
      }
    else if( BLOB == prevMode[which] )
      {
	yakkas[ which ].setState( *blobX[which] );
      }
    xblob = yakkas[ which ].estimate();
    t[ which ].copy( xblob.subVector( noErfBlob::TX, 3 ) );
    q[ which ].copy( xblob.subVector( noErfBlob::Q0, 4 ) );
    l[ which ].copy( xblob.subVector( noErfBlob::LX, 3 ) );
}

void
ComputeGeometry::computeNotReally( int i,
				   qmVectorRef* , 
				   int* ,
				   qmVectorRef* t, 
				   qmUnitQuaternionRef* q,
				   qmVectorRef* l )
{
    const real dfltPos[3] = { 0, .6, 0 };
    const real dfltSiz[3] = { .1, .05, .05 };

    t[i].setData( dfltPos );
    q[i].setIdentity();
    l[i].setData( dfltSiz );
}

void
ComputeGeometry::setMode( int i, int m )
{
    switch( m )
       {
	case TRIANGLE:
	   initTriangle(i);
	   break;
	case BLOB:
	   initBlobs(i);
	   break;
	case YAKKA:
	   initYakka(i);
	   break;
	default:
	   initNoMode(i);
       }
}
inline
void
ComputeGeometry::cameraToWorld( const qmVectorRef& pcv, qmVectorRef& pw )
{
    pct.setSubtract( pcv, Twc );
    pw.setMultiply( Rwc, pct );
    pw.setScaleEq( scaleToWorld );
}
inline
void
ComputeGeometry::worldToCamera( const qmVectorRef& pw, qmVectorRef& pc )
{
    pct.setScale( 1./scaleToWorld, pw );
    pc.setMultiply( Rwc_inv, pct );
    pc += Twc;
}
void
ComputeGeometry::convertToWorld( int* val,
				 qmVectorRef* t,
				 qmUnitQuaternionRef* q,
				 qmVectorRef* l,
				 qmVectorRef* tw,
				 qmUnitQuaternionRef* qw,
				 qmVectorRef* lw )
{
    static qmMatrix Rc( 3, 3 );
    static qmMatrix Rw( 3, 3 );
    for( int i = 0; i < 3; i++ )
       {
	   if( !val[i] )
	      continue;
	   cameraToWorld( t[i], tw[i] );
	   lw[i].setScale( scaleToWorld, l[i] );
	   q[i].getRotationMatrix( Rc );
	   Rw.setMultiply( Rwc, Rc );
	   qw[i].set( Rw );
       }
}
void
ComputeGeometry::setWorldTransform( qmVectorRef& t, qmMatrixRef& r, real s )
{
    Twc.copy( t );
    Rwc.copy( r );
    Rwc_inv.setInvert( r );
    scaleToWorld = s; 
}
ostream&
ComputeGeometry::alert( char* fn, char* msg )
{
    return( cerr << "ComputeGeometry: ALERT: " << fn << "(): " << msg << endl );
}

void
ComputeGeometry::error( char* fn, char* msg )
{
    cerr << "ComputeGeometry: ERROR: " << fn << "(): " << msg << endl;
    assert(0);
}
void
ComputeGeometry::calibrate( int ncorr, float* data )
{
    noErfRelativeOrientation erf;
    qmVector uL( ncorr );
    qmVector vL( ncorr );
    qmVector uR( ncorr );
    qmVector vR( ncorr );
    for( int i = 0; i < ncorr; i++ )
       {
	   uL(i) = data[0];
	   vL(i) = data[1];
	   uR(i) = data[2];
	   vR(i) = data[3];
	   data += 4;
       }
    qmVector actualState( 7 + ncorr );
    qmVectorRef trel( 3, actualState.getArray() );
    qmUnitQuaternionRef qrel
       ( actualState.getArray( noErfRelativeOrientation::Q0 ) );
    qmVectorRef alpha
       ( ncorr, actualState.getArray( noErfRelativeOrientation::ALPHA ) );
    qmVector state;
    state.set( 7 + ncorr );
    xinit[0] = 0;
    xinit[1] = 0;
    xinit[2] = 0;
    xinit[3] = 1;
    xinit[4] = 0;
    xinit[5] = 0;
    xinit[6] = 0;
    xinit[7] = 0;
    float fixAlpha = 2;
    state( noErfRelativeOrientation::TX ) = xinit[0];
    state( noErfRelativeOrientation::TY ) = xinit[1];
    state( noErfRelativeOrientation::TZB ) = xinit[2];
    state( noErfRelativeOrientation::Q0 ) = xinit[3];
    state( noErfRelativeOrientation::QX ) = xinit[4];
    state( noErfRelativeOrientation::QY ) = xinit[5];
    state( noErfRelativeOrientation::QZ ) = xinit[6];
    for( i = 0; i < ncorr; i++ )
      state( noErfRelativeOrientation::ALPHA + i ) = fixAlpha;
    noLevenbergMarquardt estimator;
    erf.init( uL, vL, uR, vR );
    real beta = 1;
    erf.setBeta( beta );
    estimator.init( erf );
    estimator.setGradientTolerance( calibTol );
    estimator.setAbsoluteTolerance( calibAbstol );
    estimator.setIterationLimit( 30 );
    estimator.setState( state );
    state = estimator.estimate();
    cerr << state << endl;
    initCalibration( state.getArray(), state.getArray( 3 ), beta );
}

void
ComputeGeometry::compute2DQuickPoints( qmVectorRef& mean, qmMatrixRef& cov,
				       qmVectorRef& qpNeg, qmVectorRef& qpPos )
{
    static qmEigenStructure eig;
    static qmMatrix evec( 2, 2 );
    static qmVector eval( 2 );
    if( qmEigenStructure::eigenSymmetric2( cov, eval, evec ) )
       alert( "compute2DQuickPoints", "eigen2" );
    int imax;
    int imin;
    real lmax = sqrt( eval.getMax( &imax ) );
    real lmin = sqrt( eval.getMin( &imin ) );
    real e = sqrt(eval(1-imax)) / sqrt(eval(imax));
    real alpha = ( e < .5 ) ? 1 : ( 1-2*(e-.5) );
    alpha *= 2*lmax;
    qpNeg.copy( mean ).setScaleAddEq( -alpha, evec.getRow(imax) );
    qpPos.copy( mean ).setScaleAddEq( alpha, evec.getRow(imax) );
}

void
ComputeGeometry::computeQuickPoint( real* yL, real* yR, 
				    real* qpNeg, real* qpPos )
{
    static qmVector qLNeg( 2 );
    static qmVector qLPos( 2 );
    static qmVector qRNeg( 2 );
    static qmVector qRPos( 2 );
    static qmMatrix cov( 2, 2 );
    cov(0,0) = yL[2];
    cov(0,1) = cov(1,0) = yL[4];
    cov(1,1) = yL[3];
    compute2DQuickPoints( qmVectorRef(2,yL), cov, qLNeg, qLPos );
    cov(0,0) = yR[2];
    cov(0,1) = cov(1,0) = yR[4];
    cov(1,1) = yR[3];
    compute2DQuickPoints( qmVectorRef(2,yR), cov, qRNeg, qRPos );
    real res_same_0;
    real res_same_1;
    real res_diff_0;
    real res_diff_1;
    static qmVector qpNP( 3 );
    static qmVector qpPN( 3 );
    triangle.point( qLNeg, qRNeg, qmVectorRef( 3, qpNeg ), &res_same_0 );
    triangle.point( qLPos, qRPos, qmVectorRef( 3, qpPos ), &res_same_1 );
    triangle.point( qLNeg, qRPos, qpNP, &res_diff_0 );
    triangle.point( qLPos, qRNeg, qpPN, &res_diff_1 );
    if( res_diff_0 + res_diff_1 < res_diff_0 + res_diff_1 )
       {
	   qmVectorRef( 3, qpNeg ).copy( qpNP );
	   qmVectorRef( 3, qpPos ).copy( qpPN );
       }
}

int
ComputeGeometry::computeQuickPoints( qmVectorRef* y, int* val, real* qp )
{
    for( int i = 0; i < 3; i++ )
       { 
	   if( !val[i] )
	      continue;
	   computeQuickPoint( y[i].getArray(0), y[i].getArray(5), 
			      qp+6*i, qp+6*i+3 );
       }
    return 0;
}
