// ComputeGeometry.h++

#ifndef _ComputeGeometry_H_
#define _ComputeGeometry_H_

#define QmDouble
#include "QuickMath/QuickMath.h++"
#include "noLevenbergMarquardt.h++"
#include "noYakkaYakka.h++"
#include "noErfBlob.h++"
#include "noErfRelativeOrientation.h++"
#include <sys/time.h>
#include "qmComputeGeometry/cgTriangleStereo.h++"

class ComputeGeometry
{
    int mode[3];
    int prevMode[3];
    int personMode;
    cgTriangleStereo triangle;
    static real ballSize[3];
    void computeTriangle( int which,
			  qmVectorRef* y, 
			  int* val,
			  qmVectorRef* t, 
			  qmUnitQuaternionRef* q,
			  qmVectorRef* l );
    void computeBlob( int which,
		      qmVectorRef* y, 
		      int* val,
		      qmVectorRef* t, 
		      qmUnitQuaternionRef* q,
		      qmVectorRef* l );
    void computeNotReally( int which,
			   qmVectorRef* y, 
			   int* val,
			   qmVectorRef* t, 
			   qmUnitQuaternionRef* q,
			   qmVectorRef* l );
    qmVector* yakkaX[3];
    qmVector* blobX[3];
    void computeYakkaYakka( int which,
			    qmVectorRef* y, 
			    int* val,
			    qmVectorRef* t, 
			    qmUnitQuaternionRef* q,
			    qmVectorRef* l );
    qmMatrix Rwc;
    qmMatrix Rwc_inv;
    qmVector Twc;
    real scaleToWorld;
    qmVector pct;
    inline void cameraToWorld( const qmVectorRef& pcv, qmVectorRef& pw );
    inline void worldToCamera( const qmVectorRef& pw, qmVectorRef& pc );
    static ostream& alert( char* fn, char* msg );
    static void error( char* fn, char* msg );
    float calibData[8];
    float xinit[8];
    void compute2DQuickPoints( qmVectorRef&, qmMatrixRef&,
			       qmVectorRef&, qmVectorRef& );
    void computeQuickPoint( real*, real*, real*, real* );
    void yakkaNormalizeBlobs( int which, qmVector** blobs );
 public:
    static float calibTol;
    static float calibAbstol;
    static int kalmanMode;
    static real mmtNoiseVar;
    static real dynNoiseVar;
    static int vtiIter;
    enum { NOMODE=0, TRIANGLE, BLOB, YAKKA };
    enum { HEAD=0, LEFTHAND=1, RIGHTHAND=2 };
    ComputeGeometry();
    void initCalibration( real* t, real* q, real b );
    void initNoMode(int);
    void initTriangle(int);
    void initBlobs(int);
    void initYakka(int);
    inline void compute( qmVectorRef* y,
			 int* val,
			 qmVectorRef* t, 
			 qmUnitQuaternionRef* q,
			 qmVectorRef* l,
			 qmVectorRef* tw, 
			 qmUnitQuaternionRef* qw,
			 qmVectorRef* lw );
    static real yakkaDL;
    static real yakkaDLs;
    static real yakkaUV;
    static real lmAbsTol;
    static int lmIterLimit;
    static real yakkaT;
    static real yakkaW;
    static real yakkaL;
    static real yakkaS;
    static int wakkaFooMode;
    static real wakkaFoo[3];
    static real wakkaHoo[3];
    void setMode( int, int );
    void setMode( int mode ) { for (int i=0; i<3; i++ ) setMode(i,mode); };

    void convertToWorld( int* val,
			 qmVectorRef* t,
			 qmUnitQuaternionRef* q,
			 qmVectorRef* l,
			 qmVectorRef* tw,
			 qmUnitQuaternionRef* qw,
			 qmVectorRef* lw );
    static real yakkaDT;
    static real yakkaDW;
    void setWorldTransform( qmVectorRef& t, qmMatrixRef& r, real s );
    inline float* getCalibration() { return calibData; }
    void calibrate( int, float* );
    int computeQuickPoints( qmVectorRef* y, int* val, real* );
};
inline
void
ComputeGeometry::compute( qmVectorRef* y, 
			  int* val,
			  qmVectorRef* t, 
			  qmUnitQuaternionRef* q,
			  qmVectorRef* l,
			  qmVectorRef* tw, 
			  qmUnitQuaternionRef* qw,
			  qmVectorRef* lw )
{
  for( int i = 0; i < 3; i++ )
    {
      if( TRIANGLE == mode[i] )
	computeTriangle(i, y, val, t, q, l );
      else if( BLOB == mode[i] )
	computeBlob(i, y, val, t, q, l );
      else if( YAKKA == mode[i] )
	computeYakkaYakka(i, y, val, t, q, l );
      else
	computeNotReally(i, y, val, t, q, l );
      prevMode[i] = mode[i];
    }

  convertToWorld( val, t, q, l, tw, qw, lw );
}


#endif /* _ComputeGeometry_H_ */
