// FfinderClient.c++

#include "FfinderClient.h++"

FfinderClient::FfinderClient()
  : leftClient( 0 ),
    rightClient( 0 ),
    leftServerHost( "eastfinchley" ),
    rightServerHost( "westfinchley" ),
    currentFrameNumber( -1 )
{}

int
FfinderClient::connect( char* lhost, char* rhost )
{
  /* STIVE-PC has the convention that lhost and rhost are the same	*/
  /* and that the 2nd camera's program_number is incremented by 1.	*/
  /* Note: Ali & Egon have opposite conventions for naming left & right */
  /* cameras.  Ali names from camera point of view (e.g. left eye, right*/
  /* eye) while Egon names as you face the cameras.  To be consistent	*/
  /* with all this, and to accept input from STIVE-PC as well as SGI, we*/
  /* do the junk with program_number ++; program_number --;.		*/
  u_long program_number = FFINDERPROG;
  leftServerHost = lhost;
  rightServerHost = rhost;
  const int tryMax = 20;

  if (strcmp(lhost,rhost) == 0)
    program_number ++;

  for( int tryCount = 0; tryCount < tryMax &&
	 !( leftClient = clnt_create
	    ( leftServerHost, FFINDERPROG, FFINDERVERS, "tcp")); 
       tryCount++ )
    clnt_pcreateerror( leftServerHost ), sginap( 100 );
  if( tryCount >= tryMax )
    return 1;

  if (strcmp(lhost,rhost) == 0)
    program_number --;
  
  for( tryCount = 0; tryCount < tryMax &&
	 !( rightClient = clnt_create
	    ( rightServerHost, program_number, FFINDERVERS, "tcp")); 
       tryCount++ )
    clnt_pcreateerror( rightServerHost ), sginap( 100 );
  if( tryCount >= tryMax )
    return 1;
  // initClient( leftConfig, leftClient, leftServerHost );
  // initClient( rightConfig, rightClient, rightServerHost );
  return 0;
}

// static
// inline
// void
// initClient( ffConfig& data, CLIENT *c, char *s = "" )
// {
//   void *result;
//   if( c )
//      if ( !( result = ffinit_1( &data, c ) ) )
// 	clnt_perror( c, s ), exit(1);
// }

static
inline
void
frame( CLIENT *c, char *s = "" )
{
  void *result;
  if( c )
     if ( !( result = ffframe_1( (void *)0, c ) ) )
	clnt_perror( c, s ), exit(1);
}

static
inline
void
get_frame_result( ffPerson& data, CLIENT *c, char *s = "" )
{
  ffPerson *result = 0;
  if( c )
     {
       if ( !( result = ffget_frame_result_1( (void *)0, c ) ) )
	  clnt_perror( c, s ), exit(1);
       data = *result;
     }
}

static
inline
void
get_current( ffPerson& data, CLIENT *c, char *s = "" )
{
  ffPerson *result = 0;
  if( c )
     {
       if ( !( result = ffget_current_1( (void *)0, c ) ) )
	  clnt_perror( c, s ), exit(1);
       data = *result;
   }
}

void
FfinderClient::callCaptureFrames()
{
    frame( leftClient, leftServerHost );
    frame( rightClient, rightServerHost );
}

void
FfinderClient::callGetResults()
{
    get_frame_result( leftPerson, leftClient, leftServerHost );
    get_frame_result( rightPerson, rightClient, rightServerHost );
}

void
FfinderClient::callGetPerson()
{
//     callCaptureFrames();
    callGetResults();
    callCaptureFrames();
}

int
FfinderClient::callGetCurrent()
{
    get_current( leftPerson, leftClient, leftServerHost );
    get_current( rightPerson, rightClient, rightServerHost );
    if( currentFrameNumber == leftPerson.frameNumber )
       return 1;
    currentFrameNumber = leftPerson.frameNumber;
    return 0;
}
