// IvCameraView.h++

#ifndef _IvCameraView_H_
#define _IvCameraView_H_

#include <Inventor/Xt/SoXtRenderArea.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoCallback.h>
#include <Inventor/nodes/SoSpotLight.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoTransformSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/events/SoMouseButtonEvent.h>
#include <Inventor/SoOffscreenRenderer.h>
#include <Inventor/draggers/SoJackDragger.h>
#include <Inventor/draggers/SoTranslate1Dragger.h>
#include <Inventor/sensors/SoNodeSensor.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <X11/IntrinsicP.h>
#include "QuickMath/qmMinMax.h++"

class IvCameraView
{
private:
    static const int defaultHeight;
    static const int defaultWidth;
    int height;
    int width;
    SoXtRenderArea* renderArea;
    SoSeparator* root;
    SoPerspectiveCamera* camera;
    SoSpotLight* light;
    SoLightModel* lightModel;
    SoTransformSeparator* sceneRoot;
    SoTransform* coordinateTransform;
    SoEventCallback* keyEventNode;
    int (*userKeyCB)( void*, SoEventCallback* );
    void* userKeyData;
    static SoEventCallbackCB keypressCB;
    int (*userMouseCB)( void*, SoEventCallback* );
    void* userMouseData;
    static SoEventCallbackCB mousepressCB;
    void* backgroundData;
    GLenum backgroundFormat;
    SoCallback* backgroundNode;
    static SoCallbackCB backgroundCB;
    static SoSensorCB interiorOrientationSensorCB;
    static SoSensorCB referenceTransformSensorCB;
    static SoSensorCB coordinateTransformSensorCB;
    SoNodeSensor* interiorOrientationSensor;
    SoNodeSensor* relativeOrientationSensor;

public:
    IvCameraView::IvCameraView( Widget = NULL );
    IvCameraView::~IvCameraView();
    inline SoXtRenderArea* getRenderArea() { return renderArea; }
    inline Widget getWidget() { return renderArea->getWidget(); }
    void setSize( int h, int w, int y, int x );
    inline void addToView( SoNode* );
    inline void setTitle( char *s ) { renderArea -> setTitle( s ); }
    void setBackground( void*, GLenum = GL_RGB );
    void setBackgroundColor( float r, float g, float b );
    void setLightModel( int model ); // SoLightModel::{BASE_COLOR,PHONG}
    inline void setKeyCB( int (*)(void*, SoEventCallback*), void* = 0 );
    inline void setMouseCB( int (*)(void*, SoEventCallback*), void* = 0 );
    void setInteriorOrientation( float beta );
    inline void setReferenceTransform( SbMatrix& );
    inline void setCoordinateTransform( SbMatrix& );
    void setCoordinateTransform( float tx, float ty, float tz,
				 float q0, float q1, float q2, float q3 );
    void attachInteriorOrientation( SoPerspectiveCamera* );
    void attachReferenceTransform( SoTransform* );
    void attachCoordinateTransform( SoTransform* );
    int findColorBlob( SbColor&, float* mean, float* cov );
    void handle();
    SoSeparator* getRoot() { return root; }
};

inline
void
IvCameraView::addToView( SoNode* node )
{
    sceneRoot -> addChild( node );
}

inline
void
IvCameraView::setReferenceTransform( SbMatrix& m )
{
    coordinateTransform -> setMatrix( m.inverse() );
}

inline
void
IvCameraView::setCoordinateTransform( SbMatrix& m )
{
    coordinateTransform -> setMatrix( m );
}

inline 
void
IvCameraView::setKeyCB( int (*fn)(void*, SoEventCallback* ),
		    void* data )
{
    userKeyCB = fn;
    userKeyData = data;
}

inline 
void
IvCameraView::setMouseCB( int (*fn)(void*, SoEventCallback* ),
			  void* data )
{
    userMouseCB = fn;
    userMouseData = data;
}

#endif /* _IvCameraView_H_ */
