// PersonViewer.h++

#ifndef _PersonViewer_H_
#define _PersonViewer_H_

#include <Inventor/Xt/SoXt.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoSeparator.h>
#include "IvStereoCamera.h++"
#include "BlobAxes.h++"
#include "Hand.h++"
#define QmDouble
#include "QuickMath/QuickMath.h++"
#include "PurplePoint.h++"

class PersonViewer
{
    static BlobAxes* triangleAxes;
    SoSeparator* triangleRoot[3];
    SoMaterial* triangleMaterial[3];
    SoSphere* triangleSphere;
    SoTransform* triangleXform[3];
    int triangleIndex[3];
    static float triColor[3][3];
    Hand* hands[2];
    PurplePoint* quickPoints[6];

    SoSeparator* worldRoot;
    SoTransform* worldXform;
    SbVec3f worldT;
    SbRotation worldR;
    SbVec3f worldS;
    Widget window;
    IvStereoCamera* view;
    static int keyCB( void*, SoEventCallback* );
    static ostream& alert( char* fn, char* msg );
    static void error( char* fn, char* msg );

 public:
    static int stopInput;
    static int jogInput;
    static int quitFlag;
    static int collectFlag;
    static int calibrateFlag;
    static float rayDepth;
    static int qpMode;
    static float qpSize;
    enum { NOMODE, TRIANGLE, BLOB, YAKKA };
    static int mode;
    PersonViewer();
    Widget init( char* name, int w, int h );
    void initCalibration( float tx, float ty, float tz,
			  float q0, float qx, float qy, float qz,
			  float beta );
    inline void initCalibration( float data[8] );
    istream& scanViewTransform( istream& );
    int scanViewTransform( char* );
    void setViewTransform( real* t, real* q, real ha );
    void handle();
    void set( qmVectorRef* y, 
	      int* valid,
	      qmVectorRef* t,
	      qmUnitQuaternionRef* q,
	      qmVectorRef* l );
    void setHands( int which[2], int minus[2] );
    void setWorldTransform( real* t, real* r, real scale );
    void initWorld( char* filename, char* pathname = "." );
    void dumpScene();
    void setExistence( int );
    Display* getDisplay() { return SoXt::getDisplay(); }
    Window getWindow() { return XtWindow( window ); }
    void setFrameNumber( int );
    void setMessage( char* );
    inline void clearMessage() { setMessage( "" ); }
    void setQuickPoints( real*, real s=qpSize ); // 18 floats
};

inline
void
PersonViewer::initCalibration( float data[8] )
{
    initCalibration( data[0], data[1], data[2],
		     data[3], data[4], data[5], data[6],
		     data[7] );
}

#endif /* _PersonViewer_H_ */
