// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.  These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.

// noLevenbergMarquardt.h++

#ifndef _noLevenbergMarquardt_H_
#define _noLevenbergMarquardt_H_

#include <math.h>
#include "noOptimizer.h++"
#include "noErrorFunction.h++"
#include "QuickMath/QuickMath.h++"

class noLevenbergMarquardt : public noOptimizer
{
    static real defaultU;
    static real defaultV;
    static real defaultTolerance;
    static real defaultAbsTolerance;
protected:
    noErrorFunction* erf;
    real u;
    real v;
    qmVector dy;
    qmVector Rivec;
    qmVector x;
    qmMatrix J;
    qmVector dx;
    qmVector new_x;
    qmVector new_dy;
    real new_loss;
    int reduced;
    real loss;
    qmVector JtDy;
    qmVector RiDy;
    qmMatrix RiJ;
    real magJtDy;
    real gradientTolerance;
    real absoluteTolerance;
    qmVector solveDx;
    qmMatrix solveH;
    qmMatrix solveJtJ;
    qmMatrix solveuI;
    qmVector solveJtDy;
    int verbose;
    int iterationLimit;
    int echoExitStats;
    int estimateStatus;
    int solveStatus;
protected:
    real computeLoss( qmVectorRef& dy );
    const qmVectorRef& solve( qmMatrixRef& J, qmVectorRef& dy, real u );
public:
    enum { SolveSuccess, SolveFail };
    enum { None, Success, Overflow, Singular };
    noLevenbergMarquardt();
    void init( noErrorFunction& a );
    const qmVectorRef& estimate();
    inline void setGradientTolerance( real );
    inline void setAbsoluteTolerance( real );
    inline void setRelativeTolerance( const qmVectorRef& );
    inline void setState( const qmVectorRef& );
    inline void setVerbose( int v = 1 ) { verbose = v; }
    inline void setIterationLimit( int l ) { iterationLimit = l; }
    inline void setEchoExitStats( int e = 1 ) { echoExitStats = e; }
    inline int getEstimateStatus() { return estimateStatus; }
    inline int estimateOK() { return( (Success==estimateStatus) ? 1 : 0 ); }
};

inline
void
noLevenbergMarquardt::setGradientTolerance( real tol )
{
    gradientTolerance = tol;
}

inline
void
noLevenbergMarquardt::setAbsoluteTolerance( real tol )
{
    absoluteTolerance = tol;
}

inline
void
noLevenbergMarquardt::setRelativeTolerance( const qmVectorRef& ri )
{
    Rivec.copy( ri );
}

inline
void
noLevenbergMarquardt::setState( const qmVectorRef& x_set )
{
    x = x_set;
}

#endif /* _noLevenbergMarquardt_H_ */
