// worldBaby.c++

#include "Parse/Parse.h++"

static int verbose = 0;

#define QmDouble
#include "qmStatistics/qmStatistics.h++"
#include "PersonPlus.h++"

static
inline
void
readAndCompute( char* filename,
		qmVectorRef& mean,
		qmVectorRef& direction,
		real* length = 0 )
{
    // read in points

    PersonPlus personPlus;
    sffPersonPlus person;
    qmSampleAcc xline( 3 );
    qmGaussian xlineStats( 3 );
    xline.setZero();
//     int forceFlag = 1;
//     personPlus.undump_force( filename, &forceFlag );
//     int ringCount = personPlus.getRingLength();
    personPlus.readRing( filename );
    int ringCount = personPlus.getRingLength();
    real* data = new real[ 3 * ringCount ];
    int pointCount = 0;
    person = personPlus.get();
    int firstFrame = person.frameNumber;
    do
       {
	   if( ! person.rightHand.valid )
	      {
		  cerr << "  INVALID POINT" << endl;
		  continue;
	      }
	   qmVectorRef( 3, data+3*pointCount ).copy
	      ( qmVectorRef( 3, person.rightHand.camera_ref.t ) );
	   // xline.addSample( qmVectorRef( 3, person.rightHand.camera_ref.t ) );
	   xline.addSample( qmVectorRef( 3, data+3*pointCount ) );
	   pointCount++;
	   // forceFlag = 1, personPlus.undump_force( filename, &forceFlag );
	   personPlus.undump_force();
	   person = personPlus.get();
       }
    while( person.frameNumber != firstFrame );
    xline.getGaussian( xlineStats );
    mean.copy( xlineStats.getMean() );
    qmMatrix cov = xlineStats.getCovariance();
    qmEigenStructure eig( cov );
    qmMatrix evectors = eig.getEigVectors();
    qmVector evalues = eig.getEigValues();
    if( verbose )
       {
	   cerr << "\teval= " << evalues << endl;
	   evectors.print( cerr << "\tevec= ", "\n\t" ) << endl;
       }
    int which;
    evalues.getMax( &which );
    direction.copy( evectors.getColumn(which) );
    
    if( length )
       {
	   qmVector dist( pointCount );
	   qmVector y_ym( 3 );
	   for( int i = 0; i < pointCount; i++ )
	      dist(i) = y_ym.setSubtract
		 ( qmVectorRef(3,data+i*3), mean ).
		 getInnerProduct( direction );
	   *length = dist.getMax() - dist.getMin();
       }
}

int
main( int argc, char *argv[] )
{
    char* xfilename = 0;
    char* yfilename = 0;
    char* sfilename = 0;
    real inches = 12;

    ParseRuleFlag( 'v', verbose, 1, "Verbose mode" );
    ParseRuleArg( xfilename, "xfile", "xfile" );
    ParseRuleArg( yfilename, "yfile", "yfile" );
    ParseRuleArg( sfilename, "sfile", "sfile" );
    ParseRuleArgFloat( inches, "inches", "inches" );
    
    if( Parse::ParseCmdLine( argc, argv ) )
       Parse::Usage( argv[0] ), exit(1);

    // compute xm, xhat
    //  xm is just the mean
    //  xhat is the eigenvector of the covariance corresp to largest evalue
    qmVector xmean(3);
    qmVector xdir(3);
    readAndCompute( xfilename, xmean, xdir );
    if( verbose )
       {
	   cerr << "xmean= " << xmean << endl;
	   cerr << "xdir= " << xdir << endl << endl;
       }

    qmVector ymean(3);
    qmVector ydir(3);
    readAndCompute( yfilename, ymean, ydir );
    if( verbose )
       {
	   cerr << "ymean= " << ymean << endl;
	   cerr << "ydir= " << ydir << endl;
       }

    qmVector smean(3);
    qmVector sdir(3);
    real lengthInUnits;
    readAndCompute( sfilename, smean, sdir, &lengthInUnits );
    if( verbose )
       {
	   cerr << "smean= " << smean << endl;
	   cerr << "sdir= " << sdir << endl;
       }

    xdir.setNormalize();  // even though they should both be normal already
    if( xdir(0) > 0 )
       xdir.setNegateEq();
    ydir.setNormalize();
    if( ydir(1) > 0 )
       ydir.setNegateEq();
    
    qmVector yp( 3 );
    real a = xdir.getInnerProduct( ydir );
    cerr << "\ta= " << a << endl;
    yp.copy( xdir );
    yp *= -a;
    yp += ydir;
    yp.setNormalize();

    qmVector zdir( 3 );
    zdir.setCross( xdir, yp );

    // compute origin
    //  A = [ xhat  -yhat ]
    //  b = (ym - xm)
    //  s = (A`A)^-1 A`b
    //  origin = xm + s(0)*xhat
    qmMatrix A( 3, 2 );
    qmVector tmp( 3 );

    A.setColumn( 0, xdir );
    A.setColumn( 1, tmp.setNegate( yp ) );
    tmp.setSubtract( ymean, xmean );
    qmMatrix AtA( 2, 2 );
    AtA.setTMultiply( A, A );
    AtA.setInvertEq();
    qmVector tmp2( 2 );
    tmp2.setTMultiply( A, tmp );
    qmVector s( 2 );
    s.setMultiply( AtA, tmp2 );
    qmVector origin( 3 );
    origin.copy( xdir );
    origin *= s( 0 );
    origin += xmean;
    cout << origin << endl;

    // compute scale

    // output T = origin
    // output R = [xhat`
    //             yhat`
    //             zhat`]
    // output scale = 
    qmMatrix R( 3, 3 );
    R.setRow( 0, xdir );
    R.setRow( 1, yp );
    R.setRow( 2, zdir );
    cout << R;

    real scale = inches / lengthInUnits;
    cout << scale << endl;

    return 0;
}
