// vlImage.c++

#include "vlImage.h++"
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

VLServer vlImage::server = 0;
VLNode vlImage::drain = 0;
VLNode vlImage::source = 0;
VLPath vlImage::path = 0;
VLBuffer vlImage::buffer = 0;
VLInfoPtr vlImage::info = 0;
VLControlValue vlImage::val;
int vlImage::transferBegun = 0;
int vlImage::captureWidth = 0;
int vlImage::captureHeight = 0;
int vlImage::captureX = 0;
int vlImage::captureY = 0;
int vlImage::captureSize = 0;
int vlImage::captureType = vlImage::RGB;
int vlImage::captureBytesPerPixel = vliRGBBytesPerPixel;
int vlImage::verbose = 0;

#if 0
int vlImage::capture_type = vlImage::RGB;
int vlImage::nbuffers = 4;
int vlImage::input = vlImage::SVIDEO_1;
int vlImage::zoom = 1;
#endif

static
void
error( char *message )
{
  char tag[128];
  sprintf( tag, "vlImage: %s", message );
  vlPerror( tag );
  vlImage::close();
  exit( 1 );
}

static
void
warning( char *message )
{
  char tag[128];
  sprintf( tag, "vlImage: %s", message );
  vlPerror( tag );
}

int
findIndyAnalogVideoInput( VLServer server )
{
    VLDevList devList;
    vlGetDeviceList( server, &devList );
    for( int i = 0; i < devList.devices[0].numNodes; i++ )
       if( !strcmp( devList.devices[0].nodes[i].name, "Analog Video Input" ) )
	  return devList.devices[0].nodes[i].number;
    return -1;
}

int
findO2SVideoInput( VLServer server )
{
    VLDevList devList;
    vlGetDeviceList( server, &devList );
    for( int i = 0; i < devList.devices[0].numNodes; i++ )
       if( !strcmp( devList.devices[0].nodes[i].name, "SVideo Input" ) )
	  return devList.devices[0].nodes[i].number;
    return -1;
}

int
findO2CompositeInput( VLServer server )
{
    VLDevList devList;
    vlGetDeviceList( server, &devList );
    for( int i = 0; i < devList.devices[0].numNodes; i++ )
       if( !strcmp( devList.devices[0].nodes[i].name, "Composite Video Input"))
	  return devList.devices[0].nodes[i].number;
    return -1;
}

int
findDigitalVideoInput( VLServer server )
{
    VLDevList devList;
    vlGetDeviceList( server, &devList );
    char* name = "Bogus";
    if( !strcmp( devList.devices[0].name, "ev1" ) )
       // we are on Galileo
       {
	   name = "IndyCam";
       }
    else if( !strcmp( devList.devices[0].name, "vino" ) )
       // we are on Indy
       {
	   name = "Digital Video Input";
       }
    for( int i = 0; i < devList.devices[0].numNodes; i++ )
       if( !strcmp( devList.devices[0].nodes[i].name, name ) )
	  return devList.devices[0].nodes[i].number;
    return -1;
}

#if 0
int
findAllInputs( VLServer server )
{
    int count = 0;
    VLDevList devList;
    vlGetDeviceList( server, &devList );
    for( int i = 0; i < devList.devices[0].numNodes; i++ )
       if( devList.devices[0].nodes[i].type == VL_SRC &&
	   devList.devices[0].nodes[i].kind == VL_VIDEO )
	  {
	      cerr << devList.devices[0].nodes[i].name << "= "
		   << devList.devices[0].nodes[i].number << endl;
	      count++;
	  }
    return count;
}
#endif

void
vlImage::init( int input, int zoom, int encoding, int aord, int nbuffers,
	       int cropx, int cropy, int cropw, int croph)
{
  if( transferBegun )
    return;
  if( !( server = vlOpenVideo("") ) )
    error( "open video" );
  drain = vlGetNode( server, VL_DRN, VL_MEM, VL_ANY );
  int inumber;
  switch (aord) 
    {
    case O2SVIDEO: 
      inumber = findO2SVideoInput( server );
      break;
    case O2COMPOSITE:
      inumber = findO2CompositeInput( server );
      break;
    case INDYANALOG: 
      inumber = findIndyAnalogVideoInput( server );
      break;
    case DIGITAL: 
      inumber = findDigitalVideoInput( server );
      break;
    default:
      inumber = -1;
    }
  if( -1 == inumber )
    error( "failed to find analog or digital input" );
  // source = vlGetNode( server, VL_SRC, VL_VIDEO, VL_ANY );
  source = vlGetNode( server, VL_SRC, VL_VIDEO, inumber );
  path = vlCreatePath( server, VL_ANY, source, drain );
  if( ( vlSetupPaths( server, 
		      (VLPathList)&path, 1, VL_SHARE, VL_SHARE ) ) < 0 )
    error( "setup paths" );
  //setInput( input );
  setEncoding( encoding );
  setZoom( zoom );
  if ((cropw != -1) && (croph != -1))
      setCrop( cropx, cropy, cropw, croph );
  buffer = vlCreateBuffer( server, path, drain, nbuffers );
  if( !buffer )
    error( "create buffers" );
  vlRegisterBuffer( server, path, drain, buffer );
  if( vlBeginTransfer( server, path, 0, NULL ) )
    error( "begin transfer" );
  transferBegun = 1;
}

void
vlImage::close()
{
    if( !transferBegun )
       return;
    vlEndTransfer( server, path );
    vlDeregisterBuffer( server, path, drain, buffer );
    vlDestroyBuffer( server, buffer );
    vlDestroyPath( server, path );
    vlCloseVideo( server );
    transferBegun = 0;
}

#if 0
int
vlImage::NBuffers( int n )
{
  if( QUERY != n && !transferBegun )
     nbuffers = n;
  return nbuffers;
}
#endif

int
vlImage::getEncoding()
{
    if( !server )
       return -1;
    vlGetControl( server, path, drain, VL_PACKING, &val );
    return val.intVal;
}

void
vlImage::setEncoding( int mode )
{
    if( !server )
       return;
    switch( mode )
       {
	  case YUV: 
	   val.intVal = VL_PACKING_YVYU_422_8; 
	   captureType = YUV;
	   captureBytesPerPixel = vliYUVBytesPerPixel;
	   break;
	  case RGB: default: 
	   val.intVal = VL_PACKING_RGB_8; 
	   captureType = RGB;
	   captureBytesPerPixel = vliRGBBytesPerPixel;
	   break;
       }
    if( vlSetControl( server, path, drain, VL_PACKING, &val ) < 0 )
       error( "set packing control" );
}

int
vlImage::getZoom( )
{
    if( !server )
       return -1;
    vlGetControl( server, path, drain, VL_ZOOM, &val );
    return val.intVal;
}

void
vlImage::setZoom( int decimation )
{
    if( !server )
       return;
    val.fractVal.numerator = 1;
    val.fractVal.denominator = decimation;
    if( vlSetControl( server, path, drain, VL_ZOOM, &val ) < 0 )
       error( "set zoom control" );
    vlGetControl( server, path, drain, VL_SIZE, &val );
    captureHeight = val.xyVal.y;
    captureWidth = val.xyVal.x;
    captureSize = captureHeight * captureWidth;
}

void
vlImage::setCrop( int x, int y, int width, int height )
{
    if( !server )
       return;

    captureY = val.xyVal.y = y;
    captureX = val.xyVal.x = x;

    if( vlSetControl( server, path, drain, VL_OFFSET, &val ) < 0 )
       warning( "set OFFSET control" );

    vlGetControl( server, path, drain, VL_OFFSET, &val );
    captureY = val.xyVal.y;
    captureX = val.xyVal.x;
    cerr << "set OFFSET to " << captureX << " " << captureY << endl;

    val.xyVal.y = height;
    val.xyVal.x = width;

    if( vlSetControl( server, path, drain, VL_SIZE, &val ) < 0 )
      warning( "set SIZE control" );

    vlGetControl( server, path, drain, VL_SIZE, &val );
    captureHeight = val.xyVal.y;
    captureWidth = val.xyVal.x;
    cerr << "set SIZE to " << captureWidth << " " << captureHeight << endl;
}

int
vlImage::getInput()
{
    if( !server )
       return -1;
    vlGetControl( server, path, source, VL_MUXSWITCH, &val );
    return val.intVal;
}

void
vlImage::setInput( int in )
{
    if( !server )
       return;
    val.intVal = in;
    if ( vlSetControl( server, path, source, VL_MUXSWITCH, &val ) < 0 )
      {
	warning( "set source control" );
      }
}

void
vlImage::setVerbose( int in )
{
    // Only set if 0 or 1; else just echo value
    if( 0 == in )
       verbose = 0;
    if( 1 == in )
       verbose = 1;
}

vliImageRef
vlImage::blindGrab()
{
    if( !transferBegun )
       error( "blindGrab: attempt to grab before initialized" );
    info = (VLInfoPtr) NULL;
    while( !( info = vlGetLatestValid( server, buffer ) ) )
      {
	if (vlBufferDone(buffer) < 0)
	  {
	    warning("didn't get a frame");
	  }
	else
	  {
	    if ( vlGetErrno() != VLSuccess )
	      warning("vlBufferDone");
	  }
	sginap( 1 );
      }
    byte* dataPtr = (byte *) vlGetActiveRegion( server, buffer, info );
    vliImageRef iref( captureWidth,
		      captureHeight,
		      captureBytesPerPixel,
		      dataPtr );
    return iref;
}

void
vlImage::grab( vliRGBImageRef& rgb )
{
    if( RGB == captureType )
       rgb = blindGrab();
    else
       error( "grab: attempt to assign non-RGB image to vliRGBImageRef" );
}

void
vlImage::grab( vliYUVImageRef& yuv )
{
    if( YUV == captureType )
       yuv = blindGrab();
    else
       error( "grab: attempt to assign non-YUV image to vliYUVImageRef" );
}

void
vlImage::free()
{
    vlPutFree( server, buffer );
}

void
vlImage::yuvToRgb( const vliYUVImageRef& yuv_image, vliRGBImageRef& rgb_image )
{
    // could check dimensions here
    vliYUVPixel yuv;
    int pixelCount = rgb_image.getPixelCount();
    for( int i = 0; i < pixelCount; i++ )
       {
	   vliRGBPixel& rgb = rgb_image.getPixel( i );
	   yuv = yuv_image.getPixel(i);
	   vliPixel::yuvToRgb
	      ( yuv.y, yuv.u, yuv.v, rgb.r(), rgb.g(), rgb.b() );
       }
}

void
vlImage::yuvToRgb2( const vliYUVImageRef& yuv_image,
		    vliRGBImageRef& rgb_image )
{
    // could check dimensions here
    vliYUVPixel yuv;
    int pixelCount = rgb_image.getPixelCount();
    for( int i = 0; i < pixelCount; i++ )
       {
	   vliRGBPixel& rgb = rgb_image.getPixel( i );
	   yuv = yuv_image.getPixel(i);
	   vliPixel::yuvToRgb2
	      ( yuv.y, yuv.u, yuv.v, rgb.r(), rgb.g(), rgb.b() );
       }
}

void
vlImage::rgbToYuv( const vliRGBImageRef& rgb_image, vliYUVImageRef& yuv_image )
{
    // could check dimensions here
    vliYUVPixel yuv;
    int pixelCount = yuv_image.getPixelCount();
    for( int i = 0; i < pixelCount; i++ )
       {
	   rgbToYuv( rgb_image.getPixel( i ), yuv );
	   yuv_image.setPixel( i, yuv );
       }
}

