// vliImageRef.h++

#ifndef _vliImageRef_H_
#define _vliImageRef_H_

#include "vlImage/vliTypes.h++"
#include <bstring.h>
#include <stream.h>

class vliImageRef
{
protected:
    int pixelCount;
    int width;
    int bytesPerPixel;
    byte* data;
public:
    enum { NOTYPE=0, RGB=1, YUV=2 };
    vliImageRef( )
      : pixelCount( 0 ), width( 0 ), bytesPerPixel( 0 ), data( 0 ) {}
    vliImageRef( int w, int h, int bpp, byte* p )
      : pixelCount( w*h ), width( w ), bytesPerPixel( bpp ), data( p ) {}
    vliImageRef( const vliImageRef& im )
      : pixelCount( im.pixelCount ),
	    width( im.width ),
	    bytesPerPixel( im.bytesPerPixel ),
	    data( im.data ) {}
    virtual istream& readRawRGB( istream& is ) {return is;}//= 0;
    virtual ostream& writeRawRGB( ostream& os ) {return os;}//= 0;
    inline istream& read( istream& );
    inline ostream& write( ostream& );
    inline byte* getPixelData( int x, int y );
    inline byte* getPixelData( int i = 0 );
    inline int copy( const vliImageRef& );
    inline int getPixelCount() const { return pixelCount; }
    inline int getWidth() const { return width; }
    inline int getHeight() const { return( (width) ? pixelCount/width : 0 ); }
    inline int getBytesPerPixel() const { return bytesPerPixel; }
    virtual int getType() { return NOTYPE; }
    inline void clear() { bzero( data, pixelCount * bytesPerPixel ); }
    virtual void black() { bzero( data, pixelCount * bytesPerPixel ); }
    // these are sort of bogus cuz packing is not always like this
    inline void setPixel( int i, byte* );
    inline void setPixel( int x, int y, byte* );
    inline void setPixelSafe( int x, int y, byte* );
};

inline
byte*
vliImageRef::getPixelData( int i )
{
    return( data + i * bytesPerPixel );
}

inline
byte*
vliImageRef::getPixelData( int x, int y )
{
    return( data + (y*width+x) * bytesPerPixel );
}

inline
int
vliImageRef::copy( const vliImageRef& src )
{
    if( pixelCount != src.pixelCount ||
	width != src.width ||
	bytesPerPixel != src.bytesPerPixel )
       return 1;
    bcopy( src.data, data, pixelCount*bytesPerPixel );
    return 0;
}

inline
istream&
vliImageRef::read( istream& is )
{
    return is.read( data, pixelCount*bytesPerPixel );
}

inline
ostream&
vliImageRef::write( ostream& os )
{
    return os.write( data, pixelCount*bytesPerPixel );
}

inline
void
vliImageRef::setPixel( int i, byte* pixel )
{
    bcopy( pixel, data + i* bytesPerPixel, bytesPerPixel );
}

inline
void
vliImageRef::setPixel( int u, int v, byte* pixel )
{
    setPixel( v*width + u, pixel );
}

inline
void
vliImageRef::setPixelSafe( int u, int v, byte* pixel )
{
    if( u < 0 || v < 0 || u >= width || v*width >= pixelCount )
       return;
    setPixel( v*width + u, pixel );
}

#endif /* _vliImageRef_H_ */
