// vliPixel.c++

#include "vlImage/vliPixel.h++"

#include <stream.h>

// inline
void
vliPixel::yuvToRgb( byte y, byte u, byte v, byte& r, byte& g, byte& b )
{
    float red = (1.164*(y-16.0) + 1.596*(v-128.0));
    float blue = (1.164*(y-16.0) + 2.017*(u-128.0));
    float green = (1.164*(y-16.0) - 0.392*(u-128.0) - 0.813*(v-128.0));
    r = (byte) limit( red, 0., 255. );
    g = (byte) limit( green, 0., 255. );
    b = (byte) limit( blue, 0., 255. );
    // cerr << "yuv= " << (int)y << ", " << (int)u << ", " << (int)v << endl;
    // cerr << " rgb= " << red << ", " << green << ", " << blue << endl;
    // cerr << " rgb= " << (int)r << ", " << (int)g << ", " << (int)b << endl;
}

static const float compress = .9;
static const float offset = 30;

// inline
void
vliPixel::yuvToRgb2( byte y, byte u, byte v, byte& r, byte& g, byte& b )
{
    r = (byte) (compress*(1.164*(y-16.0) + 1.596*(v-128.0)) + offset);
    g = (byte) (compress*(1.164*(y-16.0) - 0.392*(u-128.0) - 0.813*(v-128.0))
		+ offset);
    b = (byte) (compress*(1.164*(y-16.0) + 2.017*(u-128.0)) + offset );
}

// inline
void
vliPixel::rgbToYuv( byte r, byte g, byte b, byte& y, byte& u, byte& v )
{
#if 0
    float yf = .299 * r + .587 * g + .114 * b;
    float uf =  .5 * ( b + 227. - yf );
    float vf =  .5 * ( r + 179. - yf );
#endif
#if 0
    float yf = .2569 * r + .5042 * g + .0980 * b - 88.9476;
    float uf =  -0.1482*r - 0.2910*g + 0.4392*b + 188.5647;
    float vf =  0.4392*r - 0.3678*g - 0.0715*b - 51.4593;
#endif
    float yf = .2569 * r + .5042 * g + .0980 * b + 16.;
    float uf =  -0.1482*r - 0.2910*g + 0.4392*b + 128.;
    float vf =  0.4392*r - 0.3678*g - 0.0715*b + 128.;
    y = (byte) limit( yf, 0., 255. );
    u = (byte) limit( uf, 0., 255. );
    v = (byte) limit( vf, 0., 255. );
}

// inline
void
vliPixel::rgb2ToUyvy( byte r1, byte g1, byte b1,
		      byte r2, byte g2, byte b2,
		      byte& u, byte& y1, byte& v, byte& y2 )
{
    float y1f = .299 * r1 + .587 * g1 + .114 * b1;
    float u1f_2 =  .25 * ( b1 + 227. - y1f );
    float v1f_2 =  .25 * ( r1 + 179. - y1f );
    float y2f = .299 * r2 + .587 * g2 + .114 * b2;
    float u2f_2 =  .25 * ( b2 + 227. - y2f );
    float v2f_2 =  .25 * ( r2 + 179. - y2f );
    u = (byte) limit( u1f_2 + u2f_2, 0., 255. );
    y1 = (byte) limit( y1f, 0., 255. );
    v = (byte) limit( v1f_2 + v2f_2, 0., 255. );
    y2 = (byte) limit( y2f, 0., 255. );
}

#if 0
inline
void
vliPixel::yuvToRgb( byte y, byte u, byte v, byte& r, byte& g, byte& b )
{
    float red = (1.164*(y-16.0) + 1.596*(v-128.0));
    float blue = (1.164*(y-16.0) + 2.017*(u-128.0));
    float green = (1.164*(y-16.0) - 0.392*(u-128.0) - 0.813*(v-128.0));
    r = (byte) limit( red, 0., 255. );
    g = (byte) limit( green, 0., 255. );
    b = (byte) limit( blue, 0., 255. );
    // cerr << "yuv= " << (int)y << ", " << (int)u << ", " << (int)v << endl;
    // cerr << " rgb= " << red << ", " << green << ", " << blue << endl;
    // cerr << " rgb= " << (int)r << ", " << (int)g << ", " << (int)b << endl;
}

inline
void
vliPixel::rgbToYuv( byte r, byte g, byte b, byte& y, byte& u, byte& v )
{
#if 0
    float yf = .299 * r + .587 * g + .114 * b;
    float uf =  .5 * ( b + 227. - yf );
    float vf =  .5 * ( r + 179. - yf );
#endif
#if 0
    float yf = .2569 * r + .5042 * g + .0980 * b - 88.9476;
    float uf =  -0.1482*r - 0.2910*g + 0.4392*b + 188.5647;
    float vf =  0.4392*r - 0.3678*g - 0.0715*b - 51.4593;
#endif
    float yf = .2569 * r + .5042 * g + .0980 * b + 16.;
    float uf =  -0.1482*r - 0.2910*g + 0.4392*b + 128.;
    float vf =  0.4392*r - 0.3678*g - 0.0715*b + 128.;
    y = (byte) limit( yf, 0., 255. );
    u = (byte) limit( uf, 0., 255. );
    v = (byte) limit( vf, 0., 255. );
}

inline
void
vliPixel::rgb2ToUyvy( byte r1, byte g1, byte b1,
		      byte r2, byte g2, byte b2,
		      byte& u, byte& y1, byte& v, byte& y2 )
{
    float y1f = .299 * r1 + .587 * g1 + .114 * b1;
    float u1f_2 =  .25 * ( b1 + 227. - y1f );
    float v1f_2 =  .25 * ( r1 + 179. - y1f );
    float y2f = .299 * r2 + .587 * g2 + .114 * b2;
    float u2f_2 =  .25 * ( b2 + 227. - y2f );
    float v2f_2 =  .25 * ( r2 + 179. - y2f );
    u = (byte) limit( u1f_2 + u2f_2, 0., 255. );
    y1 = (byte) limit( y1f, 0., 255. );
    v = (byte) limit( v1f_2 + v2f_2, 0., 255. );
    y2 = (byte) limit( y2f, 0., 255. );
}
#endif
